;; File Version: 1.2
;; Last Modified: December 18, 2007
;; By: AWD

PRO CALL_RINGMAP, mainuval, data, cube, geometer_base

COMPILE_OPT idl2

;; Procedure must be called by GEOMETER. Builds a temporary widget to 
;; collect user preferences for the DIRECT_RINGMAP plot. Put into this
;; separate procedure from GEOMETER_EVENT where it was originally 
;; developed.
;; Josh Colwell
;; July 31, 2006
;; December 6, 2006: Allow for grid to be plotted on top of pixels.
;; January 16, 2007: Implement graphic state vector.
;; June 16, 2007: Implement map time. Modify for Geometer version of cube.
;; December 18, 2007: Updated to allow multiple output file formats. 

;; Create a window to get plotting preferences.
ringmapbase = WIDGET_BASE(/COLUMN, TITLE = 'Direct Ring Map Options')

;; Get the target ring limb limits and double it for the 
;; "zoomed in" option (triple if target is Saturn to 
;; allow room for rings).
IF mainuval.cubeloaded THEN limbtime = cube[0].et ELSE $
  limbtime = mainuval.etstart
targetid = WIDGET_INFO(mainuval.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, mainuval.targetdrop, GET_VALUE = $
                targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])

BODY_LIMB, target, 'CASSINI', limbtime, 90, $
           ra_limb, dec_limb, limberror



radbase = WIDGET_BASE(ringmapbase, /ROW)
quadwidget = CW_FIELD(radbase, TITLE = 'Quadrant: ',$
                       UVALUE = 'QUADRANT', VALUE = $
                       STRTRIM(STRING(0),2), XSIZE = 3)
radminwidget = CW_FIELD(radbase, TITLE = 'Min. radius: ',$
                       UVALUE = 'RADMINTEXT', VALUE = $
                       STRTRIM(STRING(0),2), XSIZE = 8)
radmaxwidget = CW_FIELD(radbase, TITLE = 'Max. radius: ',$
                       UVALUE = 'RADMAXTEXT', VALUE = $
                       STRTRIM(STRING(1.5e5),2), XSIZE = 8)

;; Check to see if values from a previous run of target exist, and if so,
;; use those values.
IF mainuval.ringstate.use THEN BEGIN
    WIDGET_CONTROL, quadwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.ringstate.quad),2)
    WIDGET_CONTROL, radminwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.ringstate.radmin),2)
    WIDGET_CONTROL, radmaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.ringstate.radmax),2)
ENDIF

IF cube[0].et GT 1E19 THEN BEGIN
    specminred = '0'
    specmaxred = '0'
    specmingreen = '0'
    specmaxgreen = '0'
    specminblue = '0'
    specmaxblue = '0'
ENDIF ELSE BEGIN
    nspec = N_ELEMENTS(WHERE(cube[0].data[0,*] NE -1.))
    IF mainuval.ringstate.use THEN BEGIN
       specminred = STRTRIM(STRING(mainuval.ringstate.spec_bounds[0]),2)
       specmaxred = STRTRIM(STRING(mainuval.ringstate.spec_bounds[1]),2)
       specmingreen = STRTRIM(STRING(mainuval.ringstate.spec_bounds[2]),2)
       specmaxgreen = STRTRIM(STRING(mainuval.ringstate.spec_bounds[3]),2)
       specminblue = STRTRIM(STRING(mainuval.ringstate.spec_bounds[4]),2)
       specmaxblue = STRTRIM(STRING(mainuval.ringstate.spec_bounds[5]),2)
    ENDIF ELSE BEGIN
       specminred = '0'
       specmaxred = STRTRIM(STRING(nspec/3),2)
       specmingreen = STRTRIM(STRING(nspec/3+1),2)
       specmaxgreen = STRTRIM(STRING(nspec-nspec/3),2)
       specminblue = STRTRIM(STRING(nspec-nspec/3+1),2)
       specmaxblue = STRTRIM(STRING(nspec-1),2)
    ENDELSE
ENDELSE

plotspecbase = WIDGET_BASE(ringmapbase, /ROW, /FRAME)
plotrowselect = CW_FIELD(plotspecbase, TITLE = 'Plot Row: ', $
                         UVALUE = 'PLOTROW', VALUE = '0', XSIZE = 4)
plottimeselect = CW_FIELD(plotspecbase, TITLE = 'Plot Index: ', $
                          UVALUE = 'PLOTTIME', VALUE = '0', XSIZE =4)
plotbutton = WIDGET_BUTTON(plotspecbase, VALUE = 'Plot Spectrum', $
                           UVALUE = 'PLOTBUTTON')
IF NOT mainuval.cubeloaded THEN WIDGET_CONTROL, plotspecbase, SENSITIVE = 0

redbase = WIDGET_BASE(ringmapbase, /ROW)
redwidget = CW_FIELD(redbase, TITLE = 'Red channels (lo,hi): ',$
                     UVALUE = 'REDTEXT', VALUE = specminred+','+specmaxred, $
                     XSIZE = 8)
redmaxwidget = CW_FIELD(redbase, TITLE = 'Red maximum: ', UVALUE = $
                        'REDMAX', VALUE = '0', XSIZE = 6)

greenbase = WIDGET_BASE(ringmapbase, /ROW)
greenwidget = CW_FIELD(greenbase, TITLE = 'Green channels (lo,hi): ',$
                       UVALUE = 'GREENTEXT', VALUE = specmingreen+','+$
                       specmaxgreen, XSIZE = 8)
greenmaxwidget = CW_FIELD(greenbase, TITLE = 'Green maximum: ', UVALUE = $
                        'GREENMAX', VALUE = '0', XSIZE = 6)

bluebase = WIDGET_BASE(ringmapbase, /ROW)
bluewidget = CW_FIELD(bluebase, TITLE = 'Blue channels (lo,hi): ',$
                      UVALUE = 'BLUETEXT', VALUE = specminblue+','+$
                      specmaxblue, XSIZE = 8)
bluemaxwidget = CW_FIELD(bluebase, TITLE = 'Blue maximum: ', UVALUE = $
                        'BLUEMAX', VALUE = '0', XSIZE = 6)

;; Allow user to choose whether to draw lat/lon lines
baseforframe2 = WIDGET_BASE(ringmapbase, /ROW, /FRAME)
label = WIDGET_LABEL(baseforframe2, VALUE = 'Draw radial grid:')
gridbase = WIDGET_BASE(baseforframe2, /ROW, /EXCLUSIVE)
showgridbutton = WIDGET_BUTTON(gridbase, VALUE = 'Yes', $
                                 UVALUE = 'GRIDYES')
hidegridbutton = WIDGET_BUTTON(gridbase, VALUE = 'No', $
                                 UVALUE = 'GRIDNO')
WIDGET_CONTROL, showgridbutton, /SET_BUTTON
label = WIDGET_LABEL(baseforframe2, VALUE = 'Grid over pixels:')
gridontopbase = WIDGET_BASE(baseforframe2, /ROW, /EXCLUSIVE)
gridontop = WIDGET_BUTTON(gridontopbase, VALUE='Yes', $
                          UVALUE = 'GRIDTOP')
gridonbottom = WIDGET_BUTTON(gridontopbase, VALUE = 'No', $
                             UVALUE = 'GRIDBOTTOM')
WIDGET_CONTROL, gridontop, /SET_BUTTON
grid = 1B
gridtop = 1B

;; Y/N option for plotting planet shadow on rings
baseforframe1 = WIDGET_BASE(ringmapbase, /ROW, /FRAME)
label = WIDGET_LABEL(baseforframe1, VALUE = 'Show shadow: ')
shadowbase = WIDGET_BASE(baseforframe1, /ROW, /EXCLUSIVE)
showshadowbutton = WIDGET_BUTTON(shadowbase, VALUE = 'Yes', $
                                 UVALUE = 'SHADOWYES')
hideshadowbutton = WIDGET_BUTTON(shadowbase, VALUE = 'No', $
                                 UVALUE = 'SHADOWNO')
WIDGET_CONTROL, showshadowbutton, /SET_BUTTON
shadow = 1B

;; Allow user to select subset of available times:
timeselectbase = WIDGET_BASE(ringmapbase, /ROW, /FRAME)
ntimewidget = CW_FIELD(timeselectbase, TITLE = 'Number of time steps to plot', $
                       VALUE = '1', UVALUE = 'TIMETEXT', XSIZE = 8)

;; Have a series of Yes/No buttons for various Sky Map options.
;; Allow user to show all FOV pointings at once (but a single
;; background).
timesbase = WIDGET_BASE(ringmapbase, /ROW, /EXCLUSIVE)
alltimesbutton = WIDGET_BUTTON(timesbase, VALUE = $
                               'Show all times', $
                               UVALUE = 'ALLTIMES')
onetimebutton = WIDGET_BUTTON(timesbase, VALUE = $
                              'Show selected time', $
                              UVALUE = 'ONETIME')
WIDGET_CONTROL, onetimebutton, /SET_BUTTON
alltimes = 0B

;; Allow user to select an animation:
animbase = WIDGET_BASE(ringmapbase, /ROW, /EXCLUSIVE)
animyes = WIDGET_BUTTON(animbase, VALUE = 'Animate', $
                        UVALUE = 'ANIMYES')
animno = WIDGET_BUTTON(animbase, VALUE = 'Single Frame', $
                       UVALUE = 'ANIMNO')
WIDGET_CONTROL, animno, /SET_BUTTON
ringanim = 0B

;; Allow user to use data cube times (corresponding to 
;; data records) rather than entered delta-times:
cubebase = WIDGET_BASE(ringmapbase, /ROW, /EXCLUSIVE)
cubeyes = WIDGET_BUTTON(cubebase, VALUE = 'Use Data Times', $
                        UVALUE = 'CUBEYES')
cubeno = WIDGET_BUTTON(cubebase, VALUE = 'Use Entered Times', $
                       UVALUE = 'CUBENO', SENSITIVE=0)
IF(N_ELEMENTS(data) NE 0) THEN $
                    WIDGET_CONTROL, cubeno, SENSITIVE = 1
cubetimes = 0B

;; Allow user to choose a specific time at which to draw map
mapbase = WIDGET_BASE(ringmapbase, /ROW, /FRAME)
IF mainuval.cubeloaded THEN $
  CSPICE_ET2UTC, cube[0].et, 'ISOD', 3L, startutc ELSE $
    CSPICE_ET2UTC, mainuval.etstart, 'ISOD', 3L, startutc
maptimewidget = CW_FIELD(mapbase, TITLE = 'Map time', $
                         VALUE = startutc, $
                         UVALUE = 'MAPTIMETEXT', XSIZE = 21, $
                        /RETURN_EVENTS)

;; Allow user to select screen, PS, JPG, or TIFF output and filename:

outputnames = ['Screen', 'PS', 'JPG', 'TIFF']
outputbuttons = CW_BGROUP(ringmapbase, outputnames, UVALUE = 'OUTPUTBUTTON', $
                    /ROW, /EXCLUSIVE, /RETURN_NAME, SET_VALUE = 0, $
                    UNAME = 'outputbutton', LABEL_LEFT = 'Output type')
savefilewidget = CW_FIELD(ringmapbase, TITLE = 'Filename:', UVALUE = $
                        'SAVEFILE', VALUE = 'ringmap', XSIZE = 20)
screen = 1B
postscript = 0B
jpg = 0B
tiff = 0B

IF mainuval.targstate.use THEN BEGIN
   savefile = STRTRIM(mainuval.targstate.savefile,2)
   WIDGET_CONTROL, savefilewidget, SET_VALUE = savefile
ENDIF ELSE savefile = ''

IF NOT MAINUVAL.ringstate.use THEN BEGIN
    mainuval.ringstate.grid = grid
    mainuval.ringstate.gridtop = gridtop
    mainuval.ringstate.shadow = shadow
    mainuval.ringstate.alltimes = alltimes
    mainuval.ringstate.animate = ringanim
    mainuval.ringstate.cubetimes = cubetimes
    mainuval.ringstate.screen = screen
    mainuval.ringstate.postscript = postscript
    mainuval.ringstate.jpg = jpg
    mainuval.ringstate.tiff = tiff
    mainuval.ringstate.savefile = savefile
ENDIF

IF mainuval.ringstate.use THEN BEGIN
    WIDGET_CONTROL, redmaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.ringstate.spec_max[0]),2)
    WIDGET_CONTROL, greenmaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.ringstate.spec_max[1]),2)
    WIDGET_CONTROL, bluemaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.ringstate.spec_max[2]),2)
    WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.ringstate.ntime),2)
    WIDGET_CONTROL, maptimewidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.ringstate.maptime),2)

    IF mainuval.ringstate.animate THEN BEGIN
       WIDGET_CONTROL, animyes, /SET_BUTTON
       ringanim = 1B
    ENDIF
    IF mainuval.ringstate.alltimes THEN BEGIN
       WIDGET_CONTROL, alltimesbutton, /SET_BUTTON
       alltimes = 1B
    ENDIF
    IF mainuval.ringstate.cubetimes THEN BEGIN
       WIDGET_CONTROL, cubeyes, /SET_BUTTON
       cubetimes = 1B
    ENDIF
    IF mainuval.ringstate.screen THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 0
       screen = 1B
    ENDIF
    IF mainuval.ringstate.postscript THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 1
       postscript = 1B
       savefile = mainuval.ringstate.savefile
    ENDIF
    IF mainuval.ringstate.jpg THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 2
       jpg = 1B
       savefile = mainuval.ringstate.savefile
    ENDIF
    IF mainuval.ringstate.tiff THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 3
       tiff = 1B
       savefile = mainuval.ringstate.savefile
    ENDIF


    IF NOT mainuval.ringstate.grid THEN BEGIN
       WIDGET_CONTROL, hidegridbutton, /SET_BUTTON
       grid = 0B
    ENDIF
    IF NOT mainuval.ringstate.gridtop THEN BEGIN
       gridtop = 0B
       WIDGET_CONTROL, gridonbottom, /SET_BUTTON
    ENDIF
    IF NOT mainuval.ringstate.shadow THEN BEGIN
       shadow = 0B
       WIDGET_CONTROL, hideshadowbutton, /SET_BUTTON
    ENDIF

ENDIF


buttonbase = WIDGET_BASE(ringmapbase, /ROW)
helpbutton = WIDGET_BUTTON(buttonbase, VALUE = 'Help', $
                           UVALUE = 'HELPBUTTON')
cancelbutton = WIDGET_BUTTON(buttonbase, VALUE = 'Cancel', $
                             UVALUE = 'CANCELBUTTON')
;; Clicking the done button on this widget results in all 
;; values being read from the widget and passed to
;; DIRECT_RINGMAP.
donebutton = WIDGET_BUTTON(buttonbase, VALUE = 'Make It So', $
                                 UVALUE = 'DONEBUTTON', SENSITIVE=0)

WIDGET_CONTROL, ringmapbase, SET_UVALUE = ringmapuval
WIDGET_CONTROL, ringmapbase, /REALIZE, GROUP = geometer_base, /MODAL

IF mainuval.ringstate.cubetimes THEN $
       WIDGET_CONTROL, donebutton, SENSITIVE = 1

;; Check this widget to see if the Done button is pressed.
;; Once the done button is pressed, read the values from the
;; various inputs in this widget, destroy the widget, and then
;; send these inputs to the ringmap routine.
done_with_ringmapinput = 0B
spec_bounds = [0,1,2,3,4,5]

REPEAT BEGIN
   ringmapevent = WIDGET_EVENT()
   CASE ringmapevent.top OF
      ringmapbase:BEGIN
         WIDGET_CONTROL, ringmapevent.id, GET_UVALUE = ev_uval
         CASE ev_uval OF
            'HELPBUTTON': XDISPLAYFILE, mainuval.geometerpath+$
                'directringmaphelp.txt', DONE_BUTTON = 'Done', $
                                        GROUP = ringmapbase, /MODAL
            'CANCELBUTTON': BEGIN
               WIDGET_CONTROL, ringmapevent.top, /DESTROY
               RETURN
            END
            'DONEBUTTON':BEGIN
               WIDGET_CONTROL, quadwidget, GET_VALUE = quadrant
               quadrant = FLOAT(quadrant[0])
               mainuval.ringstate.quad = quadrant
               WIDGET_CONTROL, radminwidget, GET_VALUE = radmin
               radmin = FLOAT(radmin[0])
               mainuval.ringstate.radmin = radmin
               WIDGET_CONTROL, radmaxwidget, GET_VALUE = radmax
               radmax = FLOAT(radmax[0])
               mainuval.ringstate.radmax = radmax

               ;; Read out the color selections:
               WIDGET_CONTROL, redwidget, GET_VALUE = red
               redlims = STRSPLIT(red,',',/EXTRACT)
               spec_bounds[0] = FIX(redlims[0])
               spec_bounds[1] = FIX(redlims[1])
               WIDGET_CONTROL, redmaxwidget, GET_VALUE = redmaxstring
               redmax = FLOAT(redmaxstring[0])
               
               WIDGET_CONTROL, greenwidget, GET_VALUE = green
               greenlims = STRSPLIT(green,',',/EXTRACT)
               spec_bounds[2] = FIX(greenlims[0])
               spec_bounds[3] = FIX(greenlims[1])
               WIDGET_CONTROL, greenmaxwidget, GET_VALUE = greenmaxstring
               greenmax = FLOAT(greenmaxstring[0])

               WIDGET_CONTROL, bluewidget, GET_VALUE = blue
               bluelims = STRSPLIT(blue,',',/EXTRACT)
               spec_bounds[4] = FIX(bluelims[0])
               spec_bounds[5] = FIX(bluelims[1])
               WIDGET_CONTROL, bluemaxwidget, GET_VALUE = bluemaxstring
               bluemax = FLOAT(bluemaxstring[0])

               mainuval.ringstate.spec_bounds = spec_bounds
               mainuval.ringstate.spec_max = [redmax, greenmax, bluemax]

               WIDGET_CONTROL, ntimewidget, GET_VALUE = ntimestring
               ntimeplot = FIX(FLOAT(ntimestring[0]))
               mainuval.ringstate.ntime = ntimeplot
               
               WIDGET_CONTROL, maptimewidget, GET_VALUE = maptimestring
               maptime = STRTRIM(maptimestring[0],2)
               mainuval.ringstate.maptime = maptime

               WIDGET_CONTROL, savefilewidget, GET_VALUE = savefile
               mainuval.ringstate.savefile = savefile

               done_with_ringmapinput = 1B
               mainuval.ringstate.use = 1B
            END
            'PLOTBUTTON': BEGIN
               WIDGET_CONTROL, plotrowselect, GET_VALUE = rowselect
               rowselect = FIX(FLOAT(rowselect[0])) > 0
                    nrows = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.)) 
                    IF rowselect GE nrows THEN BEGIN
                        WIDGET_CONTROL, plotrowselect, SET_VALUE = $
                                        STRTRIM(STRING(nrows-1),2)
                        rowselect = nrows - 1
                    ENDIF
                    WIDGET_CONTROL, plottimeselect, GET_VALUE = timeselect
                    timeselect = FIX(FLOAT(timeselect[0])) > 0
                    nrecords = N_ELEMENTS(cube.et)
                    IF timeselect GE nrecords THEN BEGIN
                        WIDGET_CONTROL, plottimeselect, SET_VALUE = $
                                        STRTRIM(STRING(nrecords-1), 2)
                        timeselect = nrecords-1
                    ENDIF
                    PLOT, cube[timeselect].data[rowselect, *], XTITLE = $
                          'Channel Number', YTITLE = 'Data Cube Value', $
                          XRANGE = [0, nspec], $
                          TITLE = 'Spectrum for: Record '+$
                          STRTRIM(STRING(timeselect),2)+'. Row '+$
                          STRTRIM(STRING(rowselect),2)+'.', $
                          YRANGE = $
                            [0., MAX(cube[timeselect].data[rowselect,*])]
            END
            'SHADOWYES': BEGIN
               shadow = 1B
               mainuval.ringstate.shadow = 1B
            END
            'SHADOWNO': BEGIN
               shadow = 0B
               mainuval.ringstate.shadow = 0B
            END
            'ALLTIMES': BEGIN
               alltimes = 1B
               mainuval.ringstate.alltimes = 1B
            END
            'ONETIME': BEGIN
               alltimes = 0B
               mainuval.ringstate.alltimes = 0B
            END
            'OUTPUTBUTTON': BEGIN
               WIDGET_CONTROL, outputbuttons, GET_VALUE = outputtype
               CASE outputtype OF
                  0: BEGIN
                     screen = 1B
                     postscript = 0B
                     jpg = 0B
                     tiff = 0B
                     mainuval.ringstate.screen = 1B
                     mainuval.ringstate.postscript = 0B
                     mainuval.ringstate.jpg = 0B
                     mainuval.ringstate.tiff = 0B
                  END
                  1: BEGIN
                     screen = 0B
                     postscript = 1B
                     jpg = 0B
                     tiff = 0B
                     mainuval.ringstate.screen = 0B
                     mainuval.ringstate.postscript = 1B
                     mainuval.ringstate.jpg = 0B
                     mainuval.ringstate.tiff = 0B
                     mainuval.ringstate.savefile = savefile
                  END
                  2: BEGIN
                     screen = 0B
                     jpg = 1B
                     postscript = 0B
                     tiff = 0B
                     mainuval.ringstate.screen = 0B
                     mainuval.ringstate.postscript = 0B
                     mainuval.ringstate.jpg = 1B
                     mainuval.ringstate.tiff = 0B
                     mainuval.ringstate.savefile = savefile
                  END
                  3: BEGIN
                     screen = 0B
                     jpg = 0B
                     postscript = 0B
                     tiff = 1B
                     mainuval.ringstate.screen = 0B
                     mainuval.ringstate.postscript = 0B
                     mainuval.ringstate.jpg = 0B
                     mainuval.ringstate.tiff = 1B
                     mainuval.ringstate.savefile = savefile
                  END
               ENDCASE
            END
            'ANIMYES': BEGIN
               ringanim = 1B
               mainuval.ringstate.animate = 1B
               screen = 1B
               postscript = 0B
               jpg = 0B
               tiff = 0B
               WIDGET_CONTROL, outputbuttons, SET_VALUE = 0
            END
            'ANIMNO': BEGIN
               ringanim = 0B
               mainuval.ringstate.animate = 0B
            END
            'GRIDYES': BEGIN
               grid = 1B
               mainuval.ringstate.grid = 1B
            END
            'GRIDNO': BEGIN
               grid = 0B
               mainuval.ringstate.grid = 0B
            END
            'GRIDTOP': BEGIN 
               gridtop = 1B
               mainuval.ringstate.gridtop = 1B
            END
            'GRIDBOTTOM': BEGIN 
               gridtop = 0B
               mainuval.ringstate.gridtop = 0B
            END
            'MAPTIMETEXT': BEGIN
               WIDGET_CONTROL, maptimewidget, GET_VALUE = $
                               maptimestring
               CSPICE_UTC2ET, maptimestring[0], maptimeet
               CSPICE_ET2UTC, maptimeet, 'ISOD', 3L, maptimestring
               WIDGET_CONTROL, maptimewidget, SET_VALUE = $
                               maptimestring
            END
            'CUBEYES': BEGIN
               cubetimes = 1B
               mainuval.ringstate.cubetimes = 1B
               limbtime = cube.et[0,0]
               ntimes = N_ELEMENTS(cube.et[0,*])
               IF NOT mainuval.ringstate.use THEN $
                  WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                                  STRING(STRTRIM(ntimes),2)
               BODY_LIMB, target, 'CASSINI', limbtime, 90, $
                          ra_limb, dec_limb, limberror
               ;; Allow user to continue
               WIDGET_CONTROL, donebutton, SENSITIVE = 1

            END
            'CUBENO': BEGIN
               IF N_ELEMENTS(data) NE 0 THEN BEGIN
               limbtime = data[0].et
               ntimes = N_ELEMENTS(data.ra_full[0])
               WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                               STRING(STRTRIM(ntimes),2)
               BODY_LIMB, target, 'CASSINI', limbtime, 90, $
                          ra_limb, dec_limb, limberror
               ;; Allow user to continue
               WIDGET_CONTROL, donebutton, SENSITIVE = 1

               cubetimes = 0B
               mainuval.ringstate.cubetimes = 0B
            ENDIF               ; check that data is defined
            END
            ELSE:
         ENDCASE
      END
      ELSE: ;; Do nothing
   ENDCASE
ENDREP UNTIL done_with_ringmapinput
WIDGET_CONTROL, ringmapbase, /DESTROY

IF cubetimes THEN ntimes = N_ELEMENTS(cube.et) ELSE $
   ntimes = N_ELEMENTS(data.ra_full[0])

IF ringanim THEN XINTERANIMATE, SET = [500, 500, ntimeplot] 

;; Identify the time to make the map from the user entry:
IF NOT cubetimes THEN BEGIN
;    WIDGET_CONTROL, mainuval.maptimetext, GET_VALUE = maptime
;    maptime = maptime[0]
    CSPICE_UTC2ET, maptime, mapet
    junk = MIN(ABS(mapet-data.et),etloc)
    CSPICE_ET2UTC, data[etloc].et, mainuval.tformat, 3L, maptime
    mainuval.ringstate.maptime = maptime
    WIDGET_CONTROL, mainuval.statustext, SET_VALUE = $
                    'Actual time for map: ' + maptime
ENDIF ELSE BEGIN
   CSPICE_UTC2ET, maptime, mapet
   junk = MIN(ABS(mapet-cube.et), etloc)
   CSPICE_ET2UTC, cube[etloc].et, mainuval.tformat, 3L, maptime
   mainuval.ringstate.maptime = maptime
ENDELSE

;; Determine which set of times to use and store in vectors
;; to pass to DIRECT_RINGMAP.

dtime_ix = FLOAT(ntimes)/(ntimeplot-1)
plot_ets = DBLARR(ntimeplot)
plot_utcs = STRARR(ntimeplot)

IF cubetimes THEN BEGIN
    IF mainuval.cubeloaded THEN BEGIN
        cube_ringmap = cube
        FOR time_ix = 0, ntimeplot-1 DO BEGIN
            gettime_ix = time_ix*dtime_ix < (ntimes-1)
            plot_ets[time_ix] = cube[gettime_ix].et
            CSPICE_ET2UTC, plot_ets[time_ix], mainuval.tformat, $
                           3L, utc0
            plot_utcs[time_ix] = utc0
        ENDFOR
    ENDIF ELSE BEGIN
        FOR time_ix = 0, ntimeplot-1 DO BEGIN
            gettime_ix = time_ix*dtime_ix < (ntimes-1)
            plot_ets[time_ix] = data[gettime_ix].et
            plot_utcs[time_ix] = data[gettime_ix].utc
        ENDFOR
        WIDGET_CONTROL, mainuval.statustext, SET_VALUE = $
                        'Cube not loaded.'
        WIDGET_CONTROL, ringmapbase, /DESTROY
        RETURN
    ENDELSE
ENDIF ELSE BEGIN
    ;; Create a dummy variable to take the place of the cube
    ;; data to pass to DIRECT_RINGMAP since cube data may
    ;; not be loaded if we're here.
    cube_ringmap = {et:FLTARR(1)+1E20} 
    ;; Give it a ridiculous value so
    ;; DIRECT_RINGMAP will know it's not real.

    FOR time_ix = 0, ntimeplot-1 DO BEGIN
        gettime_ix = time_ix*dtime_ix < (ntimes-1)
        plot_ets[time_ix] = data[gettime_ix].et
        CSPICE_ET2UTC, plot_ets[time_ix], mainuval.tformat, $
                       3L, utc0
        plot_utcs[time_ix] = utc0
    ENDFOR
ENDELSE

;; Now make the call(s) to DIRECT_RINGMAP
IF ringanim THEN BEGIN
   maptimedummy=''
    FOR time_ix = 0, ntimeplot-1 DO BEGIN
        data_ix = time_ix*dtime_ix < (ntimes-1)
        DIRECT_RINGMAP, data, mainuval, data_ix, time_ix, plot_ets, $
                        plot_utcs, cube_ringmap, spec_bounds, maptimedummy, $
                        QUADRANT = quadrant, $
                        RADMIN = radmin, RADMAX = radmax, $
                        COLORMAX = [redmax, greenmax, bluemax], $
                        ALLTIMES = alltimes, GRID = grid, TOPGRID = gridtop, $
                        SHADOW = shadow, $
                        SCREEN = screen, POSTSCRIPT = postscript, $
                        JPG = jpg, TIFF = tiff, SAVEFILE = savefile
        XINTERANIMATE, FRAME = time_ix, WINDOW = !D.WINDOW
    ENDFOR
    XINTERANIMATE
ENDIF ELSE $
  DIRECT_RINGMAP, data, mainuval, etloc, etloc, plot_ets, $
                  plot_utcs, cube_ringmap, spec_bounds, maptime, $
                  QUADRANT = quadrant, $
                  RADMIN = radmin, RADMAX = radmax, $
                  COLORMAX = [redmax, greenmax, bluemax], $
                  ALLTIMES = alltimes, GRID = grid, TOPGRID = gridtop, $
                  SHADOW = shadow, $
                  SCREEN = screen, POSTSCRIPT = postscript, $
                  JPG = jpg, TIFF = tiff, SAVEFILE = savefile
END
