;;File Version: 1.3
;;Last Modified: October 25, 2008
;;By: JEC

PRO DIRECT_BODYMAP, data, maininfo, i_time, i_plot, plot_ets, plot_utcs, $
                    cube, spec_bounds, lonmin, lonmax, latmin, latmax, $
                    maptime, $
                    ALLTIMES = ALLTIMES, COLORMAX = COLORMAX, $
                    ANIMATE = ANIMATE, GRID = GRID, $
                    AZN = AZN, AZS = AZS, STE = STE, STW = STW, $
                    SCREEN = SCREEN, POSTSCRIPT = POSTSCRIPT, $
                    JPG = JPG, TIFF = TIFF, SAVEFILE = SAVEFILE

COMPILE_OPT idl2
;;  >> Abstract <<
;;  Only to be called by GEOMETER.
;;  i_time is the index of the desired time for computing body data.
;;  maininfo is the UVALUE structure from GEOMETER.
;;  Josh Colwell
;;
;; This is a direct graphics version of the iTool version of BODYMAP
;; for GEOMETER. 
;; Rather than create a new GUI here, GUI options will be selected
;; from the main GEOMETER GUI and passed to this routine.
;;
;; Version History:
;; Sept 7, 2006: Created initial version. 
;; Dec. 5, 2006: Don't draw data pixels that are only partially on the body.
;; Jan. 16, 2007: fixed bug in lat/lon plotting option.
;; June 7, 2007: Update for new geometer cube format, and for new
;; maptime input.
;; December 18, 2007: Updated to allow multiple output file formats. 
;; October 25, 2008: Allow latitude and longitude limits in
;; non-cylindrical projections.

;;Identify the time to make the map from the user entry
IF maptime NE '' THEN CSPICE_UTC2ET, maptime, mapet

IF NOT KEYWORD_SET(POSTSCRIPT) THEN BEGIN
   red = 255                    ; Define some colors for later
   orange = 107000                
   yellow = 6553599 
   green = 65280
   cyan = 16711500
   blue = 16711680
   violet = 16600000
   white = 256L*256L*256L-1
   black = 0

   !P.THICK = 1
   !P.CHARTHICK = 1

   SET_PLOT, 'X'

;; Create a window

   IF KEYWORD_SET(GRID) THEN winsize = [800, 400] $ 
   ELSE winsize = [600, 600]

   IF (KEYWORD_SET(JPG) OR KEYWORD_SET(TIFF)) THEN BEGIN
      !P.BACKGROUND = WHITE
      !P.COLOR = BLACK
      IF KEYWORD_SET(GRID) THEN winsize = [1000, 500] $
      ELSE winsize = [1000, 1000]
   ENDIF

   WINDOW, XSIZE = winsize[0], YSIZE = winsize[1]

ENDIF ELSE BEGIN
; Color table written by Liam.Gumley@ssec.wisc.edu
; http://cow.physics.wisc.edu/~craigm/idl/archive/msg00492.html
   r = [0,255,0,255,0,255,0,255,0,255,255,112,219,127,0,255]
   g = [0,0,255,255,255,0,0,255,0,187,127,219,112,127,163,171]
   b = [0,255,255,0,0,0,255,255,115,0,127,147,219,127,255,127]
   TVLCT, r, g, b
   red = 5
   orange = 9
   yellow = 3
   green = 4
   cyan = 2
   blue = 6
   violet = 12
   white = 7
   SET_PLOT, 'PS'
   IF KEYWORD_SET(GRID) THEN BEGIN
      xplot = 40
      yplot = 20
   ENDIF ELSE BEGIN
      xplot = 20
      yplot = 20
   ENDELSE

   SET_PLOT, 'PS'
   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF (extension NE '.ps') THEN savefile = savefile + '.ps'
   DEVICE, FILENAME=savefile, XSIZE=xplot, YSIZE=yplot, /ENCAPSULATED, $
           /COLOR, BITS_PER_PIXEL=8
   !P.THICK = 2
   !P.CHARTHICK = 2
ENDELSE

;; Put the plot time on the annotation layer:
time_label = plot_utcs[i_plot]

;;Get the string value for the target body:
targetid = WIDGET_INFO(maininfo.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, maininfo.targetdrop, GET_VALUE = targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
CSPICE_BODN2C, target, targetid, found
IF NOT found THEN BEGIN
   target = 'SATURN'
   targetid = 699L 
   ;;This should only happen if the target is the rings.
ENDIF

IF cube[0].et GT 1E19 THEN BEGIN
    ;; cube.et is set to 1E20 in GEOMETER if cube data is not to be used.
    npixel = N_ELEMENTS(data[i_time].ra_full[0,*]) 
    selected_pixels = WHERE(maininfo.pxvec)
    name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)
    channel_name = 'UVIS'
    time = data[i_time].et
    IF KEYWORD_SET(ANIMATE) THEN gridtime = data[0].et ELSE gridtime = time
ENDIF ELSE BEGIN
    npixel = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))
    name = 'UVIS'
    channel_name = 'UVIS'
    IF maptime EQ '' THEN BEGIN
       time = plot_ets[i_plot]
    ENDIF ELSE time = mapet

    IF KEYWORD_SET(ANIMATE) THEN gridtime = plot_ets[0] ELSE gridtime = time
ENDELSE

IF name EQ 'Pixel 64' THEN BEGIN
    CASE maininfo.channel OF
        1: channel_name = 'UVIS EUV'
        2: channel_name = 'UVIS EUV'
        3: channel_name = 'UVIS Solar'
        4: channel_name = 'UVIS FUV'
        5: channel_name = 'UVIS FUV'
        6: channel_name = 'UVIS FUV'
        7: channel_name = 'UVIS HSP'
        8: channel_name = 'UVIS HDAC'
        ELSE: channel_name = 'UVIS'
    ENDCASE
ENDIF

; Make a circle symbol for plotting
circ = findgen(17) * (!PI*2/16.)
usersym, COS(circ), SIN(circ)

;;*****************************************************************************
;;                              MAPPING
;;*****************************************************************************
IF KEYWORD_SET(AZN) THEN BEGIN
   MAP_SET, 90, /AZIMUTHAL, /GRID, /ISOTROPIC, /HORIZON, LABEL = 2, $
            LIMIT = [latmin, 0., latmin, 270.], LATLAB = 180., LONLAB = 80.
ENDIF
IF KEYWORD_SET(AZS) THEN BEGIN
   MAP_SET, -90, /AZIMUTHAL, /GRID, /ISOTROPIC, /HORIZON, LABEL = 2, $
            LIMIT = [latmax, 0, latmax, 270.], LATLAB = 180, LONLAB = -80.
ENDIF
IF KEYWORD_SET(STE) THEN BEGIN
   MAP_SET, 0, 0, /STEREOGRAPHIC, /GRID, /ISOTROPIC, /HORIZON, LABEL = 2, $
            LIMIT = [latmin, lonmin, latmax, lonmax]
ENDIF
IF KEYWORD_SET(STW) THEN BEGIN
   MAP_SET, 0, 180, /STEREOGRAPHIC, /GRID, /ISOTROPIC, /HORIZON, LABEL = 2, $
            LIMIT = [latmin, lonmin, latmax, lonmax]
ENDIF

IF KEYWORD_SET(GRID) THEN BEGIN
   lon_inc = 30.
   lat_inc = 30.
   grid_ids = ['']
   PLOT, [lonmin,lonmax], [latmin,latmax], $
          XTITLE = 'West Longitude (Degrees)', YTITLE = 'Latitude (Degrees)', $
          XRANGE = [lonmin, lonmax], YRANGE = [latmin,latmax], $
          LINESTYLE = 6, /XSTY, /YSTY, TITLE = target, /NODATA
   FOR longrid = 360-lon_inc, lon_inc, -lon_inc DO BEGIN
      xcoords = [longrid,longrid]
      ycoords = [-90,90]
      OPLOT, xcoords, ycoords, LINESTYLE = 1
   ENDFOR
   FOR latgrid = -90+lat_inc, 90-lat_inc, lat_inc DO BEGIN
      OPLOT, [360,0], [latgrid,latgrid], LINESTYLE = 1
   ENDFOR
ENDIF

CSPICE_ET2UTC, time, 'ISOD', 3L, time_label
XYOUTS, 20, 20, time_label, SIZE = 1.5, /DEVICE

;;*****************************************************************************
;;                                TERMINATOR
;;*****************************************************************************
;; store the IDs of all markers on the body (limb, terminator, etc.)
marker_ids = ['']

;; Draw the terminator:
lon_term = [0D0]
lat_term = [0D0]

CSPICE_BODN2C, target, targetid, found
CSPICE_BODVAR, targetid, 'RADII', radii
CSPICE_SPKPOS, target, time, 'J2000', 'LT', 'CASSINI', scpos, ltime
CSPICE_SPKPOS, target, time, 'J2000', 'LT', 'SUN', sunpos, ltimesun
CSPICE_EDLIMB, radii[0], radii[1], radii[2], sunpos, term
;; term is a SPICE ellipse defining the terminator.
CSPICE_EL2CGV, term, center, v1, v2

framestring = 'IAU_'+STRUPCASE(STRCOMPRESS(target))
CSPICE_SPKPOS, target, time, framestring, 'LT', 'CASSINI', $
               scpos_p, ltime
CSPICE_PXFORM, 'J2000', framestring, time-ltime, inert2p
inert2p = TRANSPOSE(inert2p) ;;IDL thing.

;; Get 72 points along the terminator of the body:
FOR j = 0, 71 DO BEGIN
    AZ = !DPI*2.0/71D0*j-!DPI
    termvec = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    termvec = inert2p#termvec
    CSPICE_RECLAT, termvec, bodintercept, lon, lat
    lon_term = [lon_term,(2*!DPI-lon)*!RADEG MOD 360D0]
    lat_term = [lat_term, lat*!RADEG]
ENDFOR
lon_term = lon_term[1:*]
lat_term = lat_term[1:*]  ;;eliminate the dummy leading zero.

;; To avoid a crossover line connecting from 0 to 360 longitude, rearrange the
;; terminator points so that lines aren't drawn across the plot. Assume that
;; we have enough points on the terminator so that a jump in terminator 
;; longitude of 180 degrees indicates a crossover.
IF KEYWORD_SET(GRID) THEN BEGIN
j = 0
REPEAT BEGIN
    IF (ABS(lon_term[j+1]-lon_term[j]) GT 180.) THEN BEGIN
;;  Found a discontinuity. Plot up to index=j separately from j+1:*.
        lon_term1 = lon_term[0:j]
        lat_term1 = lat_term[0:j]
        lon_term = lon_term[j+1:*]
        lat_term = lat_term[j+1:*]
        IF (j GT 0) THEN OPLOT, lon_term1, lat_term1, COLOR = red
        j = N_ELEMENTS(lon_term)
    ENDIF
    j = j+1
ENDREP UNTIL j GE N_ELEMENTS(lon_term)-2
ENDIF
OPLOT, lon_term, lat_term, COLOR = red, THICK = 1.5

;;*****************************************************************************
;;                                LIMB
;;*****************************************************************************
;; Do the same thing with the limb:
lon_limb = [0D0]
lat_limb = [0D0]
CSPICE_EDLIMB, radii[0], radii[1], radii[2], scpos_p, limb
;; limb is a SPICE ellipse.
CSPICE_EL2CGV, limb, center, v1, v2

;; Get 72 points along the limb of the body:
FOR j = 0, 71 DO BEGIN
    AZ = !DPI*2.0/71D0*j-!DPI
    limbvec = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    CSPICE_RECLAT, limbvec, bodintercept, lon, lat
    lon_limb = [lon_limb, (2*!DPI-lon)*!RADEG MOD 360D0]
    lat_limb = [lat_limb, lat*!RADEG]
ENDFOR
lon_limb = lon_limb[1:*]
lat_limb = lat_limb[1:*]  ;;eliminate the dummy leading zero.

;; To avoid a crossover line connecting from 0 to 360 longitude, rearrange the
;; limb points so that lines aren't drawn across the plot. 
IF KEYWORD_SET(GRID) THEN BEGIN
j = 0
REPEAT BEGIN
    IF (ABS(lon_limb[j+1]-lon_limb[j]) GT 180.) THEN BEGIN
;;  Found a discontinuity. Plot up to index=j separately from j+1:*.
        lon_limb1 = lon_limb[0:j]
        lat_limb1 = lat_limb[0:j]
        lon_limb = lon_limb[j+1:*]
        lat_limb = lat_limb[j+1:*]
        IF (j GT 0) THEN OPLOT, lon_limb1, lat_limb1, COLOR = blue
        j = N_ELEMENTS(lon_limb)
    ENDIF
    j = j+1
ENDREP UNTIL j GE N_ELEMENTS(lon_limb)-2
ENDIF
OPLOT, lon_limb, lat_limb, COLOR = blue, THICK = 1.5

;;*****************************************************************************
;;                              SUB-SOLAR/SUB-SC
;;*****************************************************************************
;;Plot the sub-solar and sub-spacecraft points:
IF cube[0].et GT 1E19 THEN BEGIN
   subsolarlon = data[i_time].subsolarlon
   subsolarlat = data[i_time].subsolarlat
   subsclon = data[i_time].subsclon
   subsclat = data[i_time].subsclat
 ENDIF ELSE BEGIN
    junk = MIN(ABS(cube.et-time), timeloc)
   subsolarlon = cube[timeloc].sub_solar_longitude
   subsolarlat = cube[timeloc].sub_solar_latitude
   subsclon = cube[timeloc].sub_spacecraft_longitude
   subsclat = cube[timeloc].sub_spacecraft_latitude
ENDELSE

OPLOT, [subsolarlon], [subsolarlat], $
          LINESTYLE = 6, PSYM = 4, COLOR = red, THICK = 1.5
XYOUTS, [subsolarlon], [subsolarlat], 'Sub-Solar'

OPLOT, [subsclon], [subsclat], $
          LINESTYLE = 6, PSYM = 5, COLOR = blue, THICK = 1.5
XYOUTS, [subsclon], [subsclat], 'Sub-S/C'

;; Label sub-solar and sub-s/c points
subsolar_out = STRTRIM(STRING(subsolarlon, '(D8.2)'), 2) + ', ' + $
               STRTRIM(STRING(subsolarlat, '(D8.2)'), 2)
XYOUTS, [subsolarlon], [subsolarlat] - 5, subsolar_out

scpos_out = STRTRIM(STRING(subsclon, '(D8.2)'), 2) + ', ' + $
            STRTRIM(STRING(subsclat, '(D8.2)'), 2)
XYOUTS, [subsclon], [subsclat] - 5, scpos_out


;;**************************************************************************
;;                                PIXELS
;;**************************************************************************

IF cube[0].et GT 1E19 THEN BEGIN
   npixel = N_ELEMENTS(data[i_time].ra_full[0,*])

;; Indicate to the user which end of the slit is which by drawing the lowest 
;; numbered pixel filled in green.

   selected_pixels = WHERE(maininfo.pxvec)
   name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)

;; If the pixel is not entirely on the body, then use symbols and no line 
;; to draw the pixel boundary. Otherwise draw solid lines. 
   IF MIN(data[i_time].lon_full[1:*,0]) EQ -1000. AND $
      MAX(data[i_time].lon_full[1:*,0]) NE -1000. THEN BEGIN
      OPLOT, data[i_time].lon_full[1:*,0], data[i_time].lat_full[1:*,0], $
             LINESTYLE = 6 
   ENDIF ELSE BEGIN
      OPLOT, [data[i_time].lon_full[1:*,0],data[i_time].lon_full[1,0]], $
             [data[i_time].lat_full[1:*,0],data[i_time].lat_full[1,0]]
   ENDELSE

   ;; Using data from the GEOMETER data structure in the current session.
   IF KEYWORD_SET(ALLTIMES) THEN BEGIN
      ntimes = N_ELEMENTS(plot_ets)
      FOR time_ix = 0, ntimes-1 DO BEGIN
         OPLOT, [data[time_ix].lon_full[1:*,0],data[time_ix].lon_full[1,0]], $
                [data[time_ix].lat_full[1:*,0],data[time_ix].lat_full[1,0]]
         
         IF npixel GT 1 THEN BEGIN
            FOR px_index = 1, npixel-1 DO $
               OPLOT, [data[time_ix].lon_full[1:*,px_index], $
                       data[time_ix].lon_full[1,px_index]], $
                      [data[time_ix].lat_full[1:*,px_index], $
                       data[time_ix].lat_full[1,px_index]]
                
         ENDIF
      ENDFOR  ;; Looping over times.
;; Just label the last time:
      XYOUTS, data[ntimes-1].lon_full[1,0],data[ntimes-1].lat_full[1,0], $
              channel_name
   ENDIF ELSE BEGIN
      OPLOT, [data[i_time].lon_full[1:*,0],data[i_time].lon_full[1,0]], $
             [data[i_time].lat_full[1:*,0],data[i_time].lat_full[1,0]]
        
      XYOUTS, data[i_time].lon_full[1,0],data[i_time].lat_full[1,0], $
              channel_name
        
      IF npixel GT 1 THEN BEGIN
         FOR px_index = 1, npixel-1 DO BEGIN
            OPLOT, [data[i_time].lon_full[1:*,px_index], $
                    data[i_time].lon_full[1,px_index]], $
                   [data[i_time].lat_full[1:*,px_index], $
                    data[i_time].lat_full[1,px_index]]
            
         ENDFOR
         name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)
         XYOUTS, data[i_time].lon_full[1,0], $
                 data[i_time].lat_full[1,0], name
      ENDIF
   ENDELSE  ;;Only plotting at one selected time.

ENDIF ELSE BEGIN
   npixel = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))

;; Plotting data from loaded image cube (also computed by GEOMETER, but
;; run and saved at a different time).

   nrows = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))
   ntimes = N_ELEMENTS(plot_ets)

;; Figure out colors. Scale each channel to 255.
    IF colormax[0] EQ 0. THEN BEGIN
        redmax = 1.E-20
        FOR time_ix = 0, ntimes-1 DO $
          FOR px_index = 0, npixel-1 DO $
            redmax = redmax > $
          MAX(TOTAL(cube[time_ix].data[px_index,$
                              spec_bounds[0]:spec_bounds[1]]))
    ENDIF ELSE redmax = colormax[0]
    IF colormax[1] EQ 0. THEN BEGIN
        greenmax = 1.E-20
        FOR time_ix = 0, ntimes-1 DO $
          FOR px_index = 0, npixel-1 DO $
            greenmax = greenmax > $
          MAX(TOTAL(cube[time_ix].data[px_index,$
                              spec_bounds[2]:spec_bounds[3]]))
    ENDIF ELSE greenmax = colormax[1]
    IF colormax[2] EQ 0. THEN BEGIN
        bluemax = 1.E-20
        FOR time_ix = 0, ntimes-1 DO $
          FOR px_index = 0, npixel-1 DO $
            bluemax = bluemax > $
          MAX(TOTAL(cube[time_ix].data[px_index,$
                              spec_bounds[4]:spec_bounds[5]]))
    ENDIF ELSE bluemax = colormax[2]

   IF KEYWORD_SET(ALLTIMES) THEN BEGIN
      FOR time_ix = 0, ntimes-1 DO BEGIN

         FOR px_index = 0, npixel-1 DO BEGIN
            ;; Build vectors for plotting the boundary of the slit:
            lonvec = [cube[time_ix].lon_full[*,px_index], $
                     cube[time_ix].lon_full[0,px_index]]
            latvec = $
               [cube[time_ix].lat_full[*,px_index], $
                cube[time_ix].lat_full[0,px_index]]
            
            ;; Make sure the pixel is entirely on the body. Otherwise
            ;; we won't plot it.
            IF MIN(latvec) GT -1000. THEN BEGIN

                redcolor = TOTAL(cube[time_ix].data[px_index,$
                                           spec_bounds[0]:spec_bounds[1]]) $
                           < redmax
                redcolor /= redmax
                redcolor *= 255.

                greencolor = TOTAL(cube[time_ix].data[px_index,$
                                             spec_bounds[2]:spec_bounds[3]]) $
                             < greenmax
                greencolor /= greenmax
                greencolor *= 255.

                bluecolor = TOTAL(cube[time_ix].data[px_index,$
                                            spec_bounds[4]:spec_bounds[5]]) $
                            < bluemax
                bluecolor /= bluemax
                bluecolor *= 255.

                ;; Now combine these three shades into a single color:
;; problem: PS doesn't accept TrueColor plots. Only images. 
                TVLCT, INDGEN(256), INDGEN(256), INDGEN(256)

                ;; Now combine these three shades into a single color:
                pxcolor = LONG(redcolor)+256L*LONG(greencolor)+$
                          256L*256L*LONG(bluecolor)

                OPLOT, lonvec, latvec, COLOR = pxcolor
                POLYFILL, lonvec, latvec, COLOR = pxcolor
             ENDIF   ;;pixel is entirely on body
         ENDFOR      ;; Building pixel vectors.
      ENDFOR         ;; Looping over times.
;; Just label the last time:

;; This line was crashing the program. - AWD 9/28/07
;      XYOUTS, lonvec[time_ix-1], latvec[time_ix-1], channel_name
        
   ENDIF ELSE BEGIN
;; Just plotting at one selected time.
      FOR px_index = 0, npixel-1 DO BEGIN
         lonvec = [cube[i_time].lon_full[*,px_index], $
                   cube[i_time].lon_full[0,px_index]]
         
         latvec = $
            [cube[i_time].lat_full[*,px_index], $
             cube[i_time].lat_full[0,px_index]]
            
         IF MIN(latvec GT -1000.) THEN BEGIN
            ;;Get the color of the pixel:
            redcolor = TOTAL(cube[i_time].data[px_index,$
                                       spec_bounds[0]:spec_bounds[1]]) $
                       < redmax
            redcolor /= redmax
            redcolor *= 255.
            
            greencolor = TOTAL(cube[i_time].data[px_index,$
                                         spec_bounds[2]:spec_bounds[3]]) $
                         < greenmax
            greencolor /= greenmax
            greencolor *= 255.
            
            bluecolor = TOTAL(cube[i_time].data[px_index,$
                                        spec_bounds[4]:spec_bounds[5]]) $
                        < bluemax
            bluecolor /= bluemax
            bluecolor *= 255.
            
            ;; Now combine these three shades into a single color:
            pxcolor = LONG(redcolor)+256L*LONG(greencolor)+$
                      256L*256L*LONG(bluecolor)
            
            IF ((MIN(lonvec) NE -1000) AND (MIN(latvec) NE -1000)) THEN BEGIN
               OPLOT, lonvec, latvec, COLOR = pxcolor
               POLYFILL, lonvec, latvec, COLOR = pxcolor
            ENDIF
         ENDIF
      ENDFOR
      name = 'Pixel '+STRTRIM(STRING(px_index-1),2)
      XYOUTS, cube[i_time].lon_full[0, px_index-1], $
              cube[i_time].lat_full[0, px_index-1], name
   ENDELSE  ;;Only plotting at one selected time.
    
ENDELSE  ;;Plotting data cube pixels.


;; Write JPEG or TIFF files:

IF KEYWORD_SET(JPG) THEN BEGIN
   win = TVRD(TRUE = 1)
   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF ((extension NE '.jpg') AND (extension NE 'jpeg')) THEN $
      savefile = savefile + '.jpg'
   WRITE_JPEG, savefile, win, QUALITY = 100, TRUE = 1
ENDIF

IF KEYWORD_SET(TIFF) THEN BEGIN
   win = TVRD(TRUE = 1)
   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF ((extension NE '.tif') AND (extension NE 'tiff')) THEN $
      savefile = savefile + '.tif'
   WRITE_TIFF, savefile, win, ORIENTATION = 0
ENDIF

IF KEYWORD_SET(POSTSCRIPT) THEN DEVICE, /CLOSE

END
