;;File Version: 1.13
;;Last Modified: December 18, 2007
;;By: AWD
FORWARD_FUNCTION HA2LON

PRO DIRECT_RINGMAP, data, maininfo, i_time, i_plot, plot_ets, plot_utcs, $
                    cube, spec_bounds, maptime, ALLTIMES = ALLTIMES, $
                    QUADRANT = quadrant, $
                    RADMIN = radmin, RADMAX = radmax, $
                    COLORMAX = COLORMAX, ANIMATE = ANIMATE, $
                    GRID = grid, TOPGRID = TOPGRID, SHADOW = SHADOW, $
                    SCREEN = SCREEN, POSTSCRIPT = POSTSCRIPT, $
                    JPG = JPG, TIFF = TIFF, SAVEFILE = SAVEFILE

;;  >> Abstract <<
;;  Only to be called by GEOMETER.

;;  i_time is the index of the desired time for computing body data.
;;  maininfo is the UVALUE structure from GEOMETER.
;;  Josh Colwell
;;
;; This is a direct graphics version of the iTool version of BODYMAP
;; for GEOMETER. 
;; Rather than create a new GUI here, GUI options will be selected
;; from the main GEOMETER GUI and passed to this routine.
;;
;; Version History:
;; Sept 7, 2006: Created initial version. 
;; December 6, 2006: Implement pixel plotting and data plotting. Allow
;; grid to be drawn over pixels (gridtop keyword).
;; June 16, 2007: Implement maptime.
;; December 18, 2007: Updated to allow multiple output file formats. 

;;Identify the time to make the map from the user entry
IF maptime NE '' THEN CSPICE_UTC2ET, maptime, mapet

;;Get the string value for the target body:
targetid = WIDGET_INFO(maininfo.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, maininfo.targetdrop, GET_VALUE = targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
CSPICE_BODN2C, target, targetid, found
IF NOT found THEN BEGIN
   target = 'SATURN'
   targetid = 699L 
   ;;This should only happen if the target is the rings.
ENDIF

IF cube[0].et GT 1E19 THEN BEGIN
    ;; cube.et is set to 1E20 in GEOMETER if cube data is not to be used.
    npixel = N_ELEMENTS(data[i_time].ra_full[0,*]) 
    selected_pixels = WHERE(maininfo.pxvec)
    name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)
    channel_name = 'UVIS'
    IF maptime EQ '' THEN $
       time = data[i_time].et $
    ELSE time = mapet

    IF KEYWORD_SET(ANIMATE) THEN gridtime = data[0].et ELSE gridtime = time
ENDIF ELSE BEGIN
    npixel = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))
    name = 'UVIS'
    channel_name = 'UVIS'
    IF maptime EQ '' THEN $
       time = plot_ets[i_plot] $
    ELSE time = mapet
    IF KEYWORD_SET(ANIMATE) THEN gridtime = plot_ets[0] ELSE gridtime = time
ENDELSE

;; Indicate to the user which end of the slit is which by drawing the lowest 
;; numbered pixel filled in green:


IF name EQ 'Pixel 64' THEN BEGIN
    CASE maininfo.channel OF
        1: channel_name = 'UVIS EUV'
        2: channel_name = 'UVIS EUV'
        3: channel_name = 'UVIS Solar'
        4: channel_name = 'UVIS FUV'
        5: channel_name = 'UVIS FUV'
        6: channel_name = 'UVIS FUV'
        7: channel_name = 'UVIS HSP'
        8: channel_name = 'UVIS HDAC'
        ELSE: channel_name = 'UVIS'
    ENDCASE
ENDIF

IF NOT KEYWORD_SET(POSTSCRIPT) THEN BEGIN
   red = 255                    ; Define some colors for later
   orange = 107000                
   yellow = 6553599 
   green = 65280
   cyan = 16711500
   blue = 16711680
   violet = 16600000
   white = 256L*256L*256L-1
   black = 0

   !P.THICK = 1
   !P.CHARTHICK = 1

;; Create the window:

   SET_PLOT, 'X'

   IF KEYWORD_SET(SCREEN) THEN winsize = 500

   IF (KEYWORD_SET(JPG) OR KEYWORD_SET(TIFF)) THEN BEGIN
      !P.BACKGROUND = WHITE
      !P.COLOR = BLACK
      winsize = 1000
   ENDIF

   WINDOW, 0, XSIZE = winsize, YSIZE = winsize

ENDIF ELSE BEGIN
; Color table written by Liam.Gumley@ssec.wisc.edu
; http://cow.physics.wisc.edu/~craigm/idl/archive/msg00492.html
   r = [0,255,0,255,0,255,0,255,0,255,255,112,219,127,0,255]
   g = [0,0,255,255,255,0,0,255,0,187,127,219,112,127,163,171]
   b = [0,255,255,0,0,0,255,255,115,0,127,147,219,127,255,127]
   TVLCT, r, g, b

   red = 5
   orange = 9
   yellow = 3
   green = 4
   cyan = 2
   blue = 6
   violet = 12
   white = 7
   SET_PLOT, 'PS'

   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF (extension NE '.ps') THEN savefile = savefile + '.ps'

   DEVICE, FILENAME=savefile, XSIZE=20, YSIZE=20, /ENCAPSULATED, $
           /COLOR, BITS_PER_PIXEL=8
   !P.THICK = 2
   !P.CHARTHICK = 2
ENDELSE

; Make a circle symbol for plotting
circ = findgen(17) * (!PI*2/16.)
usersym, COS(circ), SIN(circ)

;;Get the string value for the target body:
targetid = WIDGET_INFO(maininfo.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, maininfo.targetdrop, GET_VALUE = targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
CSPICE_BODN2C, target, targetid, found
IF NOT found THEN BEGIN
    target = 'SATURN'
    targetid = 699L 
    ;;This should only happen if the target is the rings.
ENDIF

;;*************************************************************************
;;                              RINGS
;;*************************************************************************
ringradlimit = 1.5E5

;; Plot the critical ring boundaries.
;; Would be better eventually to read these values in from PCK.
CSPICE_BODVAR, 699L, 'RADII', satrad

ringrad = [satrad[0], 66970.,74510.,92000.,117580.,122170.,133405.,$
           133730.,136780.,140180.,140270.]
nring = N_ELEMENTS(ringrad)    

nringpoints = 72
;;Use 72 points for each ring

ringradvec = DBLARR(nringpoints)
ringlon = FINDGEN(nringpoints)/(nringpoints-1D0)*!DPI*2D0

CASE quadrant OF
    1: radrange = [0, radmax, 0, radmax]
    2: radrange = [-radmax, 0, 0, radmax]
    3: radrange = [-radmax, 0, -radmax, 0]
    4: radrange = [0, radmax, -radmax, 0]
ELSE: RADRANGE = [-RADMAX, RADMAX, -RADMAX, RADMAX]
ENDCASE

ringradvec[*] = ringrad[0]
ringindex = 0
PLOT, ringradvec, ringlon, /POLAR, $
      XR = RADRANGE[0:1], YR = RADRANGE[2:3], /ISO

FOR ringindex = 1, nring-1 DO BEGIN
    ringradvec[*] = ringrad[ringindex]
    OPLOT, ringradvec, ringlon, /POLAR
ENDFOR  ;;ringindex

;;*************************************************************************
;;                             PLANET SHADOW
;;*************************************************************************
IF KEYWORD_SET(SHADOW) THEN BEGIN

   ;; Plot the sub-solar point:
   IF cube[0].et GT 1E19 THEN BEGIN
      subsolarlon = data[i_time].subsolarlon
   ENDIF ELSE BEGIN
      junk = MIN(ABS(cube.et-time),cubetime)
      subsolarlon = cube[cubetime].sub_solar_longitude
   ENDELSE
   OPLOT, [60000], !DTOR*[subsolarlon], PSYM=2, COLOR=orange, THICK=1.5, /POLAR

;; Compute the location of the boundary of the shadow of Saturn on the
;; ring plane:
   ringin = [66970., 74500., 87350., 92000., 122000., 133750., 140160.]
   ringout = [74500., 87600., 92000., 117500., 133425., 136775., 140200.]
   nrad = 100     ;; Do for 100 evenly spaced points in ring plane radius.
   radvec = FINDGEN(100)/99.*(MAX(ringout)-MIN(ringin))+MIN(ringin)

;; At each radius, find the shadow boundaries.  Start at hour angle of
;; 12 and increase.
;; This means the first boundary crossing will be the dusk boundary, and
;; the second will be the dawn boundary.  Ignore the change in the
;; boundary with time due to the Sun's changing elevation and Saturn's
;; oblateness.  Pick a time in the middle of the tour and assume the
;; hour angle of the boundaries don't change significantly with time.

;; So, first get Sun location:
   abcorr = 'LT+S'
   sunloc = DBLARR(3)
   sunet = 0D0
   CSPICE_SPKPOS, 'SUN', time, 'IAU_SATURN', abcorr, 'SATURN', sunloc, ltcorr
   polarsun = CV_COORD(FROM_RECT = sunloc[0:1], /TO_POLAR)
   sunlon = polarsun[0]*!RADEG

;; Also need size of Saturn:
   axes = DBLARR(3)
   CSPICE_BODVAR, 699L, 'RADII', axes

   duskha = DBLARR(nrad)
   dawnha = DBLARR(nrad)
   point = DBLARR(3)
   found = 0L

   FOR boundary = 1,-1,-2 DO BEGIN
;; Do once for dusk boundary (boundary = 1) and once for dawn boundary
;; (boundary = -1).
      FOR rindex = 0, nrad-1 DO BEGIN
         radius = radvec[rindex]
         guessha = 12D0  ;; start at noon.
         guesslon = HA2LON(guessha, sunlon)
         lonstep = 10D0  ;; initial step size in longitude in degrees.
         shadow = 0B
         direction = boundary
         REPEAT BEGIN
            oldshadow = shadow
            guesslon = guesslon+lonstep*direction
            shadow = 0B
            cartpoint = CV_COORD(FROM_POLAR = [guesslon*!DTOR, radius], /TO_RECT)
            ;; cartpoint is a point on the rings in the Saturn frame.
            lookvec = [cartpoint,0D0]-sunloc
            ;; lookvec is the 3D vector from Sun to the point being tested
            ;; on the rings.
            CSPICE_SURFPT, sunloc, lookvec, axes[0], axes[1], axes[2], point, $
                           found
            ;; Test to see if intersection point is between test point and
            ;; Sun:
            IF found THEN $
               shadow = (CSPICE_VDIST(sunloc, point) LT CSPICE_VNORM(lookvec)) $
            ELSE shadow = 0B
            IF shadow NE oldshadow THEN BEGIN
               lonstep = lonstep/2D0
               direction = -direction
            ENDIF
            
         ENDREP UNTIL lonstep LT 5D-1
         
         IF boundary EQ 1 THEN $
            duskha[rindex] = 12D0-(guesslon-sunlon)/15D0 $
         ELSE $
            dawnha[rindex] = 12D0-(guesslon-sunlon)/15D0
      ENDFOR
   ENDFOR
   
   smalldusk = WHERE(duskha LT -12.)
   IF smalldusk[0] NE -1 THEN duskha[smalldusk] = duskha[smalldusk]+24.
   bigdawn = WHERE(dawnha GT 12.)
   IF bigdawn[0] NE -1 THEN dawnha[bigdawn] = dawnha[bigdawn]-24.
   
   FOR i = 0, N_ELEMENTS(dawnha) - 1 DO BEGIN
      dawnha[i] = HA2LON(dawnha[i], sunlon)
      duskha[i] = HA2LON(duskha[i], sunlon)
   ENDFOR

   dawnha = (!DTOR*dawnha)
   duskha = (!DTOR*duskha)

   OPLOT, radvec, dawnha, COLOR = blue, /POLAR
   OPLOT, radvec, duskha, COLOR = red, /POLAR

   save, file='ring.sav', radvec, dawnha, duskha

ENDIF
;;*************************************************************************
;;                                    GRID
;;*************************************************************************
;; Now draw some radial grid lines at constant longitude.
lon_inc = 30.

IF KEYWORD_SET(GRID) THEN BEGIN
   FOR longrid = 0, 360.-lon_inc, lon_inc DO BEGIN
      radcoords = [MIN(ringrad), MAX(ringrad)]
      loncoords = [longrid, longrid]*!DTOR
      OPLOT, radcoords, loncoords, /POLAR
   ENDFOR
ENDIF

;;**************************************************************************
;;                              Plot pixels.
;;**************************************************************************
IF cube[0].et GT 1E19 THEN BEGIN
;; Indicate to the user which end of the slit is which by drawing the lowest 
;; numbered pixel filled in green.

   selected_pixels = WHERE(maininfo.pxvec)
   name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)

   radvec = [data[i_time].rrad_full[1:*,0], data[i_time].rrad_full[1,0]]
   lonvec = [data[i_time].lon_full[1:*,0], data[i_time].lon_full[1,0]]
   lonvec *= !DTOR
   OPLOT, radvec, lonvec, /POLAR

    ;; Using data from the GEOMETER data structure in the current session.
   IF KEYWORD_SET(ALLTIMES) THEN BEGIN
      ntimes = N_ELEMENTS(plot_ets)
      FOR time_ix = 0, ntimes-1 DO BEGIN
          radvec =[data[time_ix].rrad_full[1:*,0],data[time_ix].rrad_full[1,0]]
          lonvec=[data[time_ix].lon_full[1:*,0],data[time_ix].lon_full[1,0]]
          lonvec *= !DTOR
          OPLOT, radvec, lonvec, /POLAR
         
          IF npixel GT 1 THEN BEGIN
              FOR px_index = 1, npixel-1 DO BEGIN
                radvec = [data[time_ix].rrad_full[1:*,px_index], $
                          data[time_ix].rrad_full[1,px_index]]
                lonvec = [data[time_ix].lon_full[1:*,px_index], $
                          data[time_ix].lon_full[1,px_index]]
                lonvec *= !DTOR
                OPLOT, radvec, lonvec, /POLAR
            ENDFOR
         ENDIF
      ENDFOR  ;; Looping over times.
;; Just label the last time:
      XYOUTS, data[ntimes-1].rrad_full[1,0],data[ntimes-1].lon_full[1,0]*$
              !DTOR, channel_name
  ENDIF ELSE BEGIN
      OPLOT, [data[i_time].rrad_full[1:*,0],data[i_time].rrad_full[1,0]], $
             [data[i_time].lon_full[1:*,0],data[i_time].lon_full[1,0]]*!DTOR,$
             /POLAR
        
      XYOUTS, data[i_time].rrad_full[1,0],data[i_time].lon_full[1,0]*!DTOR, $
              channel_name
      
      IF npixel GT 1 THEN BEGIN
         FOR px_index = 1, npixel-1 DO BEGIN
            OPLOT, [data[i_time].rrad_full[1:*,px_index], $
                    data[i_time].rrad_full[1,px_index]], $
                   [data[i_time].lon_full[1:*,px_index], $
                    data[i_time].lon_full[1,px_index]]*!DTOR, /POLAR
            
         ENDFOR
         name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)
         XYOUTS, data[i_time].rrad_full[1,0], $
                 data[i_time].lon_full[1,0]*!DTOR, name
      ENDIF
   ENDELSE  ;;Only plotting at one selected time.

ENDIF ELSE BEGIN

;; Plotting data from loaded image cube (also computed by GEOMETER, but
;; run and saved at a different time).

   nrows = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))
   ntimes = N_ELEMENTS(plot_ets)

;; Figure out colors. Scale each channel to 255.
   IF colormax[0] EQ 0. THEN BEGIN
      redmax = 1.E-20
      FOR time_ix = 0, ntimes-1 DO $
         FOR px_index = 0, npixel-1 DO $
            redmax = redmax > $
         MAX(TOTAL(cube[time_ix].data[px_index,$
                                      spec_bounds[0]:spec_bounds[1]]))
   ENDIF ELSE redmax = colormax[0]
   IF colormax[1] EQ 0. THEN BEGIN
      greenmax = 1.E-20
      FOR time_ix = 0, ntimes-1 DO $
         FOR px_index = 0, npixel-1 DO $
            greenmax = greenmax > $
         MAX(TOTAL(cube[time_ix].data[px_index,$
                             spec_bounds[2]:spec_bounds[3]]))
   ENDIF ELSE greenmax = colormax[1]
   IF colormax[2] EQ 0. THEN BEGIN
      bluemax = 1.E-20
      FOR time_ix = 0, ntimes-1 DO $
         FOR px_index = 0, npixel-1 DO $
            bluemax = bluemax > $
         MAX(TOTAL(cube[time_ix].data[px_index,$
                             spec_bounds[4]:spec_bounds[5]]))
   ENDIF ELSE bluemax = colormax[2]
   
   IF KEYWORD_SET(ALLTIMES) THEN BEGIN
      ntimes = N_ELEMENTS(plot_ets)
      FOR time_ix = 0, ntimes-1 DO BEGIN
          
         FOR px_index = 0, npixel-1 DO BEGIN
            ;; Build vectors for plotting the boundary of the slit:
            lonvec = [cube[time_ix].lon_full[*,px_index], $
                      cube[time_ix].lon_full[0,px_index]]
            lonvec *= !DTOR

            radvec = $
               [cube[time_ix].rrad_full[*,px_index], $
                cube[time_ix].rrad_full[0,px_index]]

            ;;Get the brightness of the pixel:
            redcolor = TOTAL(cube[time_ix].data[px_index,$
                                       spec_bounds[0]:spec_bounds[1]]) $
                       < redmax
            redcolor /= redmax
            redcolor *= 255.

            greencolor = TOTAL(cube[time_ix].data[px_index,$
                                         spec_bounds[2]:spec_bounds[3]]) $
                         < greenmax
            greencolor /= greenmax
            greencolor *= 255.

            bluecolor = TOTAL(cube[time_ix].data[px_index,$
                                        spec_bounds[4]:spec_bounds[5]]) $
                        < bluemax
            bluecolor /= bluemax
            bluecolor *= 255.

                ;; Now combine these three shades into a single color:
            pxcolor = LONG(redcolor)+256L*LONG(greencolor)+$
                      256L*256L*LONG(bluecolor)

            OPLOT, radvec, lonvec, COLOR = pxcolor, /POLAR
            POLYFILL, radvec, lonvec, COLOR = pxcolor

         ENDFOR  ;; Building pixel vectors.
      ENDFOR     ;; Looping over times.
;; Just label the last time:
      XYOUTS, cube[time_ix-1].rrad_full[0,0],$
              cube[time_ix-1].lon_full[0,0]*!DTOR, channel_name
        
   ENDIF ELSE BEGIN
;; Just plotting at one selected time.
      FOR px_index = 0, npixel-1 DO BEGIN
         lonvec = [cube[time_ix].lon_full[*,px_index], $
                   cube[time_ix].lon_full[0,px_index]]
         lonvec *= !DTOR

         radvec = $
            [cube[time_ix].rrad_full[*,px_index], $
             cube[time_ix].rrad_full[0,px_index]]
         
         ;;Get the color of the pixel:
         redcolor = TOTAL(cube[i_time].data[px_index,$
                                    spec_bounds[0]:spec_bounds[1]])
         redcolor /= redmax
         redcolor *= 255.
            
         greencolor = TOTAL(cube[i_time].data[px_index,$
                                      spec_bounds[2]:spec_bounds[3]])
         greencolor /= greenmax
         greencolor *= 255.
            
         bluecolor = TOTAL(cube[i_time].data[px_index,$
                                     spec_bounds[4]:spec_bounds[5]])
         bluecolor /= bluemax
         bluecolor *= 255.
            
         ;; Now combine these three shades into a single color:
         pxcolor = LONG(redcolor)+256L*LONG(greencolor)+$
                   256L*256L*LONG(bluecolor)

         OPLOT, radvec, lonvec, COLOR = pxcolor, /POLAR
         POLYFILL, radvec, lonvec, COLOR = pxcolor
      ENDFOR
      name = 'Pixel '+STRTRIM(STRING(px_index-1),2)
      XYOUTS, cube[i_time-1].rrad_full[0,px_index-1], $
              cube[i_time-1].lon_full[0,px_index-1]*!DTOR, name
   ENDELSE  ;;Only plotting at one selected time.
    
ENDELSE  ;;Plotting data cube pixels.

IF KEYWORD_SET(TOPGRID) THEN BEGIN
;;*************************************************************************
;; Draw grid again, now that the pixel data is on the plot.
;;*************************************************************************

    ringindex = 0
    ringradvec[*] = ringrad[0]
    OPLOT, ringradvec, ringlon, /POLAR
    
    FOR ringindex = 1, nring-1 DO BEGIN
        ringradvec[*] = ringrad[ringindex]
        OPLOT, ringradvec, ringlon, /POLAR
    ENDFOR  ;;ringindex
    
    IF KEYWORD_SET(GRID) THEN BEGIN
        FOR longrid = 0, 360.-lon_inc, lon_inc DO BEGIN
            radcoords = [MIN(ringrad), MAX(ringrad)]
            loncoords = [longrid, longrid]*!DTOR
            OPLOT, radcoords, loncoords, /POLAR
        ENDFOR
    ENDIF
    
ENDIF


CSPICE_ET2UTC, time, 'ISOD', 3L, time_label
XYOUTS, 10, 10, time_label, SIZE = 1.5, /DEVICE

;; Write JPEG or TIFF files:

IF KEYWORD_SET(JPG) THEN BEGIN
   win = TVRD(TRUE = 1)
   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF ((extension NE '.jpg') AND (extension NE 'jpeg')) THEN $
      savefile = savefile + '.jpg'
   WRITE_JPEG, savefile, win, QUALITY = 100, TRUE = 1
ENDIF

IF KEYWORD_SET(TIFF) THEN BEGIN
   win = TVRD(TRUE = 1)
   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF ((extension NE '.tif') AND (extension NE 'tiff')) THEN $
      savefile = savefile + '.tif'
   WRITE_TIFF, savefile, win, ORIENTATION = 0
ENDIF


IF KEYWORD_SET(POSTSCRIPT) THEN DEVICE, /CLOSE

END


FUNCTION HA2LON, hourangle, sunlon
;; convert hourangle to longitude.
;; sunlon is input in degrees.
;; By: Joshua Colwell.
;; April 3, 2001.

result = sunlon+(12D0-hourangle)*15D0
IF result LT 0D0 THEN result = result+360D0
result = result MOD 360D0
RETURN, result
END
