;;File Version: 1.4
;;Last Modified: June 2, 2009
;;By: JEC

PRO DIRECT_SKYMAP, data, maininfo, i_time, i_plot, plot_ets, plot_utcs, $
                   cube, spec_bounds, maptime, STARS = STARS, $
                   ALLTIMES = ALLTIMES, $
                   RA_RANGE = RA_RANGE, DEC_RANGE = DEC_RANGE, $
                   COLORMAX = COLORMAX, LATLON = LATLON, $
                   SCREEN = SCREEN, POSTSCRIPT = POSTSCRIPT, $
                   JPG = JPG, TIFF = TIFF, SAVEFILE = SAVEFILE, $
                   XSIZE = XSIZE, YSIZE = YSIZE

;;  >> Abstract <<
;;  Only to be called by GEOMETER.

;;  i_time is the index of the desired time for computing body data.
;;  maininfo is the UVALUE structure from GEOMETER.
;;  Josh Colwell
;;
;; This is a direct graphics version of the iTool version of SKYMAP
;; for GEOMETER. This uses code from the OCCMAP program as well.
;; Rather than create a new GUI here, GUI options will be selected
;; from the main GEOMETER GUI and passed to this routine.
;;
;; Version History:
;; July 27, 2006: Created initial version with basic drawing at a single
;; time. 
;; July 28, 2006: Add ALLTIMES, RA_RANGE, DEC_RANGE
;; July 30, 2006: Animation and ALLTIMES plotting working with cube data.
;; add rgb_bounds to color in pixels.
;; December 2006: Add support for postscript output. AWD
;; January 3, 2007: draw Titan exobase. JEC.
;; June 5, 2007: Modify for Geometer version of data cube. Sync up
;; with target.
;; December 18, 2007: Updated to allow multiple output file formats. 
;; June 2, 2009: Fix animation bug. Add XSIZE and YSIZE keywords.

;;Identify the time to make the map from the user entry
IF maptime NE '' THEN CSPICE_UTC2ET, maptime, mapet

IF NOT KEYWORD_SET(POSTSCRIPT) THEN BEGIN
   red = 255                    ; Define some colors for later
   orange = 107000                
   yellow = 6553599 
   green = 65280
   cyan = 16711500
   blue = 16711680
   violet = 16600000
   white = 256L*256L*256L-1
   black = 0

   !P.THICK = 1
   !P.CHARTHICK = 1
;; Create the skymap window:

   SET_PLOT, 'X'
   IF (KEYWORD_SET(JPG) OR KEYWORD_SET(TIFF)) THEN BEGIN
      !P.BACKGROUND = WHITE
      !P.COLOR = BLACK
   ENDIF

   WINDOW, XSIZE = xsize, YSIZE = ysize

ENDIF ELSE BEGIN
; Color table written by Liam.Gumley@ssec.wisc.edu
; http://cow.physics.wisc.edu/~craigm/idl/archive/msg00492.html
   r = [0,255,0,255,0,255,0,255,0,255,255,112,219,127,0,255]
   g = [0,0,255,255,255,0,0,255,0,187,127,219,112,127,163,171]
   b = [0,255,255,0,0,0,255,255,115,0,127,147,219,127,255,127]
   TVLCT, r, g, b
   red = 5
   orange = 9
   yellow = 3
   green = 4
   cyan = 2
   blue = 6
   violet = 12
   white = 7
   SET_PLOT, 'PS'

   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF (extension NE '.ps') THEN savefile = savefile + '.ps'

   DEVICE, FILENAME=savefile, XSIZE=80, YSIZE=30, /ENCAPSULATED, $
           /COLOR, BITS_PER_PIXEL=8
   !P.THICK = 2
   !P.CHARTHICK = 2
ENDELSE

;; Put the plot time on the annotation layer:

time_label = plot_utcs[i_plot]

;;Get the string value for the target body:
targetid = WIDGET_INFO(maininfo.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, maininfo.targetdrop, GET_VALUE = targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
CSPICE_BODN2C, target, targetid, found
IF NOT found THEN BEGIN
   target = 'SATURN'
   targetid = 699L 
   ;;This should only happen if the target is the rings.
ENDIF

IF cube[0].et GT 1E19 THEN BEGIN
    ;; cube.et is set to 1E20 in GEOMETER if cube data is not to be used.
    npixel = N_ELEMENTS(data[i_time].ra_full[0,*]) 
    selected_pixels = WHERE(maininfo.pxvec)
    name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)
    channel_name = 'UVIS'
    IF maptime EQ '' THEN $
       time = data[i_time].et $
    ELSE time = mapet
ENDIF ELSE BEGIN
    npixel = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))
    name = 'UVIS'
    channel_name = 'UVIS'
    IF maptime EQ '' THEN $
       time = plot_ets[i_plot] $
    ELSE time = mapet
 ENDELSE

;; Get the camera rotation matrix used to create perspective view:
CSPICE_SPKPOS, target, time, 'J2000', 'LT', 'CASSINI', targpos, ltime
CSPICE_RECRAD, targpos/CSPICE_VNORM(targpos), unit, ratarg, dectarg
CSPICE_EUL2M, !DPI/2D0, !DPI/2D0-dectarg, ratarg, 3L, 2L, 3L, cam_mat

;; Indicate to the user which end of the slit is which by drawing the lowest 
;; numbered pixel filled in green:


IF name EQ 'Pixel 64' THEN BEGIN
    CASE maininfo.channel OF
        1: channel_name = 'UVIS EUV'
        2: channel_name = 'UVIS EUV'
        3: channel_name = 'UVIS Solar'
        4: channel_name = 'UVIS FUV'
        5: channel_name = 'UVIS FUV'
        6: channel_name = 'UVIS FUV'
        7: channel_name = 'UVIS HSP'
        8: channel_name = 'UVIS HDAC'
        ELSE: channel_name = 'UVIS'
    ENDCASE
ENDIF

; Make a circle symbol for plotting
circ = findgen(17) * (!PI*2/16.)
usersym, COS(circ), SIN(circ)

;;Get the string value for the target body:
targetid = WIDGET_INFO(maininfo.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, maininfo.targetdrop, GET_VALUE = targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
CSPICE_BODN2C, target, targetid, found
IF NOT found THEN BEGIN
    target = 'SATURN'
    targetid = 699L 
    ;;This should only happen if the target is the rings.
ENDIF

PLOT, [[0,0],[0,0]], /NODATA, XRANGE = ra_range, YRANGE = dec_range, $
      XTITLE = 'Right Ascension', YTITLE = 'Declination', /XSTYLE, $
      /YSTYLE
ra_delta = ABS(ra_range[1]-ra_range[0])
dec_delta = ABS(dec_range[1]-dec_range[0])
XYOUTS, ra_range[0]-ra_delta/30.,dec_range[0]+dec_delta/40., $
        time_label, SIZE = 1.5

;; Now draw the target body:
;;Determine a color to fill in the target body on the plot:
CASE target OF
    'SATURN': body_color = yellow
    'TITAN': body_color = orange
    ELSE: body_color = cyan
ENDCASE
CSPICE_UTC2ET, time_label, limbtime
BODY_LIMB, target, 'CASSINI', limbtime, 90, ra_target, dec_target, limberror
targ_ra_size = ABS(MAX(ra_target)-MIN(ra_target))

IF NOT limberror THEN $
  OPLOT, [ra_target,ra_target[0]], [dec_target,dec_target[0]], $
         COLOR = body_color

;;**************************************************************************
;;                                      STARS
;;**************************************************************************
IF stars GT 0. THEN BEGIN
starlim = stars
;; Load the star data file:
starfile = FILE_WHICH('star_table.dat')
IF FILE_TEST(starfile) THEN BEGIN
    RESTORE, starfile
    sp_fields = STRMID(STRTRIM(star_table.sp,2),0,1)
    ostarids = WHERE(sp_fields EQ 'O')
    ostars = star_table[ostarids]
    bstarids = WHERE(sp_fields EQ 'B')
    bstars = star_table[bstarids]
    astarids = WHERE(sp_fields EQ 'A')
    astars = star_table[astarids]
    otherstars = star_table[WHERE((sp_fields NE 'A') AND (sp_fields NE 'B') $
                                  AND (sp_fields NE 'O'))]

    bright_o = WHERE(ostars.mag LE 1.0 AND ostars.mag LT starlim)
    bright2_o = WHERE(ostars.mag GT 1.0 AND ostars.mag LE 2.0 AND $
        ostars.mag LT starlim)
    mid_o = WHERE(ostars.mag GT 2.0 AND ostars.mag LE 3.0 AND $
        ostars.mag LT starlim)
    faint_o = WHERE(ostars.mag GT 3.0 AND ostars.mag LE starlim)
    
    bright_b = where(bstars.mag LE 1.0 AND bstars.mag LT starlim)
    bright2_b = WHERE(bstars.mag GT 1.0 AND bstars.mag LE 2.0 AND $
        bstars.mag LT starlim)
    mid_b = WHERE(bstars.mag GT 2.0 AND bstars.mag LE 3.0 AND $
        bstars.mag LT starlim)
    faint_b = WHERE(bstars.mag GT 3.0 AND bstars.mag LE starlim AND $
        bstars.mag LT starlim)
        
    bright_a = WHERE(astars.mag LE 1.0 AND astars.mag LT starlim)
    bright2_a = WHERE(astars.mag GT 1.0 AND astars.mag LE 2.0 AND $
        astars.mag LT starlim)
    mid_a = WHERE(astars.mag GT 2.0 AND astars.mag LE 3.0 AND $
        astars.mag LT starlim)
    faint_a = WHERE(astars.mag GT 3.0 AND astars.mag LE starlim)
    
    ;;O Stars*************************************************************
    IF bright_o[0] NE -1 THEN $
      OPLOT, ostars[bright_o].ra, ostars[bright_o].dec, $
             PSYM=8, SYMSIZE = .7, COLOR = white

    IF bright2_o[0] NE -1 THEN $
        OPLOT, ostars[bright2_o].ra, ostars[bright2_o].dec, $
               PSYM = 8, COLOR = blue, SYMSIZE = .5

    IF mid_o[0] NE -1 THEN $
      OPLOT, ostars[mid_o].ra, ostars[mid_o].dec, $
             PSYM = 8, COLOR = cyan, SYMSIZE = .3

    IF faint_o[0] NE -1 THEN $
      OPLOT, ostars[faint_o].ra, ostars[faint_o].dec, $
        PSYM = 8, COLOR = cyan, SYMSIZE = .2

    ;;B Stars*************************************************************
    IF bright_b[0] NE -1 THEN $
        OPLOT, bstars[bright_b].ra, bstars[bright_b].dec, $
        PSYM = 8, COLOR = green, SYMSIZE = .7

    IF bright2_b[0] NE -1 THEN $
        OPLOT, bstars[bright2_b].ra, bstars[bright2_b].dec, $
        PSYM = 8, COLOR = green, SYMSIZE = .5

    IF mid_b[0] NE -1 THEN $
        OPLOT, bstars[mid_b].ra, bstars[mid_b].dec, $
        PSYM = 8, COLOR = yellow, SYMSIZE = .3

    IF faint_b[0] NE -1 THEN $
        OPLOT, bstars[faint_b].ra, bstars[faint_b].dec, $
        PSYM = 8, COLOR = yellow, SYMSIZE = .2

    ;;A Stars*************************************************************
    IF bright_a[0] NE -1 THEN $
        OPLOT, astars[bright_a].ra, astars[bright_a].dec, $
        PSYM = 8, COLOR = orange, SYMSIZE = .7

    IF bright2_a[0] NE -1 THEN $
        OPLOT, astars[bright2_a].ra, astars[bright2_a].dec, $
        PSYM = 8, COLOR = orange, SYMSIZE = .5

    IF mid_a[0] NE -1 THEN $
        OPLOT, astars[mid_a].ra, astars[mid_a].dec, $
        PSYM = 8, COLOR = red, SYMSIZE = .3

    IF faint_a[0] NE -1 THEN $
        OPLOT, astars[faint_a].ra, astars[faint_a].dec, $
        PSYM = 8, COLOR = red, SYMSIZE = .2

    ;;All other spectral types:
    OPLOT, otherstars.ra, otherstars.dec, PSYM=8, COLOR = red, SYMSIZE=.2

ENDIF ;; If star file found
ENDIF ;; If star keyword set

;;**************************************************************************
;; Plot pixels.
;;**************************************************************************

IF cube[0].et GT 1E19 THEN BEGIN
    ;; cube.et is set to 1E20 in GEOMETER if cube data is not to be used.
    ;; Using data from the GEOMETER data structure in the current session.
    IF KEYWORD_SET(ALLTIMES) THEN BEGIN
        ntimes = N_ELEMENTS(plot_ets)
        FOR time_ix = 0, ntimes-1 DO BEGIN
            OPLOT, [data[time_ix].ra_full[1:*,0],data[time_ix].ra_full[1,0]], $
                   [data[time_ix].dec_full[1:*,0],data[time_ix].dec_full[1,0]]
    
            IF npixel GT 1 THEN BEGIN
                FOR px_index = 1, npixel-1 DO $
                  OPLOT, [data[time_ix].ra_full[1:*,px_index], $
                          data[time_ix].ra_full[1,px_index]], $
                         [data[time_ix].dec_full[1:*,px_index], $
                          data[time_ix].dec_full[1,px_index]]
                
            ENDIF
        ENDFOR  ;; Looping over times.
;; Just label the last time:
        XYOUTS, data[ntimes-1].ra_full[1,0],data[ntimes-1].dec_full[1,0], $
                channel_name
        
    ENDIF ELSE BEGIN
        OPLOT, [data[i_time].ra_full[1:*,0],data[i_time].ra_full[1,0]], $
               [data[i_time].dec_full[1:*,0],data[i_time].dec_full[1,0]]
        
        XYOUTS, data[i_time].ra_full[1,0],data[i_time].dec_full[1,0], $
                channel_name
        
        IF npixel GT 1 THEN BEGIN
            FOR px_index = 1, npixel-1 DO BEGIN
                OPLOT, [data[i_time].ra_full[1:*,px_index], $
                        data[i_time].ra_full[1,px_index]], $
                       [data[i_time].dec_full[1:*,px_index], $
                        data[i_time].dec_full[1,px_index]]
                
            ENDFOR
            name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)
            XYOUTS, data[i_time].ra_full[1,0], $
                    data[i_time].dec_full[1,0], name
        ENDIF
    ENDELSE  ;;Only plotting at one selected time.

ENDIF ELSE BEGIN

;; Plotting data from loaded image cube (also computed by GEOMETER, but
;; run and saved at a different time).

    nrows = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))
    ntimes = N_ELEMENTS(plot_ets)

;; Figure out colors. Scale each channel to 255.
    IF colormax[0] EQ 0. THEN BEGIN
        redmax = 1.E-20
        FOR time_ix = 0, ntimes-1 DO $
          FOR px_index = 0, npixel-1 DO $
            redmax = redmax > $
          MAX(TOTAL(cube[time_ix].data[px_index,$
                              spec_bounds[0]:spec_bounds[1]]))
    ENDIF ELSE redmax = colormax[0]
    IF colormax[1] EQ 0. THEN BEGIN
        greenmax = 1.E-20
        FOR time_ix = 0, ntimes-1 DO $
          FOR px_index = 0, npixel-1 DO $
            greenmax = greenmax > $
          MAX(TOTAL(cube[time_ix].data[px_index,$
                              spec_bounds[2]:spec_bounds[3]]))
    ENDIF ELSE greenmax = colormax[1]
    IF colormax[2] EQ 0. THEN BEGIN
        bluemax = 1.E-20
        FOR time_ix = 0, ntimes-1 DO $
          FOR px_index = 0, npixel-1 DO $
            bluemax = bluemax > $
          MAX(TOTAL(cube[time_ix].data[px_index,$
                              spec_bounds[4]:spec_bounds[5]]))
    ENDIF ELSE bluemax = colormax[2]

    IF KEYWORD_SET(ALLTIMES) THEN BEGIN
        FOR time_ix = 0, ntimes-1 DO BEGIN

            FOR px_index = 0, npixel-1 DO BEGIN
                ;; Build vectors for plotting the boundary of the slit:
                ravec = [cube[time_ix].ra_full[*,px_index], $
                         cube[time_ix].ra_full[0,px_index]]
                decvec = $
                  [cube[time_ix].dec_full[*,px_index], $
                   cube[time_ix].dec_full[0,px_index]]

                ;;Get the brightness of the pixel:
                redcolor = TOTAL(cube[time_ix].data[px_index,$
                                           spec_bounds[0]:spec_bounds[1]]) $
                           < redmax
                redcolor /= redmax
                redcolor *= 255.

                greencolor = TOTAL(cube[time_ix].data[px_index,$
                                             spec_bounds[2]:spec_bounds[3]]) $
                             < greenmax
                greencolor /= greenmax
                greencolor *= 255.

                bluecolor = TOTAL(cube[time_ix].data[px_index,$
                                            spec_bounds[4]:spec_bounds[5]]) $
                            < bluemax
                bluecolor /= bluemax
                bluecolor *= 255.

                ;; Now combine these three shades into a single color:
;; problem: PS doesn't accept TrueColor plots. Only images. 
                TVLCT, INDGEN(256), INDGEN(256), INDGEN(256)

                pxcolor = LONG(redcolor)+256L*LONG(greencolor)+$
                          256L*256L*LONG(bluecolor)

                OPLOT, ravec, decvec, COLOR = pxcolor
                POLYFILL, ravec, decvec, COLOR = pxcolor

            ENDFOR  ;; Building pixel vectors.
        ENDFOR  ;; Looping over times.
;; Just label the last time:
        XYOUTS, cube[time_ix-1].ra_full[0,0], $
                cube[time_ix-1].dec_full[0,0], channel_name
        
    ENDIF ELSE BEGIN
;; Just plotting at one selected time.
        FOR px_index = 0, npixel-1 DO BEGIN
            ravec = [cube[i_time].ra_full[*,px_index], $
                     cube[i_time].ra_full[0,px_index]]
            
            decvec = $
              [cube[i_time].dec_full[*,px_index], $
               cube[i_time].dec_full[0,px_index]]

            ;;Get the color of the pixel:
            redcolor = TOTAL(cube[i_time].data[px_index,$
                                       spec_bounds[0]:spec_bounds[1]]) $
                       < redmax
            redcolor /= redmax
            redcolor *= 255.
            
            greencolor = TOTAL(cube[i_time].data[px_index,$
                                         spec_bounds[2]:spec_bounds[3]]) $
                         < greenmax
            greencolor /= greenmax
            greencolor *= 255.
            
            bluecolor = TOTAL(cube[i_time].data[px_index,$
                                        spec_bounds[4]:spec_bounds[5]]) $
                        < bluemax
            bluecolor /= bluemax
            bluecolor *= 255.
            
            ;; Now combine these three shades into a single color:
            pxcolor = LONG(redcolor)+256L*LONG(greencolor)+$
                      256L*256L*LONG(bluecolor)

            OPLOT, ravec, decvec, COLOR = pxcolor
            POLYFILL, ravec, decvec, COLOR = pxcolor
        ENDFOR
        name = 'Pixel '+STRTRIM(STRING(px_index-1),2)
        XYOUTS, cube[i_time].ra_full[0,px_index-1], $
                cube[i_time].dec_full[0,px_index-1], channel_name
    ENDELSE  ;;Only plotting at one selected time.
    
ENDELSE  ;;Plotting data cube pixels.
;;*************************************************************************
;;                              TARGET BODY
;;*************************************************************************
;; Draw the terminator on the target body:
ra_term = [0D0]
dec_term = [0D0]
CSPICE_BODVAR, targetid, 'RADII', radii
CSPICE_SPKPOS, target, plot_ets[i_plot], 'J2000', 'LT', 'CASSINI', scpos, ltime
CSPICE_SPKPOS, target, plot_ets[i_plot], 'J2000', 'LT', 'SUN', sunpos, ltimesun
CSPICE_EDLIMB, radii[0], radii[1], radii[2], sunpos, term
;; term is a SPICE ellipse defining the terminator.
CSPICE_EL2CGV, term, center, v1, v2

;; Get RA/Dec points along the terminator of the body:
FOR j = 0, 71 DO BEGIN
    AZ = !DPI*2.0/71D0*j-!DPI
    VEC2 = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    termvec = scpos-vec2
    ;; This is actually a less then fully accurate check to
    ;; see if the terminator line is behind the limb...
    IF CSPICE_VNORM(termvec) LT CSPICE_VNORM(scpos) THEN BEGIN
        CSPICE_RECRAD, termVEC, RANGE, RA1, DEC1
        ra_term = [ra_term, RA1*!RADEG]
        dec_term = [dec_term, DEC1*!RADEG]
    ENDIF
ENDFOR

IF N_ELEMENTS(ra_term) GT 1 THEN ra_term = ra_term[1:*]
IF N_ELEMENTS(dec_term) GT 1 THEN dec_term = dec_term[1:*]

;; Remove discontinuity across the body:
delta_ra_term = ra_term-SHIFT(ra_term,1)
IF N_ELEMENTS(delta_ra_term) GT 1 THEN delta_ra_term = delta_ra_term[1:*]

void = MAX(ABS(delta_ra_term),breakloc)
;; Rearrange to avoid drawing across the body:
IF void GT targ_ra_size/2. THEN BEGIN
    ra_term_temp = [ra_term[breakloc:*], ra_term[0:breakloc-1]]
    dec_term_temp = [dec_term[breakloc:*], dec_term[0:breakloc-1]]
    ra_term = ra_term_temp[1:*]
    dec_term = dec_term_temp[1:*]
ENDIF
delta_dec_term = dec_term-SHIFT(dec_term,1)
IF N_ELEMENTS(delta_dec_term) GT 1 THEN delta_dec_term = delta_dec_term[1:*]
void = MAX(ABS(delta_dec_term),breakloc)
;; Rearrange to avoid drawing across the body:
targ_dec_size = MAX(dec_target)-MIN(dec_target)
IF void GT targ_dec_size/2. THEN BEGIN
    ra_term_temp = [ra_term[breakloc:*], ra_term[0:breakloc-1]]
    dec_term_temp = [dec_term[breakloc:*], dec_term[0:breakloc-1]]
    ra_term = ra_term_temp[1:*]
    dec_term = dec_term_temp[1:*]
ENDIF
OPLOT, ra_term, dec_term, COLOR = red

;;Label the target body. 
XYOUTS, MAX(ra_target), MEAN(dec_target), target

;; ************************LATITUDE/LONGITUDE*****************************
CSPICE_PXFORM, 'J2000', 'IAU_'+target, time, inert2p


IF KEYWORD_SET(LATLON) THEN BEGIN
;; To get the Lat/Lon points: get vectors from body center to lat/lon
;; positions, then add those to vector from S/C to planet, convert to RA/Dec
;; and then convert to U/V to plot.
ndrawpoints = 100
ra_lat_vec = DBLARR(ndrawpoints)
dec_lat_vec = DBLARR(ndrawpoints)
visible_vec = BYTARR(ndrawpoints)
long_int = 30D0*!DTOR
lat_int = 15D0*!DTOR
lat_min = -!DPI/2.+lat_int
lat_max = !DPI/2-lat_int
nlatlines = (lat_max-lat_min)/lat_int+1

FOR lat_ix = 0, nlatlines-1 DO BEGIN
    latvec = DBLARR(ndrawpoints)+lat_min+lat_ix*lat_int
    lonvec = DINDGEN(ndrawpoints)/(ndrawpoints-1D0)*!DPI*2D0
    CSPICE_LATREC, DBLARR(ndrawpoints)+radii[0], lonvec, latvec, $
                   latlonpos

    FOR vec_ix = 0, ndrawpoints-1 DO BEGIN
       latlonvec = REFORM(latlonpos[*,vec_ix])
       latlonvec /= CSPICE_VNORM(latlonvec)
       latlonvec *= SQRT((radii[0]*COS(latvec[lat_ix]))^2+$
                         (radii[2]*SIN(latvec[lat_ix]))^2)
       ;; latlonvec is the vector to the particular point at this 
       ;; latitude and longitude on the surface of oblate spheroid
       ;; target (note, not coded for triaxial ellipsoid, just oblate
       ;; spheroid).
       CSPICE_MTXV, (inert2p), latlonvec, latlonvec_inert
       vec2point = latlonvec_inert-scpos
       IF CSPICE_VNORM(vec2point) LE CSPICE_VNORM(scpos) THEN $
          visible_vec[vec_ix] = 1 ELSE visible_vec[vec_ix] = 0
       CSPICE_RECRAD, -vec2point, dist, ra_pnt, dec_pnt
       ra_lat_vec[vec_ix] = ra_pnt*!RADEG
       dec_lat_vec[vec_ix] = dec_pnt*!RADEG
    ENDFOR

    good = WHERE(visible_vec)
    IF N_ELEMENTS(good) GT 2 THEN BEGIN
        delta_good = good-SHIFT(good,1)
        junk = MAX(delta_good, breakloc)

        good_rs = ra_lat_vec[good[0:breakloc-1]]
        good_ds = dec_lat_vec[good[0:breakloc-1]]

        latlabel_r = MEAN(good_rs)
        junk = MIN(ABS(latlabel_r-good_rs), loc)
        latlabel_d = good_ds[loc]

        IF lat_ix EQ FIX(nlatlines/2) THEN $
          OPLOT, good_rs, good_ds, COLOR = white ELSE $
            OPLOT, good_rs, good_ds, COLOR = cyan
        good_rs = ra_lat_vec[good[breakloc:*]]
        good_ds = dec_lat_vec[good[breakloc:*]]
        IF lat_ix EQ FIX(nlatlines/2) THEN $ 
           OPLOT, good_rs, good_ds, COLOR = white ELSE $
              OPLOT, good_rs, good_ds, COLOR = cyan


        IF lat_ix EQ FIX(nlatlines/2-1) THEN $
          XYOUTS, latlabel_r, latlabel_d, $
                  STRMID(STRING(STRTRIM(ROUND(lat_int*!RADEG),2)),0,2) ELSE $
                    IF lat_ix EQ FIX(nlatlines/2+1) THEN $
                      XYOUTS, latlabel_r, latlabel_d, '-'+$
          STRMID(STRING(STRTRIM(ROUND(lat_int*!RADEG),2)),0,2)
    ENDIF
ENDFOR

;; Same thing for longitude:
ra_lon_vec = DBLARR(ndrawpoints)
dec_lon_vec = DBLARR(ndrawpoints)
lon_min = 0D0
lon_max = !DPI*2.-long_int
nlonlines = (lon_max-lon_min)/long_int+1
FOR lon_ix = 0, nlonlines-1 DO BEGIN
    lonvec = DBLARR(ndrawpoints)+lon_min+lon_ix*long_int
    latvec = DINDGEN(ndrawpoints)/(ndrawpoints-1D0)*!DPI-!DPI/2D0
    CSPICE_LATREC, DBLARR(ndrawpoints)+radii[0], lonvec, latvec, $
                   latlonpos

    ;; I think I can get rid of the following loop by using vectorized
    ;; CSPICE calls, but just get it working for now and worry about
    ;; performance later.
    FOR vec_ix = 0, ndrawpoints-1 DO BEGIN
        latlonvec = REFORM(latlonpos[*,vec_ix])
        latlonvec /= CSPICE_VNORM(latlonvec)
        latlonvec *= SQRT((radii[0]*COS(latvec[lat_ix]))^2+$
                          (radii[2]*SIN(latvec[lat_ix]))^2)
        ;; latlonvec is the vector to the particular point at this 
        ;; latitude and longitude on the surface of oblate spheroid
        ;; target (note, not coded for triaxial ellipsoid, just oblate
        ;; spheroid).
        CSPICE_MTXV, (inert2p), latlonvec, latlonvec_inert
        vec2point = latlonvec_inert-scpos
        IF CSPICE_VNORM(vec2point) LE CSPICE_VNORM(scpos) THEN $
          visible_vec[vec_ix] = 1 ELSE visible_vec[vec_ix] = 0
        CSPICE_RECRAD, -vec2point, dist, ra_pnt, dec_pnt
        ra_lon_vec[vec_ix] = ra_pnt*!RADEG
        dec_lon_vec[vec_ix] = dec_pnt*!RADEG
    ENDFOR
    good = WHERE(visible_vec)
    IF N_ELEMENTS(good) GT 2 THEN BEGIN
        delta_good = good-SHIFT(good,1)
        junk = MAX(delta_good, breakloc)
        good_rs = ra_lon_vec[good[0:breakloc-1]]
        good_ds = dec_lon_vec[good[0:breakloc-1]]
        OPLOT, good_rs, good_ds, COLOR = cyan
        good_rs = ra_lon_vec[good[breakloc:*]]
        good_ds = dec_lon_vec[good[breakloc:*]]
        OPLOT, good_rs, good_ds, COLOR = cyan
        lon_d_label = MEAN(good_ds)
        junk = MIN(ABS(lon_d_label-good_ds), loc)
        IF lon_ix EQ 0 THEN $
          XYOUTS, good_rs[loc], lon_d_label, '0' ELSE $
            IF lon_ix EQ FIX(nlonlines/2-1) THEN $
              XYOUTS, good_rs[loc], lon_d_label, '180'
    ENDIF
ENDFOR
ENDIF  ;; LATLON keyword set.

;;*************************************************************************
;;                            OTHER BODIES
;;*************************************************************************
;; Draw the limbs of all bodies:

body_codes = LONG([10, 399, 599, 699, 601, 602, 603, 604, 605, $
    606, 607, 608, 609, 666])
nbodies = N_ELEMENTS(body_codes)
body_ids = ['']
FOR bodycount = 0, nbodies-1 DO BEGIN
   CSPICE_BODC2N, body_codes[bodycount], bodyname, found
    IF found AND bodyname NE target THEN BEGIN
        BODY_LIMB, bodyname, 'CASSINI', plot_ets[i_plot], 36, ra_limb, $
                   dec_limb, limberror
        IF NOT limberror THEN $
            OPLOT, [ra_limb,ra_limb[0]], [dec_limb,dec_limb[0]]
    ENDIF
    ;;Label the body:
    XYOUTS, ra_limb[0], dec_limb[0], bodyname
ENDFOR

;; Do the Titan exobase:
TITAN_EXO_LIMB, 'CASSINI', time, 36, ra_exo, dec_exo
OPLOT, [ra_exo, ra_exo[0]], [dec_exo, dec_exo[0]]

;;*************************************************************************
;;                              RINGS
;;*************************************************************************
;; Do Saturn's rings. For now, hardwire the critical ring boundaries.
;; Would be better eventually to read these values in from PCK.
ringrad = [66970.,74510.,92000.,117580.,122170.,133405.,133730.,136780.,$
    140180.,140270.]
nring = N_ELEMENTS(ringrad)    

CSPICE_SPKPOS, 'SATURN', plot_ets[i_plot], 'J2000', 'LT', 'CASSINI', $
    scpos, lt_corr
;; Get the rotation matrix between J2000 and Saturn:
CSPICE_PXFORM, 'J2000', 'IAU_SATURN', plot_ets[i_plot]-lt_corr, inert2p
inert2p = TRANSPOSE(inert2p)
scpos_p = inert2p#scpos
CSPICE_BODVAR, 699L, 'RADII', satrad

nringpoints = 36
;;Use 36 points for each ring

FOR ringindex = 0, nring-1 DO BEGIN
  ringra = DBLARR(nringpoints)
  ringdec = DBLARR(nringpoints)

  FOR ringpoint = 0, nringpoints-1 DO BEGIN
  az = !DPI*2.0*DOUBLE(ringpoint)/DOUBLE(nringpoints)-!DPI
  rvec = [COS(az), SIN(az), 0.0]*ringrad[ringindex]
  ;; RVEC is the vector to the ring in body coordinates.

  vec = TRANSPOSE(inert2p)#(scpos_p+rvec)
  ;; VEC is the vector from the S/C to the ring in IREF.
  ;; Check to see if it intercepts the planet.

  CSPICE_SURFPT,-scpos_p,inert2p#VEC,satrad[0],satrad[1],satrad[2], $
    POINT, found
  HIDDEN = 0B
  IF found THEN BEGIN
    ;; See if the point is behind the planet.
    IF (CSPICE_VNORM(SCPOS) LT CSPICE_VNORM(VEC)) THEN HIDDEN = 1B
  ENDIF
  IF (NOT HIDDEN) THEN BEGIN
    CSPICE_RECRAD, VEC, RANGE, RA1, DEC1
    ringra[ringpoint] = RA1*!RADEG
    ringdec[ringpoint] = DEC1*!RADEG
  ENDIF
  ENDFOR    ;;ringpoint

  ;; Hidden parts of the rings have coordinate values of [0.,0.].
  ;; Identify which points are not hidden:
  good = WHERE(ringra NE 0D0)
  ;; Find where there is a break in good:
  void = MAX(good-SHIFT(good,1), breakloc)
  ;; Rearrange the order of points so the ring is drawn without 
  ;; connecting behind the body:
  IF void GT 1 THEN good = [good[breakloc:*],good[0:breakloc-1]]

  OPLOT, ringra[good], ringdec[good]

ENDFOR  ;;ringindex

;; Write JPEG or TIFF files:

IF KEYWORD_SET(JPG) THEN BEGIN
   win = TVRD(TRUE = 1)
   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF ((extension NE '.jpg') AND (extension NE 'jpeg')) THEN $
      savefile = savefile + '.jpg'
   WRITE_JPEG, savefile, win, QUALITY = 100, TRUE = 1
ENDIF

IF KEYWORD_SET(TIFF) THEN BEGIN
   win = TVRD(TRUE = 1)
   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF ((extension NE '.tif') AND (extension NE 'tiff')) THEN $
      savefile = savefile + '.tif'
   WRITE_TIFF, savefile, win, ORIENTATION = 0
ENDIF

IF KEYWORD_SET(POSTSCRIPT) THEN DEVICE, /CLOSE

END
