@bodymap_panel.pro
@bodymap_panel_event.pro
@bodymap_panel_callback.pro
@bodymap_tool.pro
@bodymap_tool__define.pro

;;File Version: 1.1
;;Last Modified: November 17, 2004
;;By: JEC

PRO GEOMETER_BODYMAP, data, maininfo, i_time, GRID = GRID, UPDATE = UPDATE

;;  >> Abstract <<
;;  Only to be called by GEOMETER.

;;  i_time is the index of the desired time for computing body data.
;;  maininfo is the UVALUE structure from GEOMETER.
;;  Josh Colwell
;; November 17, 2004: Add Panel and use custom Bodymap iTool.

;; Create the Bodymap iTool with its custom panel:
IF NOT KEYWORD_SET(UPDATE) THEN BODYMAP_TOOL

;; Get the tool's object reference for sending text labels to the vis layer:
idtool = ITGETCURRENT(TOOL = oTool)

;; Change the current_zoom property of the view so that we don't have
;; so much white space around it:
view = idtool+'/WINDOW/VIEW_1'
IF NOT KEYWORD_SET(UPDATE) THEN BEGIN
    void = oTool->doSetProperty(view, 'CURRENT_ZOOM', 1.5)
    oTool->setproperty, PTR_NEW(data), 'DATA'
    oTool->setproperty, PTR_NEW(maininfo), 'MAININFO'
    oTool->setproperty, PTR_NEW(i_time), 'ITIME'
ENDIF

oTool->commitactions

;; Put the plot time on the annotation layer:
time_label = data[i_time].utc
;; Not yet implemented. Need to find method to add text to annotation layer.


npixel = N_ELEMENTS(data[i_time].lon_full[0,*])

;; Indicate to the user which end of the slit is which by drawing the lowest 
;; numbered pixel filled in green.
selected_pixels = WHERE(maininfo.pxvec)
name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)

;; Store the object IDs of all plotted pixels.
pixel_ids = ['']

;; If the pixel is not entirely on the body, then use symbols and no line 
;; to draw the pixel boundary. Otherwise draw solid lines. 
IF MIN(data[i_time].lon_full[1:*,0]) EQ -1000. AND $
  MAX(data[i_time].lon_full[1:*,0]) NE -1000. THEN $
  IPLOT, data[i_time].lon_full[1:*,0], data[i_time].lat_full[1:*,0], $
         NAME = name, /FILL_BACKGROUND, FILL_COLOR = [0,255,0], $
         TRANSPARENCY = 50, $
         XTITLE = 'West Longitude (Degrees)', YTITLE = 'Latitude (Degrees)', $
         IDENTIFIER = bodymap_id, XRANGE = [0, 360], YRANGE = [-90,90], $
         SYM_INDEX = 1, LINESTYLE = 6, /OVERPLOT ELSE $
  IPLOT, [data[i_time].lon_full[1:*,0],data[i_time].lon_full[1,0]], $
         [data[i_time].lat_full[1:*,0],data[i_time].lat_full[1,0]], $
         NAME = name, /FILL_BACKGROUND, FILL_COLOR = [0,255,0], $
         TRANSPARENCY = 50, $
         XTITLE = 'West Longitude (Degrees)', YTITLE = 'Latitude (Degrees)', $
         IDENTIFIER = bodymap_id, XRANGE = [0, 360], YRANGE = [-90,90], $
         /OVERPLOT

pixel_ids = [pixel_ids, (oTool->getselecteditems())->getfullidentifier()]

;; Also store the object IDs of all labels.
label_ids = ['']

IF MIN(data[i_time].lon_full[*,0]) GE 0 THEN BEGIN
    oTextp = OBJ_NEW('IDLitVisText')
    oTool->Add, oTextp
    gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
    gTextp->SetProperty, STRING = name, LOCATION = $
     [data[i_time].lon_full[1,0],data[i_time].lat_full[1,0]]
    label_ids = [label_ids, gTextp->getfullidentifier()]
    pixel_ids = [pixel_ids, gTextp->getfullidentifier()]
ENDIF

;; Get the object reference for the data space so we can control the
;; plotting range.
dataspaceid = 'WINDOW/VIEW_1/VISUALIZATION LAYER/DATA SPACE'
oDataSpace = oTool->getbyidentifier(dataspaceid)
oDataSpace->setproperty, x_auto_update= 0B
oDataSpace->setProperty, y_auto_update= 0B

IF npixel GT 1 THEN BEGIN
    label_flag = 0  ;;Use this to label only every other pixel.
    FOR px_index = 1, npixel-1 DO BEGIN
        name = 'Pixel '+STRTRIM(STRING(selected_pixels[px_index]),2)
        IF MIN(data[i_time].lon_full[1:*,px_index]) EQ -1000. AND $
          MAX(data[i_time].lon_full[1:*,px_index]) NE -1000. THEN $
          IPLOT, data[i_time].lon_full[1:*,px_index], $
                 data[i_time].lat_full[1:*,px_index], $
                 NAME = name, XRANGE = [0, 360], YRANGE = [-90,90], $
                 SYM_INDEX = 1, LINESTYLE = 6, /OVERPLOT ELSE $
          IF MAX(data[i_time].lon_full[1:*,px_index]) NE -1000. THEN $
          IPLOT, [data[i_time].lon_full[1:*,px_index],$
                  data[i_time].lon_full[1,px_index]], $
                 [data[i_time].lat_full[1:*,px_index],$
                  data[i_time].lat_full[1,px_index]], /OVERPLOT, $
                 NAME = name, XRANGE = [0, 360], YRANGE = [-90,90]
        pixel_ids = [pixel_ids, (oTool->getselecteditems())->$
                     getfullidentifier()]

        IF label_flag AND $
          MIN(data[i_time].lon_full[*,px_index]) GE 0 THEN BEGIN
            label_flag = 0
            oTextp = OBJ_NEW('IDLitVisText')
            oTool->Add, oTextp
            gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
            gTextp->SetProperty, STRING = name, LOCATION = $
                                 [data[i_time].lon_full[1,px_index], $
                                  data[i_time].lat_full[1,px_index]]
            label_ids = [label_ids, gTextp->getfullidentifier()]
            ;; Including the label in the pixel list so that it goes
            ;; away when the pixels are turned off or when the labels
            ;; are turned off.
            pixel_ids = [pixel_ids, gTextp->getfullidentifier()]
        ENDIF ELSE label_flag = 1
    ENDFOR
ENDIF
;; All done with pixels, so get rid of the dummy first element.
pixel_ids = pixel_ids[1:*]
oTool->setproperty, PTR_NEW(pixel_ids), 'PIXELS'

;;*****************************************************************************
;;                                TERMINATOR
;;*****************************************************************************
;; store the IDs of all markers on the body (limb, terminator, etc.)
marker_ids = ['']

;; Draw the terminator:
lon_term = [0D0]
lat_term = [0D0]
;;Get the string value for the target body:
targetindex = WIDGET_INFO(maininfo.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, maininfo.targetdrop, GET_VALUE = targetdropvalue
target = STRUPCASE(targetdropvalue[targetindex])
CSPICE_BODN2C, target, targetid, found
CSPICE_BODVAR, targetid, 'RADII', radii
CSPICE_SPKPOS, target, data[i_time].et, 'J2000', 'LT', 'CASSINI', scpos, ltime
CSPICE_SPKPOS, target, data[i_time].et, 'J2000', 'LT', 'SUN', sunpos, ltimesun
CSPICE_EDLIMB, radii[0], radii[1], radii[2], sunpos, term
;; term is a SPICE ellipse defining the terminator.
CSPICE_EL2CGV, term, center, v1, v2

framestring = 'IAU_'+STRUPCASE(STRCOMPRESS(target))
CSPICE_SPKPOS, target, data[i_time].et, framestring, 'LT', 'CASSINI', $
               scpos_p, ltime
CSPICE_PXFORM, 'J2000', framestring, data[i_time].et-ltime, inert2p
inert2p = TRANSPOSE(inert2p) ;;IDL thing.

;; Get 72 points along the terminator of the body:
FOR j = 0, 71 DO BEGIN
    AZ = !DPI*2.0/71D0*j-!DPI
    termvec = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    termvec = inert2p#termvec
    CSPICE_RECLAT, termvec, bodintercept, lon, lat
    lon_term = [lon_term,(2*!DPI-lon)*!RADEG MOD 360D0]
    lat_term = [lat_term, lat*!RADEG]
ENDFOR
lon_term = lon_term[1:*]
lat_term = lat_term[1:*]  ;;eliminate the dummy leading zero.

;; To avoid a crossover line connecting from 0 to 360 longitude, rearrange the
;; terminator points so that lines aren't drawn across the plot. Assume that
;; we have enough points on the terminator so that a jump in terminator 
;; longitude of 180 degrees indicates a crossover.
j = 0
REPEAT BEGIN
    IF (ABS(lon_term[j+1]-lon_term[j]) GT 180.) THEN BEGIN
;;  Found a discontinuity. Plot up to index=j separately from j+1:*.
        lon_term1 = lon_term[0:j]
        lat_term1 = lat_term[0:j]
        lon_term = lon_term[j+1:*]
        lat_term = lat_term[j+1:*]
        IF (j GT 0) THEN IPLOT, lon_term1, lat_term1, /OVERPLOT, $
          NAME = 'Terminator 2', COLOR = [255, 0, 0], $
          XRANGE = [0,360], YRANGE = [-90,90]
        j = N_ELEMENTS(lon_term)
        marker_ids = [marker_ids, (oTool->getselecteditems())->$
                      getfullidentifier()]
    ENDIF
    j = j+1
ENDREP UNTIL j GE N_ELEMENTS(lon_term)-2
IPLOT, lon_term, lat_term, /OVERPLOT, NAME = 'Terminator 1', COLOR = $
       [255, 0, 0], XRANGE = [0, 360], YRANGE = [-90, 90]
marker_ids = [marker_ids, (oTool->getselecteditems())->getfullidentifier()]
;;*****************************************************************************
;;                                LIMB
;;*****************************************************************************
;; Do the same thing with the limb:
lon_limb = [0D0]
lat_limb = [0D0]
CSPICE_EDLIMB, radii[0], radii[1], radii[2], scpos_p, limb
;; limb is a SPICE ellipse.
CSPICE_EL2CGV, limb, center, v1, v2

;; Get 72 points along the limb of the body:
FOR j = 0, 71 DO BEGIN
    AZ = !DPI*2.0/71D0*j-!DPI
    limbvec = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    CSPICE_RECLAT, limbvec, bodintercept, lon, lat
    lon_limb = [lon_limb, (2*!DPI-lon)*!RADEG MOD 360D0]
    lat_limb = [lat_limb, lat*!RADEG]
ENDFOR
lon_limb = lon_limb[1:*]
lat_limb = lat_limb[1:*]  ;;eliminate the dummy leading zero.

;; To avoid a crossover line connecting from 0 to 360 longitude, rearrange the
;; limb points so that lines aren't drawn across the plot. 

j = 0
REPEAT BEGIN
    IF (ABS(lon_limb[j+1]-lon_limb[j]) GT 180.) THEN BEGIN
;;  Found a discontinuity. Plot up to index=j separately from j+1:*.
        lon_limb1 = lon_limb[0:j]
        lat_limb1 = lat_limb[0:j]
        lon_limb = lon_limb[j+1:*]
        lat_limb = lat_limb[j+1:*]
        IF (j GT 0) THEN IPLOT, lon_limb1, lat_limb1, /OVERPLOT, $
          NAME = 'Limb 2', COLOR = [0, 0, 255], $
          XRANGE = [0,360], YRANGE = [-90,90]
        j = N_ELEMENTS(lon_limb)
        marker_ids = [marker_ids, (oTool->getselecteditems())->$
                      getfullidentifier()]
    ENDIF
    j = j+1
ENDREP UNTIL j GE N_ELEMENTS(lon_limb)-2
IPLOT, lon_limb, lat_limb, /OVERPLOT, NAME = 'Limb 1', COLOR = $
       [0, 0, 255], XRANGE = [0, 360], YRANGE = [-90, 90]
marker_ids = [marker_ids, (oTool->getselecteditems())->$
              getfullidentifier()]

;;Plot the sub-solar and sub-spacecraft points:
IPLOT, [-1000, data[i_time].subsolarlon], [-1000, data[i_time].subsolarlat], $
       LINESTYLE = 6, SYM_INDEX = 4, SYM_COLOR = [255,0,0], /OVERPLOT
marker_ids = [marker_ids, (oTool->getselecteditems())->$
              getfullidentifier()]

IPLOT, [-1000, data[i_time].subsclon], [-1000, data[i_time].subsclat], $
       LINESTYLE = 6, SYM_INDEX = 5, SYM_COLOR = [0, 0, 255], /OVERPLOT
marker_ids = [marker_ids, (oTool->getselecteditems())->$
              getfullidentifier()]
;;Label these points:
oTextp = OBJ_NEW('IDLitVisText')
oTool->Add, oTextp
gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
gTextp->SetProperty, STRING = 'Sub-Solar', LOCATION = $
     [data[i_time].subsolarlon,data[i_time].subsolarlat]
label_ids = [label_ids, gTextp->getfullidentifier()]
marker_ids = [marker_ids, gTextp->getfullidentifier()]

oTextp = OBJ_NEW('IDLitVisText')
oTool->Add, oTextp
gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
gTextp->SetProperty, STRING = 'Sub-S/C', LOCATION = $
     [data[i_time].subsclon,data[i_time].subsclat]
label_ids = [label_ids, gTextp->getfullidentifier()]
marker_ids = [marker_ids, gTextp->getfullidentifier()]

;; That's it for the markers.
marker_ids = marker_ids[1:*]
oTool->setProperty, PTR_NEW(marker_ids), 'MARKERS'
;;*****************************************************************************
;;                              GRID
;;*****************************************************************************
IF NOT KEYWORD_SET(UPDATE) AND KEYWORD_SET(GRID) THEN BEGIN
;; Now draw some Lat/Lon gridlines:
lon_inc = 30.
lat_inc = 30.
grid_ids = ['']

FOR longrid = 360-lon_inc, lon_inc, -lon_inc DO BEGIN
    xcoords = [longrid,longrid]
    ycoords = [-90,90]
    IPLOT, xcoords, ycoords, /OVERPLOT
    grid_ids = [grid_ids, (oTool->getselecteditems())->getfullidentifier()]
ENDFOR

FOR latgrid = -90+lat_inc, 90-lat_inc, lat_inc DO BEGIN
    IPLOT, [360,0],[latgrid,latgrid], /OVERPLOT
    grid_ids = [grid_ids, (oTool->getselecteditems())->getfullidentifier()]
ENDFOR
grid_ids = grid_ids[1:*]
oTool->setproperty, PTR_NEW(grid_ids), 'GRID'

ENDIF

label_ids = label_ids[1:*]
oTool->setProperty, PTR_NEW(label_ids), 'VISLABELS'
END

