;;File Version: 1.0
;;Last Modified: June 5, 2009
;;Modified by: JEC

PRO GEOMETER_CKWRITER, pstate

;;+
;;
;; This program writes a C-smithed C kernel by applying rotations
;; determined by the user to an existing C kernel.
;;
;; << Inputs >>
;;
;; Time: the time at which computations will be made, in NAIF format.
;; ckin: the C kernel to be modified.
;; pstate: The pointer to the main Geometer state structure.
;;
;; << Outputs >>
;;
;; ckout: the new modified C kernel.
;;
;; << Restrictions >>
;;
;; Uses the CSPICE DLM for IDL.
;;
;; << History >>
;;
;; Joshua Colwell
;; Begun: June 4, 2009.

COMPILE_OPT idl2

;; Select the C kernel to be modified:
ckin = DIALOG_PICKFILE(FILTER = "*.bc", TITLE = "Select the input C kernel", $
                      /MUST_EXIST, PATH = (*pstate).kernelpath)

ckout = DIALOG_PICKFILE(TITLE = "Enter the output C kernel name", $
                       PATH = (*pstate).kernelpath)

sc_id = -82000L
clock_id = -82L

;;Get times to compute new C matrices:
WIDGET_CONTROL, (*pstate).starttext, GET_VALUE = startvalue
WIDGET_CONTROL, (*pstate).endtext, GET_VALUE = endvalue
WIDGET_CONTROL, (*pstate).dttext, GET_VALUE = dtvalue

CSPICE_UTC2ET, startvalue[0], et_start
CSPICE_UTC2ET, endvalue[0], et_end
dt = FLOAT(dtvalue[0])
ntime = LONG((et_end-et_start)/dt+1L)
etvec = DINDGEN(ntime)*dt+et_start
tickvec = DBLARR(ntime)
CSPICE_SCE2T, clock_id, et_start, tick_start
CSPICE_SCE2T, clock_id, et_end, tick_end

;; Get the offset rotation matrix:
CSPICE_EUL2M, (*pstate).zoffset, (*pstate).yoffset, (*pstate).xoffset, $
              3L, 2L, 1L, csmith

;; Open a new C kernel file:
CSPICE_CKOPN, ckout, "C-Smithed kernel written by Geometer", 10, ck_handle

;; Define the vector of quaternion values:
quats = DBLARR(4, ntime)
avvs = DBLARR(3, ntime)
FOR tt = 0, ntime-1 DO BEGIN
   CSPICE_SCE2T, clock_id, etvec[tt], ticks
   tickvec[tt] = ticks
   CSPICE_CKGPAV, -82000L, ticks, dt*2, 'J2000', cmat, av, clkout, found
   IF found THEN BEGIN
      newcmat = csmith##cmat
      ;; Convert the new rotation matrix to a quaternion:
      CSPICE_M2Q, newcmat, work_quat
      quats[(tt*4):(tt*4)+3] = work_quat
      avvs[(tt*3):(tt*3)+2] = av
   ENDIF ELSE PRINT,'C matrix not retrieved from kernel at index ',tt
ENDFOR
numint = ntime/2
starts = DBLARR(numint)
FOR jj = 0, numint-1 DO starts[jj] = tickvec[2*jj]

;; Write all the quaternions, no angular velocity information, to a
;; single segment C kernel.

CSPICE_CKW03, ck_handle, tick_start, tick_end, sc_id, 'J2000', 1B, $
              "Type 1 C Kernel Segment", tickvec, quats, avvs, starts

CSPICE_CKCLS, ck_handle

WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = 'C-smithed C kernel written.'
END
