;; File Version: 1.1
;; Last Modified: June 7, 2007
;; By: JEC

PRO GEOMETER_CUBE_TRANSFORM, version, cube, geomcube

;; This program takes either of the data structures from Kris Larsen's
;; cube_generator program and extracts the data needed by GEOMETER and
;; stores in a new common format so that GEOMETER drawing routines don't
;; need to worry about handling two formats of the data cubes.
;; June 7, 2007: add lat/lon information to geomcubestruct.

COMPILE_OPT idl2

CASE version OF
    1: nrecord = N_ELEMENTS(cube.et[0,*])
    2: nrecord = N_ELEMENTS(cube.et)
    ELSE: STOP,'Invalid cube version number sent to geometer_cube_transform.'
ENDCASE

geomcubestruct = {data: DBLARR(64, 1024)-1D0, $
                  ra_full: DBLARR(4, 64), $
                  dec_full: DBLARR(4, 64), $
                  lat_full: DBLARR(4, 64), $
                  lon_full: DBLARR(4, 64), $
                  rrad_full: DBLARR(4, 64), $
                  et: 0D0, $
                 sub_solar_longitude: -1000D0, $
                 sub_solar_latitude: -1000D0, $
                 sub_spacecraft_longitude: -1000D0, $
                 sub_spacecraft_latitude: -1000D0}
                  
geomcube = REPLICATE(geomcubestruct, nrecord)

CASE version OF
    1: BEGIN
        nrow = N_ELEMENTS(cube.uvis[*,0,0])
        ncol = N_ELEMENTS(cube.uvis[0,0,*])
        FOR rec = 0, nrecord-1 DO BEGIN
            FOR px = 0, nrow-1 DO BEGIN
                geomcube[rec].ra_full[0,px] = cube.pixel_lower_left_ra[px,rec]
                geomcube[rec].ra_full[1,px] = cube.pixel_upper_left_ra[px,rec]
                geomcube[rec].ra_full[2,px] = cube.pixel_upper_right_ra[px,rec]
                geomcube[rec].ra_full[3,px] = cube.pixel_lower_right_ra[px,rec]
                geomcube[rec].dec_full[0,px] = $
                  cube.pixel_lower_left_declination[px,rec]
                geomcube[rec].dec_full[1,px] = $
                  cube.pixel_upper_left_declination[px,rec]
                geomcube[rec].dec_full[2,px] = $
                  cube.pixel_upper_right_declination[px,rec]
                geomcube[rec].dec_full[3,px] = $
                  cube.pixel_lower_right_declination[px,rec]

                geomcube[rec].lon_full[0,px] = $
                   cube.pixel_lower_left_longitude[px,rec]
                geomcube[rec].lon_full[1,px] = $
                   cube.pixel_upper_left_longitude[px,rec]
                geomcube[rec].lon_full[2,px] = $
                   cube.pixel_upper_right_longitude[px,rec]
                geomcube[rec].lon_full[3,px] = $
                   cube.pixel_lower_right_longitude[px,rec]
                geomcube[rec].lat_full[0,px] = $
                  cube.pixel_lower_left_latitude[px,rec]
                geomcube[rec].lat_full[1,px] = $
                  cube.pixel_upper_left_latitude[px,rec]
                geomcube[rec].lat_full[2,px] = $
                  cube.pixel_upper_right_latitude[px,rec]
                geomcube[rec].lat_full[3,px] = $
                  cube.pixel_lower_right_latitude[px,rec]

                geomcube[rec].rrad_full[0,px] = $
                  cube.lower_left_ring_plane_radii[px,rec]
                geomcube[rec].rrad_full[1,px] = $
                  cube.upper_left_ring_plane_radii[px,rec]
                geomcube[rec].rrad_full[2,px] = $
                  cube.upper_right_ring_plane_radii[px,rec]
                geomcube[rec].rrad_full[3,px] = $
                  cube.lower_right_ring_plane_radii[px,rec]

                FOR sp = 0, ncol-1 DO $
                    geomcube[rec].data[px, sp] = cube.uvis[px, rec, sp]
            ENDFOR
            geomcube[rec].et = cube.et[0,rec]
            geomcube[rec].sub_solar_longitude = $
               cube.sub_solar_longitude[0,rec]
            geomcube[rec].sub_solar_latitude = $
               cube.sub_solar_latitude[0,rec]
            geomcube[rec].sub_spacecraft_longitude = $
               cube.sub_spacecraft_longitude[0,rec]
            geomcube[rec].sub_spacecraft_latitude = $
               cube.sub_spacecraft_latitude[0,rec]
        ENDFOR
    END

    2: BEGIN
        FOR rec = 0, nrecord-1 DO BEGIN
            FOR px = 0, 63 DO BEGIN
                geomcube[rec].ra_full[0,px] = cube[rec].pixel_lower_left_ra[px]
                geomcube[rec].ra_full[1,px] = cube[rec].pixel_upper_left_ra[px]
                geomcube[rec].ra_full[2,px] = cube[rec].pixel_upper_right_ra[px]
                geomcube[rec].ra_full[3,px] = cube[rec].pixel_lower_right_ra[px]
                geomcube[rec].dec_full[0,px] = $
                  cube[rec].pixel_lower_left_declination[px]
                geomcube[rec].dec_full[1,px] = $
                  cube[rec].pixel_upper_left_declination[px]
                geomcube[rec].dec_full[2,px] = $
                  cube[rec].pixel_upper_right_declination[px]
                geomcube[rec].dec_full[3,px] = $
                  cube[rec].pixel_lower_right_declination[px]

                geomcube[rec].lon_full[0,px] = $
                   cube[rec].pixel_lower_left_longitude[px]
                geomcube[rec].lon_full[1,px] = $
                   cube[rec].pixel_upper_left_longitude[px]
                geomcube[rec].lon_full[2,px] = $
                   cube[rec].pixel_upper_right_longitude[px]
                geomcube[rec].lon_full[3,px] = $
                   cube[rec].pixel_lower_right_longitude[px]
                geomcube[rec].lat_full[0,px] = $
                  cube[rec].pixel_lower_left_latitude[px]
                geomcube[rec].lat_full[1,px] = $
                  cube[rec].pixel_upper_left_latitude[px]
                geomcube[rec].lat_full[2,px] = $
                  cube[rec].pixel_upper_right_latitude[px]
                geomcube[rec].lat_full[3,px] = $
                  cube[rec].pixel_lower_right_latitude[px]

;                geomcube[rec].rrad_full[0,px] = $
;                  cube[rec].lower_left_ring_plane_radii[px]
;                geomcube[rec].rrad_full[1,px] = $
;                  cube[rec].upper_left_ring_plane_radii[px]
;                geomcube[rec].rrad_full[2,px] = $
;                  cube[rec].upper_right_ring_plane_radii[px]
;                geomcube[rec].rrad_full[3,px] = $
;                  cube[rec].lower_right_ring_plane_radii[px]

                FOR sp = 0, 1023 DO $
                  geomcube[rec].data[px, sp] = cube[rec].uvis[px, sp]
            ENDFOR
            geomcube[rec].et = cube[rec].et
            geomcube[rec].sub_solar_longitude = $
               cube[rec].sub_solar_longitude[0]
            geomcube[rec].sub_solar_latitude = $
               cube[rec].sub_solar_latitude[0]
            geomcube[rec].sub_spacecraft_longitude = $
               cube[rec].sub_spacecraft_longitude[0]
            geomcube[rec].sub_spacecraft_latitude = $
               cube[rec].sub_spacecraft_latitude[0]
        ENDFOR
    END
ENDCASE

END
