@ringmap_panel.pro
@ringmap_panel_event.pro
@ringmap_panel_callback.pro
@ringmap_tool.pro
@ringmap_tool__define.pro

;;File Version: 1.01
;;Last Modified: October 27, 2004
;;By: JEC

PRO GEOMETER_RINGMAP, data, maininfo, i_time, UPDATE = UPDATE

;;  >> Abstract <<
;;  Only to be called by GEOMETER.

;;  i_time is the index of the desired time for computing body data.
;;  maininfo is the UVALUE structure from GEOMETER.
;;  Josh Colwell

;; Create the Ringmap iTool with its customized panel:
IF NOT KEYWORD_SET(UPDATE) THEN RINGMAP_TOOL

;; Get the tool's object reference for sending text labels to the vis layer:
idtool = ITGETCURRENT(TOOL=oTool)

;; Change the current_zoom property of the view so that we don't have
;; so much white space around it:
view = idtool+'/WINDOW/VIEW_1'
IF NOT KEYWORD_SET(UPDATE) THEN BEGIN
    void = oTool->doSetProperty(view, 'CURRENT_ZOOM', 1.5)
    oTool->setproperty, PTR_NEW(data), 'DATA'
    oTool->setproperty, PTR_NEW(maininfo), 'MAININFO'
    oTool->setproperty, PTR_NEW(i_time), 'ITIME'
ENDIF
oTool->commitactions

npixel = N_ELEMENTS(data[i_time].ra_full[0,*])

;; Indicate to the user which end of the slit is which by drawing the lowest 
;; numbered pixel filled in green:
selected_pixels = WHERE(maininfo.pxvec)
name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)

;; Store the object IDs of all plotted pixels.
pixel_ids = ['']

IPLOT, [data[i_time].rrad_full[1:*,0],data[i_time].rrad_full[1,0]], $
       [data[i_time].lon_full[1:*,0],data[i_time].lon_full[1,0]]*!DTOR, $
       /POLAR, NAME = name, /FILL_BACKGROUND, FILL_COLOR = [0,255,0], $
       TRANSPARENCY = 50,  IDENTIFIER = ringmap_id, /OVERPLOT
pixel_ids = [pixel_ids, (otool->getselecteditems())->getfullidentifier()]

;; Store the object IDs of all labels in the visualization layer.
label_ids = ['']

oTextp = OBJ_NEW('IDLitVisText')
oTool->Add, oTextp
gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
text_x = data[i_time].rrad_full[1,0]*COS(data[i_time].lon_full[1,0]*!DTOR)
text_y = data[i_time].rrad_full[1,0]*SIN(data[i_time].lon_full[1,0]*!DTOR)
gTextp->SetProperty, STRING = name, LOCATION = [text_x, text_y]

;; Store the ID for the pixel labels too so they toggle with the pixels.
pixel_ids = [pixel_ids, gtextp->getfullidentifier()]
label_ids = [label_ids, gtextp->getfullidentifier()]

;; Get the object reference for the data space so we can control the
;; plotting range.
dataspaceid = 'WINDOW/VIEW_1/VISUALIZATION LAYER/DATA SPACE'
oDataSpace = oTool->getbyidentifier(dataspaceid)
oDataSpace->setproperty, x_auto_update= 0B
oDataSpace->setProperty, y_auto_update= 0B

IF npixel GT 1 THEN BEGIN
    label_flag = 0
    FOR px_index = 1, npixel-1 DO BEGIN
        name = 'Pixel '+STRTRIM(STRING(selected_pixels[px_index]),2)
        IPLOT, [data[i_time].rrad_full[1:*,px_index], $
                data[i_time].rrad_full[1,px_index]], $
               [data[i_time].lon_full[1:*,px_index], $
                data[i_time].lon_full[1,px_index]]*!DTOR, /OVERPLOT, $
               NAME = name, /POLAR
        pixel_ids = [pixel_ids, (otool->getselecteditems())->$
                     getfullidentifier()]

        IF label_flag THEN BEGIN
            oTextp = OBJ_NEW('IDLitVisText')
            oTool->Add, oTextp
            gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
            text_x = data[i_time].rrad_full[1,px_index]*$
                     COS(data[i_time].lon_full[1,px_index]*!DTOR)
            text_y = data[i_time].rrad_full[1,px_index]*$
                     SIN(data[i_time].lon_full[1,px_index]*!DTOR)
            gTextp->SetProperty, STRING = name, LOCATION = [text_x, text_y]
            
            label_flag = 0
            pixel_ids = [pixel_ids, gtextp->getfullidentifier()]
            label_ids = [label_ids, gtextp->getfullidentifier()]
        ENDIF ELSE label_flag = 1

    ENDFOR

ENDIF
pixel_ids = pixel_ids[1:*]
otool->setproperty, PTR_NEW(pixel_ids), 'PIXELS'

;; That's it for labels on the vis layer, so set the label_ids property.
label_ids = label_ids[1:*]
otool->setproperty, PTR_NEW(label_ids), 'VISLABELS'

;;*****************************************************************************
;;
;;*****************************************************************************

;; Set the data space plot range.
ringradlimit = 1.5E5
oDataSpace->setProperty, x_minimum = -ringradlimit
oDataSpace->setProperty, x_maximum = ringradlimit
oDataSpace->setProperty, y_minimum = -ringradlimit
oDataSpace->setProperty, y_maximum = ringradlimit

;; Here it would be good to compute the location of the planet's
;; shadow on the rings. 

;;*************************************************************************
;;                              RINGS
;;*************************************************************************
IF NOT KEYWORD_SET(UPDATE) THEN BEGIN
;; Plot the critical ring boundaries.
;; Would be better eventually to read these values in from PCK.
CSPICE_BODVAR, 699L, 'RADII', satrad

ringrad = [satrad[0], 66970.,74510.,92000.,117580.,122170.,133405.,$
           133730.,136780.,140180.,140270.]
nring = N_ELEMENTS(ringrad)    

nringpoints = 72
;;Use 72 points for each ring

ringradvec = DBLARR(nringpoints)
ringlon = FINDGEN(nringpoints)/(nringpoints-1D0)*!DPI*2D0

ringids = ['']
FOR ringindex = 0, nring-1 DO BEGIN
    ringradvec[*] = ringrad[ringindex]
    IPLOT, ringradvec, ringlon, /POLAR, /OVERPLOT, $
           NAME = 'Ring Edge'+STRING(ringindex)
    ringids = [ringids, (otool->getselecteditems())->getfullidentifier()]
ENDFOR  ;;ringindex
ringids = ringids[1:*]
otool->setproperty, PTR_NEW(ringids), 'RINGS'

;;*************************************************************************
;;                                    GRID
;;*************************************************************************
;; Now draw some radial grid lines at constant longitude.
lon_inc = 30.

grid_ids = ['']
FOR longrid = 0, 360.-lon_inc, lon_inc DO BEGIN
    radcoords = [MIN(ringrad), MAX(ringrad)]
    loncoords = [longrid, longrid]*!DTOR
    IPLOT, radcoords, loncoords, /POLAR, /OVERPLOT, NAME = 'Radial Lines'
    rad_id = (otool->getselecteditems())->getfullidentifier()
    grid_ids = [grid_ids, rad_id]
ENDFOR

grid_ids = grid_ids[1:*]
otool->setproperty, PTR_NEW(grid_ids), 'GRID'
ENDIF ;;UPDATE keyword

END
