;;File Version: 3.00
;;Last Modified: August 23, 2006
;;Modified by: JEC

;;++
;; Program to allow interactive plotting and viewing of C kernel and SP 
;; kernel geometry.
;; January 8, 2004: Begin GUI definition.
;; Josh Colwell, LASP, University of Colorado.
;; January 14, 2004: Successfully calls GEOMETER_ENGINE for all times 
;; and pixels. Verifying output. Spot checks of RA/Dec and Lat/Lon of 
;; boresight against SOA give consistent results to a visual check of 
;; graphing (longitudes are opposite in sign and may be in different 
;; coordinates frames). Still need to verify precise pixel coordinates 
;; and do full quantitative comparison with CASPER and SOA.
;; January 21, 2004: RA/Dec positions continue to check for boresight, but 
;; pixel positions are off, and body coordinates are not matching CASPER. 
;; Upgraded iPlot capabilities and removed draw widget and standard plot 
;; options.
;; January 27, 2004: Passes 0th order checks of ring plane radius checks.
;; January 29, 2004: Add star option.
;; February 3, 2004: Add user control of light aberration correction.
;; February 14, 2004: Cosmetic changes.
;; February 25, 2004: Add losdist. Add auto axis titling of iPlot.
;; March 17, 2004: Add switch to compute data for corners of pixels as 
;; well as centers, plus new output options for corner data.
;; March 23, 2004: Add EUV Solar Occ port, several new targets, fix some 
;; typo bugs for some slit selections, and improve target handling to 
;; allow easy expansion. Fixed handling of preferences.
;; April 23, 2004: Correct bug in plotting 'Target Body Parameter.Altitude'
;; July 12, 2004: Allow X axis Target Body Parameter selections. Upgrade 
;; SKYMAP to draw Pixel boundaries and target body limb. 
;; July 15, 2004: Add crude body map routine.
;; July 16, 2004: Add C kernel error display.
;; July 27, 2004: Add star magnitude limit control for skymap.
;; August 26, 2004: Fix bug in selected text output option.
;; September 16, 2004: Changed main uvalue from state vector to pointer
;; to a state vector.
;; September 23, 2004: Use new customized skymap iTool.
;; October 22, 2004: Use new star catalog, compute star position corrected
;; for proper motion and parallax.
;; November 17, 2004: Release Version 2.1. Updates to BodyMap and Pixel 
;; selection widget.
;; Version 2.2: Incorporate Target View selection.
;; Version 2.21: Confirm when quitting program. Cosmetic changes.
;; Version 2.3: Enable selection of Moon for relative longitude calculation.
;; Version 2.4: Save comments with IDL save file. Include file path
;; preferences in user preference file. Automatically load default
;; preferences file. 
;; Version 2.41: Replace single precision !RADEG and !DTOR with double
;; precision values. Get ringoccphi value from geometer_engine. Allow 
;; regular plotting in addition to iPlotting.
;; Version 2.42: Add moons and "F Ring" to list of targets. Include f_peri
;; output from GEOMETER_ENGINE.
;; Version 2.43: Fix bug in selected text file output. File was not being
;; properly closed, and single pixel branch not being used in some cases.
;; Version 3.00: Allow direct graphics mapping in addition to iTools. Read
;; image cube data from cube_generator.pro, including data backplanes.
;; Fill in pixels with colors corresponding to data values using cube data.
;; July 29, 2006: Direct Sky Map working with animation.
;; August 15, 2006: Direct Target working. Allow kernel unload.
;;-
@cw_pdmenu2.pro
@cw_text_input.pro
@geom_pixel_widget.pro
@geometer_engine.pro
@body_limb.pro
@call_skymap.pro
@direct_skymap.pro
@call_target.pro
@direct_target.pro
@geometer_skymap.pro
@geometer_bodymap.pro
@geometer_ringmap.pro
@geometer_target.pro
@geometer_convert.pro
@geometer_objdestroy.pro

PRO GEOMETER_EVENT, event

COMMON GEOMETER_COMMON, data, xdata, ydata, radeg, dtor, cube

;; Event handler for events generated by GEOMETER.
WIDGET_CONTROL, event.id, GET_UVALUE = uval
WIDGET_CONTROL, event.top, GET_UVALUE = pstate

CASE uval.type OF
    'BUTTON': BEGIN
        CASE uval.id OF
            'Pixel': BEGIN
                ;;Create a new widget that allows the user to select one or 
                ;;more pixels of the selected field of view.
                GEOM_PIXEL_WIDGET, (*pstate).pxvec, event.top
            END
            
            'Set Plot Pixel': BEGIN
                dummypxvec = DBLARR(65)
                GEOM_PIXEL_WIDGET, dummypxvec, event.top, /OUTPUT
            END
            
            'Pixel Corners': (*pstate).corners = 1
            'Pixel Centers': (*pstate).corners = 0
            
            'sun': BEGIN
                set = WIDGET_INFO(event.id, /BUTTON_SET)
                IF set THEN (*pstate).solar = 1B ELSE (*pstate).solar = 0B
                ;; Set RA/Dec to something other than -1000 so that
                ;; GEOMETER_ENGINE doesn't use C kernel data.
                (*pstate).ra_inert = -500.0
                (*pstate).dec_inert = -500.0
            END

            'star': BEGIN
                set = WIDGET_INFO(event.id, /BUTTON_SET)
                IF set THEN BEGIN
                    starfile = (*pstate).geometerpath+'GeometerStars.txt'
                    IF NOT FILE_TEST(starfile) THEN BEGIN
                        WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                                        'GeometerStars.txt not found.'
                        starfile = DIALOG_PICKFILE(TITLE = $
                                                   'Select GeometerStars.txt')
                    ENDIF
                    WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                                    'Star file located'
                    starbase = WIDGET_BASE(/COLUMN, TITLE = "Enter Star Name")
                    starname = CW_FIELD(starbase, TITLE = "Star Name: ", $
                                        /RETURN_EVENTS)
                    staruval = {starname:starname,$
                                rdtrue:0B, $
                                ra:0D0, $
                                dec:0D0, $
                                mura:0D0, $
                                mudec:0D0, $
                                parallax:0D0}
                    
                    WIDGET_CONTROL, starbase, SET_UVALUE = staruval
                    WIDGET_CONTROL, starbase, /REALIZE, GROUP = $
                                    event.top, /MODAL
                    
                    REPEAT BEGIN
                        GETSTAR, staruval, starname, starfile
                    ENDREP UNTIL staruval.rdtrue
                    
                    IF staruval.ra EQ -1000.0 THEN BEGIN
                        ;; GETSTAR couldn't find the star in the star
                        ;; file, so write a message and return.
                        WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                                        'Entered star name not found in file.'
                        WIDGET_CONTROL, starbase, /DESTROY
                        RETURN
                    ENDIF

                    ;; Correct the star position for proper motion and for
                    ;; parallax.
                    ;; Get number of years since 1991.25 epoch of Hipparcos
                    ;; star catalog position:
                    time_since_epoch = (*pstate).etstart/3.155D7+8.75
                    
;PRINT,format='("RA initial = ",f24.16)',staruval.ra
;print,format='("dec initial = ",f24.16)',staruval.dec

                    decstar = staruval.dec+staruval.mudec*$
                              time_since_epoch/3.6D6
                    rastar = staruval.ra+COS(staruval.dec*dtor)*$
                             staruval.mura*time_since_epoch/3.6D6
                    
;PRINT,format='("RA at epoch = ",f24.16)',rastar
;print,format='("dec at epoch = ",f24.16)',decstar

                    dist_to_star = 1D0/(staruval.parallax*1D-3)*3.086D13
                    CSPICE_RADREC, dist_to_star, rastar*dtor, decstar*dtor, $
                                   starpos
                    
                    CSPICE_SPKEZR, 'CASSINI', (*pstate).etstart, 'J2000', $
                                   'NONE', 'SSB', scstate, lt_corr
                    scpos = scstate[0:2]
                    starpos -=scpos
                    Starunit = starpos/CSPICE_VNORM(starpos)

                    CSPICE_RECRAD, starunit, unitdist, rastar,decstar
;print,'star position corrected for proper motion and parallax, but not'
;print,'for aberration:'
;print,'ra = '
;PRINT,rastar*radeg,format = '(f24.16)'
;print,'dec = '
;PRINT,decstar*radeg,format = '(f24.16)'
                    (*pstate).ra_inert = rastar*radeg
                    (*pstate).dec_inert = decstar*radeg
                    WIDGET_CONTROL, starbase, /DESTROY
                    WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                                    'Star Position Computed. RA = '+$
                                    STRING((*pstate).ra_inert)+' Dec = '+$
                                    STRING((*pstate).dec_inert)
                ENDIF
                ;; When doing star observations want aberration correction
                ;; in SPICE calls made by GEOMETER_ENGINE to be "NONE".
                ;; Set that, and inform the user.
                (*pstate).aberr = 'NONE'
                WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                                'Aberration correction has been set to NONE.'
                WIDGET_CONTROL, (*pstate).aberrnone, /SET_BUTTON 
            END
            'ck': BEGIN
                set = WIDGET_INFO(event.id, /BUTTON_SET)
                IF set THEN BEGIN
                    (*pstate).ra_inert = -1000D0
                    (*pstate).dec_inert = -1000D0
                ENDIF
            END
            'Full Sky': (*pstate).skyzoom = 1
            'Zoomed In': (*pstate).skyzoom = 2
            'None': (*pstate).aberr = 'NONE'
            'LT': (*pstate).aberr = 'LT'
            'LT+S': (*pstate).aberr = 'LT+S'
;;-----------------------------------------------------------------------------
;;CALCULATE BUTTON
;;-----------------------------------------------------------------------------
            'Calculate': BEGIN
                ;;Send the user inputs from this widget to the GEOMETER_ENGINE
                ;;for computation of desired parameters.
                ;;First, make sure we have valid and sufficient inputs.
                IF (*pstate).dt EQ 0D0 OR (*pstate).etend EQ 0D0 OR $
                  (*pstate).etstart EQ 0D0  OR (*pstate).etend LT $
                  (*pstate).etstart THEN BEGIN
                    WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
             'One or more time entries is invalid. Not computing geometry.'
                    RETURN
                ENDIF
                
                WIDGET_CONTROL, (*pstate).ckerrortext, SET_VALUE = 'No Errors'
                
                ;;Get the string values for the selected field of view:
                slit = 'NONE'
                CASE (*pstate).channel OF
                    0:BEGIN
                        instrument = 'CASSINI_UVIS_EUV'
                        slit = 'LO'
                    END
                    1:BEGIN
                        instrument = 'CASSINI_UVIS_EUV'
                        slit = 'HI'
                    END
                    2:BEGIN
                        instrument = 'CASSINI_UVIS_EUV'
                        slit = 'OCC'
                    END
                    3: BEGIN
                        instrument = 'CASSINI_UVIS_SOLAR'
                        slit = 'OCC'
                    END
                    4:BEGIN
                        instrument = 'CASSINI_UVIS_FUV'
                        slit = 'LO'
                    END
                    5:BEGIN
                        instrument = 'CASSINI_UVIS_FUV'
                        slit = 'HI'
                    END
                    6:BEGIN
                        instrument = 'CASSINI_UVIS_FUV'
                        slit = 'OCC'
                    END
                    7:instrument = 'CASSINI_UVIS_HSP'
                    8:instrument = 'CASSINI_UVIS_HDAC'
                    9:instrument = 'NONE'
                ENDCASE
                
                ;;Get the string value for the target body:
                targetid = WIDGET_INFO((*pstate).targetdrop, /DROPLIST_SELECT)
                WIDGET_CONTROL, (*pstate).targetdrop, GET_VALUE = $
                                targetdropvalue
                target = STRUPCASE(targetdropvalue[targetid])
                
                ;;Figure out how many pixels and time steps we are 
                ;;dealing with:
                setpixels = WHERE((*pstate).pxvec)
                npixels = N_ELEMENTS(setpixels) > 1
                ntime = FIX(((*pstate).etend-(*pstate).etstart)/(*pstate).dt)+1
                nstep = ntime*npixels
                ;;Define the data structures that will store the output 
                ;;from GEOMETER_ENGINE:
                datastruc = {ra:DBLARR(npixels), dec:DBLARR(npixels), $
                             phase:DBLARR(npixels),$
                             emission:DBLARR(npixels), $
                             incidence:DBLARR(npixels), lat:DBLARR(npixels),$
                             lon:DBLARR(npixels), subsclon:0D0, $
                             subsclat:0D0, subsolarlat:0D0, $
                             subsolarlon:0D0, ringplanerad:DBLARR(npixels), $
                             scloc:DBLARR(3), $
                             scvel:DBLARR(3), alt:0D0, ratarg:0D0, $
                             dectarg:0D0, phasetarg:0D0, $
                             incidencetarg:0D0, utc:'', et:0D0, $
                             rayheight:DBLARR(npixels), $
                             latocc:DBLARR(npixels), lonocc:DBLARR(npixels), $
                             losdist:DBLARR(npixels), $
                             ra_full:DBLARR(5, npixels), $
                             dec_full:DBLARR(5, npixels), $
                             lat_full:DBLARR(5, npixels), $
                             lon_full:DBLARR(5, npixels), $
                             rrad_full:DBLARR(5, npixels), $
                             inc_full:DBLARR(5, npixels), $
                             ringoccphi:DBLARR(npixels), $
                             f_peri: 0D0 }
                data = REPLICATE(datastruc,ntime)
                
                ;;Loop through selected time range:
                statuscount = 0
;                systime0 = systime(/SECONDS)
                FOR i = 0, ntime-1 DO BEGIN
                    et = (*pstate).etstart+(*pstate).dt*DOUBLE(i)
                    CSPICE_ET2UTC, et, 'C', 3L, time
                    
                    ;;This can take a while, so keep the user 
                    ;;apprised of progress:
                    WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                                    'Computing time step'+$
                                    STRING(i)+' of'+STRING(ntime)
                    
                    ;;Loop through all selected pixels:
                    FOR pxindex = 0, (npixels-1)>0 DO BEGIN
                        px = setpixels[pxindex]>0
                        IF px EQ 0 THEN px = -1 
;;A value of 0 will be interpreted by GEOMETER_ENGINE
;;as "not set", so we need to send a special value for row 0.
                        IF px EQ 64 OR setpixels[0] EQ -1 THEN px = 0 
;;A value of 64 is for the boresight, so instead
;;send a zero which GEOMETER_ENGINE interprets as no pixel. Also, if
;;no pixels have been selected then setpixels will be a one-element
;;vector with a value of -1, so change that to 0 to use the boresight
;;to prevent GEOMETER_ENGINE from treating the -1 as the signal for
;;pixel 0.
                        ra = (*pstate).ra_inert
                        dec = (*pstate).dec_inert 
;;If these values have been entered, GEOMETER_ENGINE
;;will not use C kernel pointing, but will use pointing for these coords.
                        ckerror = 0B

;;Make the call to GEOMETER_ENGINE *************************************
                        GEOMETER_ENGINE, time, instrument, target, phase, $
                                         emission, incidence, $
                                         ra, dec, lat, lon, ringplanerad, $
                                         scloc, scvel, ra_targ, dec_targ, $
                                         lat_subpt, lon_subpt, lat_sunpt, $
                                         lon_sunpt, alt, phase_targ, $
                                         emission_targ,$
                                         incidence_targ, sunelevation, $
                                         earthelevation, sunearth, ckerror, $
                                         rayheight, latocc, lonocc, losdist, $
                                         ra_full, dec_full, lat_full, $
                                         lon_full, rrad_full, sc_wrt_plan, $
                                         state_sun, state_planet, inc_full, $
                                         sc_sun, occpoint_i, ringoccphi, $
                                         f_peri, $
                                         PIXEL = px, ABCORR = (*pstate).aberr,$
                                         CORNERS = (*pstate).corners, $
                                         SLIT = slit, MOON = (*pstate).moon, $
                                         SOLAR = (*pstate).solar
;;**********************************************************************
                        IF ckerror THEN WIDGET_CONTROL, (*pstate).ckerrortext,$
                          SET_VALUE = 'Error at: '+time
                        
                        data[i].ra[pxindex] = ra
                        data[i].dec[pxindex] = dec
                        data[i].phase[pxindex] = phase
                        data[i].emission[pxindex] = emission
                        data[i].incidence[pxindex] = incidence
                        data[i].lat[pxindex] = lat
                        data[i].lon[pxindex] = lon
                        data[i].ringplanerad[pxindex] = ringplanerad
                        data[i].rayheight[pxindex] = rayheight
                        data[i].latocc[pxindex] = latocc
                        data[i].lonocc[pxindex] = lonocc
                        data[i].losdist[pxindex] = losdist
                        data[i].ringoccphi[pxindex] = ringoccphi
                        IF (*pstate).corners THEN BEGIN
;;Must be a more elegant way to do the following, but for some reason
;;data[i].ra_full[0,pxindex] = ra_full does not work.
                            FOR corner_index = 0,4 DO BEGIN
                                data[i].ra_full[corner_index,pxindex] = $
                                  ra_full[corner_index]
                                data[i].dec_full[corner_index,pxindex] = $
                                  dec_full[corner_index]
                                data[i].lat_full[corner_index,pxindex] = $
                                  lat_full[corner_index]
                                data[i].lon_full[corner_index,pxindex] = $
                                  lon_full[corner_index]
                                data[i].rrad_full[corner_index,pxindex] = $
                                  rrad_full[corner_index]
                                data[i].inc_full[corner_index,pxindex] = $
                                  inc_full[corner_index]
                            ENDFOR   ;;loading the pixel corner data
                        ENDIF
                        
                    ENDFOR    ;;looping through selected pixels.
;;These are the same for all pixels, so save a tiny amount of time 
;;by loading them into the structure outside the pixel loop.
                    data[i].scloc = scloc
                    data[i].scvel = scvel
                    data[i].alt = alt
                    data[i].ratarg = ra_targ
                    data[i].dectarg = dec_targ
                    data[i].subsclat = lat_subpt
                    data[i].subsclon = lon_subpt
                    data[i].subsolarlat = lat_sunpt
                    data[i].subsolarlon = lon_sunpt
                    data[i].phasetarg = phase_targ
                    data[i].incidencetarg = incidence_targ
                    CSPICE_ET2UTC, et, (*pstate).tformat, 3L, time
                    data[i].utc = time
                    data[i].et = et
                    data[i].f_peri = f_peri
                    
                ENDFOR    ;;looping through time
                WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                                'Finished computing geometric parameters.'
;PRINT,'Elapsed time: ',systime(/seconds) - systime0
            END    ;;Calculate button
;;-----------------------------------------------------------------------------
        ENDCASE
    END ;;Button block.
    
    'DROP': BEGIN
        CASE uval.id OF
        'targetdrop': WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
          'New target selected.'

        'channeldrop': BEGIN
            ;;If the channel is changed, then all data change, so things
            ;;must be recomputed.
            WIDGET_CONTROL, (*pstate).xaxistext, SET_VALUE = 'Not selected'
            WIDGET_CONTROL, (*pstate).yaxistext, SET_VALUE = 'Not selected'
            data = 0
            xdata = 0
            ydata = 0

            (*pstate).channel = WIDGET_INFO(event.id, /DROPLIST_SELECT)
            WIDGET_CONTROL, (*pstate).pxbutton, /SENSITIVE
            WIDGET_CONTROL, (*pstate).plotpixelbutton, /SENSITIVE
            CASE (*pstate).channel OF
               7: BEGIN
                 ;;HSP
                 WIDGET_CONTROL, (*pstate).pxbutton, SENSITIVE = 0
                 WIDGET_CONTROL, (*pstate).plotpixelbutton, SENSITIVE = 0
                 (*pstate).pxvec[*] = 0
                 (*pstate).pxvec[64] = 1 
;;sets the boresight to be the selected "pixel".
;;The HSP has no pixels, so don't allow user to select them.
               END  ;;HSP selected
               8: BEGIN
                 ;;HDAC
                 WIDGET_CONTROL, (*pstate).pxbutton, SENSITIVE = 0
                 WIDGET_CONTROL, (*pstate).plotpixelbutton, SENSITIVE = 0
                 (*pstate).pxvec[*] = 0
                 (*pstate).pxvec[64] = 1
                 ;;The HDAC has no pixels.
               END  ;;HDAC selected
               ELSE: ;;Nothing.
            ENDCASE    ;;which FOV selected.
        END ;;Channel selection block.

        ELSE:     ;;Nothing.
        ENDCASE   ;;uval.id of droplist widgets.
    END ;;Drop block.

    'MENU': BEGIN
        CASE event.value OF
            'X Axis.Body Coordinates.Latitude': xdata = data.lat
            'X Axis.Body Coordinates.Longitude': xdata = data.lon
            'X Axis.Body Coordinates.Line of sight distance': xdata = $
              data.losdist
            'X Axis.Sky Coordinates.Right Ascension': xdata = data.ra
            'X Axis.Sky Coordinates.Declination': xdata = data.dec
            'Y Axis.Sky Coordinates.Right Ascension': ydata = data.ra
            'Y Axis.Sky Coordinates.Declination': ydata = data.dec
            'X Axis.Time.ET': xdata = data.et
            'X Axis.Time.Day of Year': BEGIN
                ntime = N_ELEMENTS(data.et)
                xdata = FLTARR(ntime)
                FOR timeindex = 0, ntime-1 DO BEGIN
                    CSPICE_ET2UTC, data[timeindex].et, 'ISOD', 3L, utc
                    xdata[timeindex] = FLOAT(STRMID(utc,5,3))+$
                                       FLOAT(STRMID(utc,9,2))/24.+$
                                       FLOAT(STRMID(utc,12,2))/1440.+$
                                       FLOAT(STRMID(utc,15,6))/86400.
                ENDFOR
            END
            'X Axis.Time.Elapsed Seconds': xdata = data.et-data[0].et
            'X Axis.Time.Elapsed Minutes': xdata = (data.et-data[0].et)/60.0
            'X Axis.Time.Elapsed Hours': xdata = (data.et-data[0].et)/3600.0
            'X Axis.Time.Elapsed Days': xdata = (data.et-data[0].et)/86400.0
            'X Axis.Target Body Parameters.Phase Angle (Sub-S/C)': $
              xdata = data.phasetarg
            'X Axis.Target Body Parameters.Incidence Angle (Sub-S/C)': $
              xdata = data.incidencetarg
            'X Axis.Target Body Parameters.Right Ascension': $
              xdata = data.ratarg
            'X Axis.Target Body Parameters.Declination': xdata = data.dectarg
            'X Axis.Target Body Parameters.Sub-solar Latitude': $
              xdata = data.subsolarlat
            'X Axis.Target Body Parameters.Sub-solar Longitude': $
              xdata = data.subsolarlon
            'X Axis.Target Body Parameters.Sub-S/C Latitude': $
              xdata = data.subsclat
            'X Axis.Target Body Parameters.Sub-S/C Longitude': $
              xdata = data.subsclon
            'X Axis.Target Body Parameters.Distance': BEGIN
                ntime = N_ELEMENTS(data.alt)
                xdata = DBLARR(ntime)
                FOR tt = 0, ntime-1 DO xdata[tt] = CSPICE_VNORM(data[tt].scloc)
            END
            'X Axis.Target Body Parameters.Altitude': xdata = data.alt
            'Y Axis.Body Coordinates.Latitude': ydata = data.lat
            'Y Axis.Body Coordinates.Longitude': ydata = data.lon
            'Y Axis.Body Coordinates.Line of sight distance': $
              ydata = data.losdist
            'Y Axis.Body Coordinates.Ring Radius': ydata = data.ringplanerad
            'X Axis.Body Coordinates.Ring Radius': xdata = data.ringplanerad
            'Y Axis.Body Coordinates.Ring Occ Phi': ydata = data.ringoccphi
            'X Axis.Body Coordinates.Ring Occ Phi': xdata = data.ringoccphi
            'Y Axis.Body Coordinates.Phase Angle': ydata = data.phase
            'Y Axis.Body Coordinates.Emission Angle': ydata = data.emission
            'Y Axis.Body Coordinates.Incidence Angle': ydata = data.incidence
            'X Axis.Body Coordinates.Phase Angle': xdata = data.phase
            'X Axis.Body Coordinates.Emission Angle': xdata = data.emission
            'X Axis.Body Coordinates.Incidence Angle': xdata = data.incidence
            'X Axis.Body Coordinates.Occultation Longitude': $
              xdata = data.lonocc
            'X Axis.Body Coordinates.Occultation Latitude': xdata = data.latocc
            'X Axis.Body Coordinates.Occultation Rayheight': $
              xdata = data.rayheight
            'Y Axis.Body Coordinates.Occultation Longitude': $
              ydata = data.lonocc
            'Y Axis.Body Coordinates.Occultation Latitude': ydata = data.latocc
            'Y Axis.Body Coordinates.Occultation Rayheight': $
              ydata = data.rayheight
            'Y Axis.Target Body Parameters.Phase Angle (Sub-S/C)': $
              ydata = data.phasetarg
            'Y Axis.Target Body Parameters.Incidence Angle (Sub-S/C)': $
              ydata = data.incidencetarg
            'Y Axis.Target Body Parameters.Right Ascension': $
              ydata = data.ratarg
            'Y Axis.Target Body Parameters.Declination': ydata = data.dectarg
            'Y Axis.Target Body Parameters.Sub-solar Latitude': $
              ydata = data.subsolarlat
            'Y Axis.Target Body Parameters.Sub-solar Longitude': $
              ydata = data.subsolarlon
            'Y Axis.Target Body Parameters.Sub-S/C Latitude': $
              ydata = data.subsclat
            'Y Axis.Target Body Parameters.Sub-S/C Longitude': $
              ydata = data.subsclon
            'Y Axis.Target Body Parameters.Distance': BEGIN
                ntime = N_ELEMENTS(data.alt)
                ydata = DBLARR(ntime)
                FOR tt = 0, ntime-1 DO ydata[tt] = CSPICE_VNORM(data[tt].scloc)
            END
            'Y Axis.Target Body Parameters.Altitude': ydata = data.alt
            'iPlot.Ydata vs Xdata': BEGIN
                ntime = N_ELEMENTS(data.alt)
                IF N_ELEMENTS(ydata) GT ntime THEN BEGIN
                   selected = WHERE((*pstate).pxvec)
                   junk = MIN(ABS(selected-(*pstate).plotpixel),plotpxindex)
                    yplotdata = REFORM(ydata[plotpxindex,*])
                ENDIF ELSE yplotdata = REFORM(ydata)
                IF N_ELEMENTS(xdata) GT ntime THEN BEGIN
                    selected = WHERE((*pstate).pxvec)
                    junk = MIN(ABS(selected-(*pstate).plotpixel),plotpxindex)
                    xplotdata = REFORM(xdata[plotpxindex,*])
                ENDIF ELSE xplotdata = REFORM(xdata)
                WIDGET_CONTROL, (*pstate).xaxistext, GET_VALUE = xstring
                xstring = xstring[0]
                dotloc = STRPOS(xstring, '.', /REVERSE_SEARCH)
                xtitle = STRMID(xstring, dotloc+1, STRLEN(xstring)-dotloc-1)
                WIDGET_CONTROL, (*pstate).yaxistext, GET_VALUE = ystring
                ystring = ystring[0]
                dotloc = STRPOS(ystring, '.', /REVERSE_SEARCH)
                ytitle = STRMID(ystring, dotloc+1, STRLEN(ystring)-dotloc-1)
                IPLOT, xplotdata, yplotdata, XTITLE = xtitle, YTITLE = ytitle
            END

         'iPlot.Ydata (vs Index)': BEGIN
            ntime = N_ELEMENTS(data.alt)
            IF N_ELEMENTS(ydata) GT ntime THEN BEGIN
             selected = WHERE((*pstate).pxvec)
             junk = MIN(ABS(selected-(*pstate).plotpixel),plotpxindex)
             yplotdata = REFORM(ydata[plotpxindex,*])
            ENDIF ELSE yplotdata = REFORM(ydata)
            WIDGET_CONTROL, (*pstate).yaxistext, GET_VALUE = ystring
            ystring = ystring[0]
            dotloc = STRPOS(ystring, '.', /REVERSE_SEARCH)
            ytitle = STRMID(ystring, dotloc+1, STRLEN(ystring)-dotloc-1)
            IPLOT, yplotdata, YTITLE = ytitle
         END

            'Overplot.Ydata vs Xdata': BEGIN
                ntime = N_ELEMENTS(data.alt)
                IF N_ELEMENTS(ydata) GT ntime THEN BEGIN
                   selected = WHERE((*pstate).pxvec)
                   junk = MIN(ABS(selected-(*pstate).plotpixel),plotpxindex)
                    yplotdata = REFORM(ydata[plotpxindex,*])
                ENDIF ELSE yplotdata = REFORM(ydata)
                IF N_ELEMENTS(xdata) GT ntime THEN BEGIN
                    selected = WHERE((*pstate).pxvec)
                    junk = MIN(ABS(selected-(*pstate).plotpixel),plotpxindex)
                    xplotdata = REFORM(xdata[plotpxindex,*])
                ENDIF ELSE xplotdata = REFORM(xdata)
                IPLOT, xplotdata, yplotdata, /OVERPLOT
            END

         'Overplot.Ydata (vs Index)': BEGIN
            ntime = N_ELEMENTS(data.alt)
            IF N_ELEMENTS(ydata) GT ntime THEN BEGIN
             selected = WHERE((*pstate).pxvec)
             junk = MIN(ABS(selected-(*pstate).plotpixel),plotpxindex)
             yplotdata = REFORM(ydata[plotpxindex,*])
            ENDIF ELSE yplotdata = REFORM(ydata)
            IPLOT, yplotdata, /OVERPLOT
         END

            'Plot.Ydata vs Xdata': BEGIN
                ntime = N_ELEMENTS(data.alt)
                IF N_ELEMENTS(ydata) GT ntime THEN BEGIN
                   selected = WHERE((*pstate).pxvec)
                   junk = MIN(ABS(selected-(*pstate).plotpixel),plotpxindex)
                    yplotdata = REFORM(ydata[plotpxindex,*])
                ENDIF ELSE yplotdata = REFORM(ydata)
                IF N_ELEMENTS(xdata) GT ntime THEN BEGIN
                    selected = WHERE((*pstate).pxvec)
                    junk = MIN(ABS(selected-(*pstate).plotpixel),plotpxindex)
                    xplotdata = REFORM(xdata[plotpxindex,*])
                ENDIF ELSE xplotdata = REFORM(xdata)
                WIDGET_CONTROL, (*pstate).xaxistext, GET_VALUE = xstring
                xstring = xstring[0]
                dotloc = STRPOS(xstring, '.', /REVERSE_SEARCH)
                xtitle = STRMID(xstring, dotloc+1, STRLEN(xstring)-dotloc-1)
                WIDGET_CONTROL, (*pstate).yaxistext, GET_VALUE = ystring
                ystring = ystring[0]
                dotloc = STRPOS(ystring, '.', /REVERSE_SEARCH)
                ytitle = STRMID(ystring, dotloc+1, STRLEN(ystring)-dotloc-1)
                PLOT, xplotdata, yplotdata, XTITLE = xtitle, YTITLE = ytitle
            END

         'Plot.Ydata (vs Index)': BEGIN
            ntime = N_ELEMENTS(data.alt)
            IF N_ELEMENTS(ydata) GT ntime THEN BEGIN
             selected = WHERE((*pstate).pxvec)
             junk = MIN(ABS(selected-(*pstate).plotpixel),plotpxindex)
             yplotdata = REFORM(ydata[plotpxindex,*])
            ENDIF ELSE yplotdata = REFORM(ydata)
            WIDGET_CONTROL, (*pstate).yaxistext, GET_VALUE = ystring
            ystring = ystring[0]
            dotloc = STRPOS(ystring, '.', /REVERSE_SEARCH)
            ytitle = STRMID(ystring, dotloc+1, STRLEN(ystring)-dotloc-1)
            PLOT, yplotdata, YTITLE = ytitle
         END

         'iMaps.Make Body Map': BEGIN
             npoints = N_ELEMENTS(data.lat)
             lat = REFORM(data.lat, npoints)
             lon = REFORM(data.lon, npoints)
             onbody = WHERE(lat NE -1000.)
             IF onbody[0] EQ -1 THEN $
               WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                               'No points intersect target body'
;;Identify the time to make the map from the user entry
             WIDGET_CONTROL, (*pstate).maptimetext, GET_VALUE = maptime
             maptime = maptime[0]
             CSPICE_UTC2et, maptime, mapet
             junk = MIN(ABS(maptime-data.et),etloc)
             GEOMETER_BODYMAP, data, *pstate, 0, /GRID
         END

         'iMaps.Make Ring Map': BEGIN
             WIDGET_CONTROL, (*pstate).maptimetext, GET_VALUE = maptime
             maptime = maptime[0]
             CSPICE_UTC2ET, maptime, mapet
             junk = MIN(ABS(mapet-data.et), etloc)
             CSPICE_ET2UTC, data[etloc].et, (*pstate).tformat, 3L, maptime
             WIDGET_CONTROL, (*pstate).maptimetext, SET_VALUE = maptime
             WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                             'Actual time for map: '+maptime
             GEOMETER_RINGMAP, data, *pstate, etloc
         END

;;***********************************************************************
;;                            SKYMAP
;;***********************************************************************
         'iMaps.Make Sky Map': BEGIN
;;          Identify the time to make the map from the user entry:
            WIDGET_CONTROL, (*pstate).maptimetext, GET_VALUE = maptime
            maptime = maptime[0]
            CSPICE_UTC2ET, maptime, mapet
            junk = MIN(ABS(mapet-data.et),etloc)
            CSPICE_ET2UTC, data[etloc].et, (*pstate).tformat, 3L, maptime
            WIDGET_CONTROL, (*pstate).maptimetext, SET_VALUE = maptime
            WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                'Actual time for map: ' + maptime
;;          Get star limiting magnitude:
            WIDGET_CONTROL, (*pstate).mapstarlimtext, GET_VALUE = starlim
            starlim = FLOAT(starlim[0])
            IF starlim LT -1. THEN $
                GEOMETER_SKYMAP, data, *pstate, etloc $
            ELSE $
                GEOMETER_SKYMAP, data, *pstate, etloc, starlim, /STARS
        END ;;Make Sky Map

        'iMaps.Target View': BEGIN
;;          Identify the time to make the map from the user entry:
            WIDGET_CONTROL, (*pstate).maptimetext, GET_VALUE = maptime
            maptime = maptime[0]
            CSPICE_UTC2ET, maptime, mapet
            junk = MIN(ABS(mapet-data.et),etloc)
            CSPICE_ET2UTC, data[etloc].et, (*pstate).tformat, 3L, maptime
            WIDGET_CONTROL, (*pstate).maptimetext, SET_VALUE = maptime
            WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                'Actual time for map: ' + maptime
;;          Get star limiting magnitude:
            WIDGET_CONTROL, (*pstate).mapstarlimtext, GET_VALUE = starlim
            starlim = FLOAT(starlim[0])
            GEOMETER_TARGET, data, *pstate, etloc, STARS = starlim
        END

        'Direct Maps.Body Map': BEGIN
print,'selected direct body map'
        END

        'Direct Maps.Ring Map': BEGIN
print,'selected direct ring map'
        END

        'Direct Maps.Sky Map': BEGIN
            IF (*pstate).cubeloaded THEN cube_skymap = cube ELSE $
              cube_skymap = {et:[1E20]}

            CALL_SKYMAP, *pstate, data, cube_skymap, event.top
        END

        'Direct Maps.Target': BEGIN
            IF (*pstate).cubeloaded THEN cube_target = cube ELSE $
              cube_target = {et:[1E20]}
            CALL_TARGET, *pstate, data, cube_target, event.top
        END

            'Quit': BEGIN
                confirm = DIALOG_MESSAGE("Are you sure you want to exit?", $
                                         /QUESTION, /DEFAULT_NO)
                IF confirm EQ "No" THEN RETURN
                PTR_FREE, pstate
                WIDGET_CONTROL, event.top, /DESTROY
             RETURN
            END

            'Stop': STOP,'Stopping program. Use .c to continue.'

            'Load SPICE Kernel': BEGIN
            kernelfile = DIALOG_PICKFILE(TITLE = $
                         'Select a SPICE kernel or list to load', PATH = $
                                        (*pstate).kernelpath)
            CSPICE_FURNSH, kernelfile
            WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                'Loaded kernel file: '+kernelfile
            END

            'Load Data Cube': BEGIN
                cubefile = DIALOG_PICKFILE(TITLE = $
                         'Select a data cube file', FILTER = '*.sav', PATH = $
                                        (*pstate).datapath)
            RESTORE, cubefile, /RELAXED_STRUCTURE_ASSIGNMENT
            cube = datastruct
            (*pstate).cubeloaded = 1B
            WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                'Loaded data cube file: '+cubefile
            END

            'Unload SPICE Kernel': BEGIN
                kernelfile = DIALOG_PICKFILE(TITLE = $
                           'Select a SPICE kernel or list to unload', PATH = $
                                             (*pstate).kernelpath)
                CSPICE_UNLOAD, kernelfile
                WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                  'Unloaded kernel file: '+kernelfile
            END

         'Save Pointing Data.IDL Save File': BEGIN
             readme = CW_TEXT_INPUT('Enter a brief description of the data')
             savefile = DIALOG_PICKFILE(TITLE = $
                       'Enter a file name for the IDL Save File', $
                                        PATH = (*pstate).outputpath)
             SAVE, data, readme, FILE = savefile
         END

         'Save Pointing Data.Selected Text File': BEGIN
         savefile = DIALOG_PICKFILE(TITLE = $
                    'Enter a file name for the ASCII file', $
                                   PATH = (*pstate).outputpath)
         ntime = N_ELEMENTS(data.alt)
         npixels = N_ELEMENTS(data[0].ra)
         OPENW, savelun, savefile, /GET_LUN
         WIDGET_CONTROL, (*pstate).xaxistext, GET_VALUE = xval
         WIDGET_CONTROL, (*pstate).yaxistext, GET_VALUE = yval
         PRINTF, savelun, "UTC   ", "ET   ", xval, "  ", yval
         pixels = WHERE((*pstate).pxvec) > 0
         IF ((*pstate).channel LT 7) AND (npixels GT 1) THEN BEGIN
            FOR px = 0, npixels-1 DO BEGIN
            PRINTF, savelun, "PIXEL = "+STRCOMPRESS(STRING(pixels[px]))
            FOR tt = 0, NTIME-1 DO $
              PRINTF, savelun, FORMAT ='(A26, 3F17.4)', data[tt].utc, $
                      data[tt].et, xdata[px,tt], ydata[px,tt]
            ENDFOR
        ENDIF ELSE BEGIN
            xdataout = REFORM(xdata)
            ydataout = REFORM(ydata)
            PRINTF, savelun, "Boresight (or star) information for HSP or HDAC"
            FOR tt = 0, ntime-1 DO $
                PRINTF, savelun, FORMAT = '(A26, 3F17.4)', data[tt].utc, $
                    data[tt].et, xdataout[tt], ydataout[tt]
        ENDELSE
        CLOSE, savelun
        FREE_LUN, savelun
         END

       'Save Pointing Data.RA/Dec Corners': BEGIN
       savefile = DIALOG_PICKFILE(TITLE = $
                  'Enter a file name for the ASCII file', PATH = $
                                 (*pstate).outputfile)
       OPENW, savelun, savefile, /GET_LUN
       ntime = N_ELEMENTS(data.alt)
       npixels = N_ELEMENTS(data[0].ra)
       PRINTF, savelun, "UTC   ", "ET   ", $
               "RA Values (center, 4 corners)   ", $
               "Dec Values (center, 4 corners)"
       pixels = WHERE((*pstate).pxvec) > 0
       FOR px = 0, npixels-1 DO BEGIN
         PRINTF, savelun, "PIXEL = "+STRCOMPRESS(STRING(pixels[px]))
         FOR tt = 0, ntime-1 DO $
          PRINTF, savelun, FORMAT = '(A26, 11F17.4)', data[tt].utc, $
                  data[tt].et, $
                  data[tt].ra_full[*,px], data[tt].dec_full[*,px]
       ENDFOR
       FREE_LUN, savelun
       END

       'Save Pointing Data.Lat/Lon Corners': BEGIN
       savefile = DIALOG_PICKFILE(TITLE = $
                                  'Enter a file name for the ASCII file', $
                                 PATH = (*pstate).outputpath)
       OPENW, savelun, savefile, /GET_LUN
       ntime = N_ELEMENTS(data.alt)
       npixels = N_ELEMENTS(data[0].ra)
       PRINTF, savelun, "UTC   ", "ET   ", $
               "Lat Values (center, 4 corners)   ", $
               "Lon Values (center, 4 corners)"
       pixels = WHERE((*pstate).pxvec) > 0
       FOR px = 0, npixels-1 DO BEGIN
         PRINTF, savelun, "PIXEL = "+STRCOMPRESS(STRING(pixels[px]))
         FOR tt = 0, ntime-1 DO $
          PRINTF, savelun, FORMAT = '(A26, 11F17.4)', data[tt].utc, $
                  data[tt].et, $
                  data[tt].lat_full[*,px], data[tt].lon_full[*,px]
       ENDFOR
       FREE_LUN, savelun
       END

       'Save Pointing Data.Ring Radius/Lon Corners': BEGIN
       savefile = DIALOG_PICKFILE(TITLE = $
                                  'Enter a file name for the ASCII file', $
                                 PATH = (*pstate).outputpath)
       OPENW, savelun, savefile, /GET_LUN
       ntime = N_ELEMENTS(data.alt)
       npixels = N_ELEMENTS(data[0].ra)
       PRINTF, savelun, "UTC   ", "ET   ", $
               "Ring Plane Radius Values (center, 4 corners)   ", $
               "Longitude Values (center, 4 corners)"
       pixels = WHERE((*pstate).pxvec) > 0
       FOR px = 0, npixels-1 DO BEGIN
         PRINTF, savelun, "PIXEL = "+STRCOMPRESS(STRING(pixels[px]))
         FOR tt = 0, ntime-1 DO $
          PRINTF, savelun, FORMAT = '(A26, 11F17.4)', data[tt].utc, $
           data[tt].et, data[tt].rrad_full[*,px], data[tt].lon_full[*,px]
       ENDFOR
       FREE_LUN, savelun
       END

       'Save Pointing Data.Full Text File': BEGIN
           savefile = DIALOG_PICKFILE(TITLE = $
                                    'Enter a file name for the ASCII file', $
                                   PATH = (*pstate).outputpath)
           ntime = N_ELEMENTS(data.alt)
           npixels = N_ELEMENTS(data[0].ra)
           OPENW, savelun, savefile, /GET_LUN
           PRINTF, savelun, FORMAT = '("Pixel  ", "UTC Time  ", "ET  ", "Target RA  ", "Target Dec  ", "RA  ", "Dec  ", "Lon  ", "Lat  ", "Phase  ", "Emission  ", "Incidence  ", "Ring Plane Radius  ", "Occultation Ray Height")'
           pixels = WHERE((*pstate).pxvec)
           FOR tt = 0, NTIME-1 DO BEGIN
               FOR px = 0, npixels-1 DO $
                 PRINTF, savelun, FORMAT = '(A4, A26, 37F17.4)', $
                         STRCOMPRESS(STRING(pixels[px])), data[tt].utc, $
                         data[tt].et, data[tt].ratarg, data[tt].dectarg, $
                         data[tt].ra[px], data[tt].dec[px], data[tt].lon[px], $
                         data[tt].lat[px], data[tt].phase[px], $
                         data[tt].emission[px], data[tt].incidence[px], $
                         data[tt].ringplanerad[px], $
                         data[tt].rayheight[px], data[tt].ra_full[*,px], $
                         data[tt].dec_full[*,px], data[tt].lat_full[*,px], $
                         data[tt].lon_full[*,px], data[tt].rrad_full[*,px]
           ENDFOR
           FREE_LUN, savelun
       END

        'Save Body Data.IDL Save File': BEGIN
            readme = CW_TEXT_INPUT('Enter a brief description of the data')
            savefile = DIALOG_PICKFILE(TITLE = $
                                   'Enter a file name for the IDL save file',$
                                      PATH = (*pstate).outputpath)
            ;;Get the string value for the target body:
            targetid = WIDGET_INFO((*pstate).targetdrop, /DROPLIST_SELECT)
            WIDGET_CONTROL, (*pstate).targetdrop, GET_VALUE = targetdropvalue
            target = STRUPCASE(targetdropvalue[targetid])
            BODY_LIMB, target, 'CASSINI', data[0].et, 90, ra_limb, dec_limb
            SAVE, ra_limb, dec_limb, readme, FILENAME = savefile
        END

       'About Geometer': XDISPLAYFILE, $
         (*pstate).geometerpath+'aboutgeometer.txt'
       'Help': XDISPLAYFILE, (*pstate).geometerpath+'geometerhelp.txt'

       'Time Format.C': (*pstate).tformat = 'C'
       'Time Format.D': (*pstate).tformat = 'D'
       'Time Format.J': (*pstate).tformat = 'J'
       'Time Format.ISOC': (*pstate).tformat = 'ISOC'
       'Time Format.ISOD': (*pstate).tformat = 'ISOD'

       'File Paths.Kernels': BEGIN
           junk = DIALOG_PICKFILE(TITLE = $
                                  'Select any file in your kernel directory', $
                                  GET_PATH = path)
           (*pstate).kernelpath = path
           WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                           'Selected directory: '+path
       END
       'File Paths.GEOMETER Output': BEGIN
           junk = DIALOG_PICKFILE(TITLE = $
                                  'Select any file in your output directory', $
                                  GET_PATH = path)
           (*pstate).outputpath = path
           WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                           'Selected directory: '+path
       END

       'Longitudes.Default': (*pstate).moon = 0
       'Longitudes.Relative to Moon': BEGIN
;; Create a little widget and check the entry until we get one that 
;; is a moon with data in the kernel pool.

           moonbase = WIDGET_BASE(/COLUMN, TITLE = "Enter Moon Name")
           moonname = CW_FIELD(moonbase, TITLE = "Moon Name: ", $
                               /RETURN_EVENTS)
                    
           WIDGET_CONTROL, moonbase, /REALIZE, GROUP = $
                           event.top, /MODAL
           
           REPEAT GETMOON, moonname, valid, code UNTIL valid
           WIDGET_CONTROL, moonbase, /DESTROY
           IF moonname NE '0' THEN BEGIN
               CATCH, /CANCEL ;;Cancel any previous error handlers.
               CATCH, spk_error_status ;;establish a new error handler.
               IF STRMID(!ERROR_STATE.MSG,0,13) EQ 'CSPICE_SPKPOS' THEN BEGIN
                   ;;Don't have data for Moon. Write a message and
                   ;;don't use Moon.
                   CATCH, /CANCEL
                   MESSAGE, /RESET_ERROR_STATE
                   WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                                   'No SPK data for this moon is loaded. '+$
                                   'Not using this moon for longitudes.'
                   code = 0
               ENDIF ELSE WIDGET_CONTROL, (*pstate).statustext, $
                 SET_VALUE = 'Moon selected for longitude: '+moonname

               CSPICE_SPKPOS, 'SATURN', (*pstate).etstart, 'J2000', 'NONE', $
                              moonname, moonstate, light_time
               (*pstate).moon = code
           ENDIF
       END

       'Save Prefs': BEGIN
         sf = DIALOG_PICKFILE(TITLE = 'Enter preferences save file name', $
                             PATH = (*pstate).geometerpath)
         tformat = (*pstate).tformat
         channel = (*pstate).channel
         pxvec = (*pstate).pxvec
         kernelpath = (*pstate).kernelpath
         outputpath = (*pstate).outputpath
         datapath = (*pstate).datapath
         targetid = WIDGET_INFO((*pstate).targetdrop, /DROPLIST_SELECT)
         SAVE, tformat, channel, targetid, pxvec, kernelpath, $
               outputpath, datapath, FILENAME = sf
         WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                         'Preference file saved.'
       END  ;;Save Preferences

       'Load Prefs': BEGIN
         sf = DIALOG_PICKFILE(TITLE = 'Select Geometer preferences file', $
                             PATH = (*pstate).geometerpath)
         RESTORE, sf
         (*pstate).tformat = tformat
         (*pstate).channel = channel
         (*pstate).pxvec = pxvec
         (*pstate).kernelpath = kernelpath
         (*pstate).outputpath = outputpath
         WIDGET_CONTROL, (*pstate).targetdrop, $
                         SET_DROPLIST_SELECT = targetid
         WIDGET_CONTROL, (*pstate).channeldrop, $
                         SET_DROPLIST_SELECT = (*pstate).channel

           WIDGET_CONTROL, (*pstate).pxbutton, /SENSITIVE
            WIDGET_CONTROL, (*pstate).plotpixelbutton, /SENSITIVE
            CASE (*pstate).channel OF
               7: BEGIN
                 ;;HSP
                 WIDGET_CONTROL, (*pstate).pxbutton, SENSITIVE = 0
                 WIDGET_CONTROL, (*pstate).plotpixelbutton, SENSITIVE = 0
                 (*pstate).pxvec[*] = 0
                 (*pstate).pxvec[64] = 1 
                 ;;sets the boresight to be the selected "pixel".
                 ;;The HSP has no pixels, so don't allow user to select them.
               END  ;;HSP selected
               8: BEGIN
                 ;;HDAC
                 WIDGET_CONTROL, (*pstate).pxbutton, SENSITIVE = 0
                 WIDGET_CONTROL, (*pstate).plotpixelbutton, SENSITIVE = 0
                 (*pstate).pxvec[*] = 0
                 (*pstate).pxvec[64] = 1
                 ;;The HDAC has no pixels.
               END  ;;HDAC selected
               ELSE: ;;Nothing.
            ENDCASE    ;;which FOV selected.

       END  ;;Load Preferences

         ELSE:  ;;do nothing.
       ENDCASE  ;;event.value case.
       IF STRMID(event.value, 0, 6) EQ 'X Axis' THEN $
        WIDGET_CONTROL, (*pstate).xaxistext, SET_VALUE = $
            STRMID(event.value,7,STRLEN(event.value)-7)
       IF STRMID(event.value, 0, 6) EQ 'Y Axis' THEN $
        WIDGET_CONTROL, (*pstate).yaxistext, SET_VALUE = $
        STRMID(event.value, 7, STRLEN(event.value)-7)

    END ;;Menu block.
;;*************************************************************************
    'TEXT':BEGIN
       WIDGET_CONTROL, event.id, GET_VALUE = value

       CASE uval.id OF
         'starttext': BEGIN
          utc = value[0]
          CSPICE_UTC2ET, utc, etstart
          (*pstate).etstart = etstart
          CSPICE_ET2UTC, etstart, (*pstate).tformat, 3L, startutc
          WIDGET_CONTROL, (*pstate).starttext, SET_VALUE = startutc
          WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
            'Entered ET start time: '+STRING(DOUBLE(etstart))
          WIDGET_CONTROL, (*pstate).maptimetext, SET_VALUE = startutc
          WIDGET_CONTROL, (*pstate).endtext, /INPUT_FOCUS

          ;; Compute the number of time steps if possible:
          IF (*pstate).dt GT 0. AND (*pstate).etend NE 0. THEN BEGIN
              nsteps = LONG(((*pstate).etend-(*pstate).etstart)/(*pstate).dt+1)
              WIDGET_CONTROL, (*pstate).nsteptext, SET_VALUE = $
                              STRTRIM(STRING(nsteps),2)
          ENDIF
         END

         'endtext': BEGIN
          utc = value[0]
          CSPICE_UTC2ET, utc, etend
          (*pstate).etend = etend
          CSPICE_ET2UTC, etend, (*pstate).tformat, 3L, endutc
          WIDGET_CONTROL, (*pstate).endtext, SET_VALUE = endutc
          WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
            'Entered ET end time: '+STRING(DOUBLE(etend))
          WIDGET_CONTROL, (*pstate).dttext, /INPUT_FOCUS
          ;; Compute the number of time steps if possible:
          IF (*pstate).dt GT 0. AND (*pstate).etstart NE 0. THEN BEGIN
              nsteps = LONG(((*pstate).etend-(*pstate).etstart)/(*pstate).dt+1)
              WIDGET_CONTROL, (*pstate).nsteptext, SET_VALUE = $
                              STRTRIM(STRING(nsteps),2)
          ENDIF
         END

         'maptime': BEGIN
          utc = value[0]
          CSPICE_UTC2ET, utc, etmap
          IF etmap LT (*pstate).etstart OR etmap GT (*pstate).etend THEN BEGIN
            WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
                'Map time must be between start time and end time.'
            WIDGET_CONTROL, (*pstate).maptimetext, SET_VALUE = $
                            'Time out of bounds'
            RETURN
          ENDIF
          CSPICE_ET2UTC, etmap, (*pstate).tformat, 3L, maputc
          WIDGET_CONTROL, (*pstate).maptimetext, SET_VALUE = maputc
          WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
            'Entered ET map time: '+STRING(DOUBLE(etmap))
         END

         'dttext': BEGIN
         dt = FLOAT(value[0])
         (*pstate).dt = dt
         WIDGET_CONTROL, (*pstate).dttext, SET_VALUE = $
                         STRTRIM(STRING(dt),2)
         ;; Compute the number of time steps if possible:
         IF (*pstate).etstart NE 0. AND (*pstate).etend NE 0. THEN BEGIN
             nsteps = LONG(((*pstate).etend-(*pstate).etstart)/dt+1)
             WIDGET_CONTROL, (*pstate).nsteptext, SET_VALUE = $
                             STRTRIM(STRING(nsteps),2)
         ENDIF

         WIDGET_CONTROL, (*pstate).statustext, SET_VALUE = $
            'Entered time step: '+STRING(DOUBLE(dt))
         WIDGET_CONTROL, (*pstate).starttext, /INPUT_FOCUS
         END

         'starlim': BEGIN
            mag = FLOAT(value[0])
            WIDGET_CONTROL, (*pstate).mapstarlimtext, SET_VALUE = $
                STRTRIM(STRING(mag),2)
         END
         ELSE:  ;;do nothing.
       ENDCASE  ;;uval.id case
    END ;;text block.

    ELSE:
ENDCASE

END



;;-----------------------------------------------------------------------------
;;-----------------------------------------------------------------------------
;;-----------------------------------------------------------------------------
;;-----------------------------------------------------------------------------




PRO GEOMETER

COMMON GEOMETER_COMMON, data, xdata, ydata, radeg, dtor, cube

radeg = 180D0/!DPI
dtor = !DPI/180D0

gefile = FILE_WHICH('geometer_engine.pro')
IF FILE_TEST(gefile) THEN BEGIN
  OPENR, gelun, gefile, /GET_LUN
  line=''
  READF, gelun, line
  FREE_LUN, gelun
  geversion = STRMID(line, 7, STRLEN(line)-7)
  geometerpath = FILE_DIRNAME(gefile, /MARK_DIRECTORY)
ENDIF ELSE BEGIN
    geversion = 'Unknown Version'
    geometerpath = ''
ENDELSE

version = 'Version 3.00 Beta, August 23, 2006'
titletext = 'GEOMETER' + ' ' + version + '  ENGINE '+geversion
mainbase = WIDGET_BASE(TITLE=titletext, /COLUMN, MBAR = menubar)

;; Set up menu bar. Use my modification of CW_PDMENU to enable
;; accelerator keys.
menuuval = {id: 'FILE', type:'MENU'}
filemenu = [ '0\Load SPICE Kernel\CTRL+L', $
             '0\Load Data Cube\CTRL+D',$
             '0\Unload SPICE Kernel\CTRL+U',$
             '5\Save Pointing Data', $
             '0\IDL Save File\CTRL+S', $
             '0\Selected Text File', $
             '0\RA/Dec Corners', $
             '0\Lat/Lon Corners', $
             '0\Ring Radius/Lon Corners', $
             '2\Full Text File', $
             '1\Save Body Data', $
             '2\IDL Save File', $
             '4\Stop\Escape', $
             '2\Quit\CTRL+Q']

filemenubutton = WIDGET_BUTTON(menubar, /MENU, VALUE = 'File')

menu = CW_PDMENU2(filemenubutton, filemenu, /MBAR, DELIMITER = '.', $
    /RETURN_FULL_NAME, UVALUE = menuuval)

plotmenubutton = WIDGET_BUTTON(menubar, /MENU, VALUE = 'Plot')
plotmenu = ['1\iPlot', $
            '0\Ydata vs Xdata\CTRL+X', $
            '2\Ydata (vs Index)\CTRL+Y', $
            '1\Overplot', $
            '0\Ydata vs Xdata', $
            '2\Ydata (vs Index)', $
            '1\Plot', $
            '0\Ydata vs Xdata', $
            '2\Ydata (vs Index)', $
            '1\iMaps', $
            '0\Make Body Map\CTRL+1', $
            '0\Make Ring Map\CTRL+2', $
            '0\Make Sky Map\CTRL+3', $
            '2\Target View\CTRL+T', $
            '1\Direct Maps', $
            '0\Body Map\CTRL+B', $
            '0\Ring Map\CTRL+R', $
            '0\Sky Map\CTRL+K', $
            '2\Target\CTRL+V']

menuuval.id = 'PLOT'
menu = CW_PDMENU2(plotmenubutton, plotmenu, /MBAR, DELIMITER = '.', $
    /RETURN_FULL_NAME, UVALUE = menuuval)

datamenubutton = WIDGET_BUTTON(menubar, /MENU, VALUE = 'Data Selection')
datamenu = ['5\X Axis', $
          '1\Time', $
          '0\ET', $
          '0\Day of Year', $
          '0\Elapsed Seconds', $
          '0\Elapsed Minutes', $
          '0\Elapsed Hours', $
          '2\Elapsed Days', $
          '1\Body Coordinates', $
          '0\Latitude', $
          '0\Longitude', $
          '0\Line of sight distance', $
          '0\Ring Radius', $
          '0\Ring Occ Phi', $
          '0\Phase Angle', $
          '0\Emission Angle', $
          '0\Incidence Angle', $
          '0\Occultation Longitude', $
          '0\Occultation Latitude', $
          '2\Occultation Rayheight', $
          '1\Sky Coordinates', $
          '0\Right Ascension', $
          '2\Declination', $
          '1\Target Body Parameters', $
          '0\Right Ascension', $
          '0\Declination', $
          '0\Phase Angle (Sub-S/C)', $
          '0\Incidence Angle (Sub-S/C)', $
          '0\Sub-solar Latitude', $
          '0\Sub-solar Longitude', $
          '0\Sub-S/C Latitude', $
          '0\Sub-S/C Longitude', $
          '0\Altitude', $
          '2\Distance', $
          '2\', $
         '1\Y Axis', $
          '1\Body Coordinates', $
          '0\Latitude', $
          '0\Longitude', $
          '0\Line of sight distance', $
          '0\Ring Radius', $
          '0\Ring Occ Phi', $
          '0\Phase Angle', $
          '0\Emission Angle', $
          '0\Incidence Angle', $
          '0\Occultation Longitude', $
          '0\Occultation Latitude', $
          '2\Occultation Rayheight', $
          '1\Sky Coordinates', $
          '0\Right Ascension', $
          '2\Declination', $
          '1\Target Body Parameters', $
          '0\Right Ascension', $
          '0\Declination', $
          '0\Phase Angle (Sub-S/C)', $
          '0\Incidence Angle (Sub-S/C)', $
          '0\Sub-solar Latitude', $
          '0\Sub-solar Longitude', $
          '0\Sub-S/C Latitude', $
          '0\Sub-S/C Longitude', $
          '0\Altitude', $
          '2\Distance']

menuuval.id = 'DATASELECT'
menu = CW_PDMENU2(datamenubutton, datamenu, /MBAR, DELIMITER = '.', $
                  /RETURN_FULL_NAME, UVALUE = menuuval)

prefmenubutton = WIDGET_BUTTON(menubar, /MENU, VALUE = 'Preferences')
prefmenu = [ '1\Time Format', $
              '0\C', $
              '0\D', $
              '0\J', $
              '0\ISOC', $
              '2\ISOD', $
             '1\Longitudes', $
             '0\Default', $
             '2\Relative to Moon', $
             '1\File Paths', $
             '0\Kernels', $
             '2\GEOMETER Output', $
          '0\Load Prefs',$
          '2\Save Prefs']
menuuval.id = 'PREFERENCES'
menu = CW_PDMENU2(prefmenubutton, prefmenu, /MBAR, DELIMITER = '.', $
    /RETURN_FULL_NAME, UVALUE = menuuval)

helpmenubutton = WIDGET_BUTTON(menubar, /MENU, VALUE = 'Help')
helpmenu = [ '0\About Geometer', '2\Help\CTRL+H']
menuuval.id = 'HELP'
menu = CW_PDMENU2(helpmenubutton, helpmenu, /MBAR, DELIMITER = '.', $
    /RETURN_FULL_NAME, UVALUE = menuuval)

;;-----------------------------------------------------------------------------
;;Done with menu bar.
;;-----------------------------------------------------------------------------

timerow = WIDGET_BASE(mainbase, /ROW)
label = WIDGET_LABEL(timerow, VALUE = 'Start Time')
tuval = {id:'starttext', type:'TEXT'}
starttext = WIDGET_TEXT(timerow, UVALUE = tuval, /EDITABLE, $
                        VALUE = 'Enter Time')

label = WIDGET_LABEL(timerow, VALUE = 'End Time')
tuval.id = 'endtext'
endtext = WIDGET_TEXT(timerow, UVALUE = tuval, /EDITABLE, $
                      VALUE = 'Enter Time')

label = WIDGET_LABEL(timerow, VALUE = 'Time Step (sec)')
tuval.id = 'dttext'
dttext = WIDGET_TEXT(timerow, UVALUE = tuval, /EDITABLE, $
                     VALUE = 'Enter DT', XSIZE = 9)

label = WIDGET_LABEL(timerow, VALUE = '# Steps')
tuval.id = 'nsteps'
nsteptext = WIDGET_TEXT(timerow, UVALUE = tuval, VALUE = 'N/A', XSIZE = 6)

row2 = WIDGET_BASE(mainbase, /ROW)
label = WIDGET_LABEL(row2, VALUE = 'Channel')
duval = {id:'channeldrop', type:'DROP'}
channeldrop = WIDGET_DROPLIST(row2, UVALUE = duval, VALUE = $
    ['EUV Lo-Res', 'EUV Hi-Res','EUV Occ','EUV Solar Port',$
     'FUV Lo-Res','FUV Hi-Res','FUV Occ', 'HSP', 'HDAC', 'None'])

label = WIDGET_LABEL(row2, VALUE = 'Target')
duval.id = 'targetdrop'
targetdrop = WIDGET_DROPLIST(row2, UVALUE = duval, VALUE = $
    ['Saturn', 'Rings','F Ring', 'Titan', 'Atlas', 'Dione', 'Enceladus', $
     'Epimetheus', 'Helene', 'Hyperion', $
     'Iapetus', 'Janus', 'Mimas', 'Pandora', 'Prometheus', 'Phoebe', $
     'Rhea','Telesto', 'Tethys', 'Jupiter','Io',$
     'Europa','Ganymede','Callisto','Moon','Venus','Sun'])

starbase = WIDGET_BASE(row2, /ROW, /EXCLUSIVE)
buval = {id:'star', type:'BUTTON'}
button = WIDGET_BUTTON(starbase, VALUE = 'Star', UVALUE = buval)
buval.id = 'sun'
button = WIDGET_BUTTON(starbase, VALUE = 'Sun', UVALUE = buval)
buval.id = 'ck'
button = WIDGET_BUTTON(starbase, VALUE = 'CK', UVALUE = buval)
WIDGET_CONTROL, button, /SET_BUTTON

label = WIDGET_LABEL(row2, VALUE = 'CK Errors:')
ckerrortext = WIDGET_TEXT(row2, VALUE = 'No Errors')

row3 = WIDGET_BASE(mainbase, /ROW)
label = WIDGET_LABEL(row3, VALUE = 'X Axis:')
xaxistext = WIDGET_TEXT(row3, VALUE = 'Not selected', XSIZE = 44)
label = WIDGET_LABEL(row3, VALUE = '  Y Axis:')
yaxistext = WIDGET_TEXT(row3, VALUE = 'Not selected', XSIZE = 44)

commandrow = WIDGET_BASE(mainbase, /ROW)
buval.id = 'Calculate'
button = WIDGET_BUTTON(commandrow, VALUE = buval.id, UVALUE = buval)

buval.id = 'Pixel'
pxbutton = WIDGET_BUTTON(commandrow, VALUE = 'Select Pixels', UVALUE = buval)

buval.id = 'Set Plot Pixel'
plotpixelbutton = WIDGET_BUTTON(commandrow, VALUE = buval.id, UVALUE = buval)

aberrrow = WIDGET_BASE(commandrow, /FRAME, /ROW)
label = WIDGET_LABEL(aberrrow, VALUE = "Aberration:")
aberrbase = WIDGET_BASE(aberrrow, /ROW, /EXCLUSIVE)
buval.id = 'None'
aberrnone = WIDGET_BUTTON(aberrbase, VALUE = buval.id, UVALUE = buval)
buval.id = 'LT'
button = WIDGET_BUTTON(aberrbase, VALUE = buval.id, UVALUE = buval)
WIDGET_CONTROL, button, /SET_BUTTON
buval.id = 'LT+S'
button = WIDGET_BUTTON(aberrbase, VALUE = buval.id, UVALUE = buval)

cornerbase = WIDGET_BASE(commandrow, /FRAME, /ROW, /EXCLUSIVE)
buval.id = 'Pixel Corners'
button1 = WIDGET_BUTTON(cornerbase, UVALUE = buval, VALUE = buval.id)
buval.id = 'Pixel Centers'
button = WIDGET_BUTTON(cornerbase, UVALUE = buval, VALUE = buval.id)
WIDGET_CONTROL, button1, /SET_BUTTON

mapbase = WIDGET_BASE(mainbase, /ROW, /FRAME)
label = WIDGET_LABEL(mapbase, VALUE = 'Map Time')
tuval.id = 'maptime'
maptimetext = WIDGET_TEXT(mapbase, VALUE = '', UVALUE = tuval, /EDITABLE)

label = WIDGET_LABEL(mapbase, VALUE = 'Star Limit (MV):')
tuval.id = 'starlim'
mapstarlimtext = WIDGET_TEXT(mapbase, VALUE = '3.5', UVALUE = tuval, $
                             XSIZE = 6, YSIZE = 1, /EDITABLE)

;; Create a zoom control base:
;; Set two initial zoom options for skymap. Send the flag to skymap routine for
;; initial skymap zoom setting.
label = WIDGET_LABEL(mapbase, VALUE = '    Initial Zoom:')
zoombase = WIDGET_BASE(mapbase, /EXCLUSIVE, /ROW)
buval.id = 'Full Sky'
button = WIDGET_BUTTON(zoombase, VALUE = buval.id, UVALUE = buval)
buval.id = 'Zoomed In'
button2 = WIDGET_BUTTON(zoombase, VALUE = buval.id, UVALUE = buval)
WIDGET_CONTROL, button, /SET_BUTTON

statusrow = WIDGET_BASE(mainbase, /ROW)
label = WIDGET_LABEL(statusrow, VALUE = 'Status:')
statustext = WIDGET_TEXT(statusrow, XSIZE = 97)

WIDGET_CONTROL, mainbase, /REALIZE

channel = WIDGET_INFO(channeldrop, /DROPLIST_SELECT)
pxvec = BYTARR(65)  
;;For pixels 0 through 63, a value of 1 indicates the pixel is
;;selected, a value of 0 indicates it is not selected. Index number 64
;;is for the boresight.
;;Create the UVALUE structure for the top level widget. This structure is 
;;used to store the user entries made in the GUI so that they can be 
;;transmitted to the GEOMETER Engine for calculation.
state = {targetdrop:targetdrop, $
         statustext:statustext, $
         etstart:0D0, $
         etend:0D0, $
         dt:0D0, $
         starttext:starttext, $
         endtext:endtext, $
         xaxistext:xaxistext, $
         yaxistext:yaxistext, $
         pxbutton:pxbutton, $
         channel:channel, $
         pxvec:pxvec, $
         dttext:dttext, $
         nsteptext:nsteptext, $
         plotpixel:0, $
         plotpixelbutton:plotpixelbutton, $
         ra_inert:-1000D0, $
         dec_inert:-1000D0, $
         tformat:'ISOD', $
         channeldrop:channeldrop, $
         aberr:'LT', $
         aberrnone:aberrnone, $
         ckerrortext:ckerrortext, $
         corners:1, $
         maptimetext:maptimetext, $
         mapstarlimtext:mapstarlimtext, $
         skyzoom:1, $
         moon:0, $
         solar: 0B, $
         cubeloaded: 0B, $
         kernelpath:'', $
         outputpath:'', $
         datapath:'', $
         geometerpath:geometerpath}

pstate = PTR_NEW(state)
WIDGET_CONTROL, mainbase, SET_UVALUE = pstate

;; Load user preferences file:
pref_file = geometerpath+'geometer_preferences.dat'
IF FILE_TEST(pref_file) THEN BEGIN
    RESTORE, pref_file
    (*pstate).tformat = tformat
    (*pstate).channel = channel
    (*pstate).pxvec = pxvec
    (*pstate).kernelpath = kernelpath
    (*pstate).outputpath = outputpath
    (*pstate).datapath = datapath
    WIDGET_CONTROL, (*pstate).targetdrop, $
                    SET_DROPLIST_SELECT = targetid
    WIDGET_CONTROL, (*pstate).channeldrop, $
                    SET_DROPLIST_SELECT = (*pstate).channel
    
    WIDGET_CONTROL, (*pstate).pxbutton, /SENSITIVE
    WIDGET_CONTROL, (*pstate).plotpixelbutton, /SENSITIVE
    CASE (*pstate).channel OF
        7: BEGIN
            ;;HSP
            WIDGET_CONTROL, (*pstate).pxbutton, SENSITIVE = 0
            WIDGET_CONTROL, (*pstate).plotpixelbutton, SENSITIVE = 0
            (*pstate).pxvec[*] = 0
            (*pstate).pxvec[64] = 1 
            ;;sets the boresight to be the selected "pixel".
            ;;The HSP has no pixels, so don't allow user to select them.
        END  ;;HSP selected
        8: BEGIN
            ;;HDAC
            WIDGET_CONTROL, (*pstate).pxbutton, SENSITIVE = 0
            WIDGET_CONTROL, (*pstate).plotpixelbutton, SENSITIVE = 0
            (*pstate).pxvec[*] = 0
            (*pstate).pxvec[64] = 1
            ;;The HDAC has no pixels.
        END          ;;HDAC selected
        ELSE:        ;;Nothing.
    ENDCASE          ;;which FOV selected.
ENDIF 

WIDGET_CONTROL, mainbase, SET_UVALUE = pstate
XMANAGER, "GEOMETER", mainbase, EVENT_HANDLER = 'GEOMETER_EVENT'
END




PRO GETSTAR, staruval, starname, starfile
ev = WIDGET_EVENT()

CASE ev.id OF
    starname:BEGIN
       WIDGET_CONTROL, ev.id, GET_VALUE = name
       name = name[0]
       OPENR, starlun, starfile, /GET_LUN
       namecheck = ''
       ra = 0D0
       dec = 0D0
       pmra = 0D0
       pmdec = 0D0
       parallax = 0D0
       REPEAT BEGIN
           READF, starlun, namecheck
           READF, starlun, ra
           READF, starlun, dec
           READF, starlun, pmra
           READF, starlun, pmdec
           READF, starlun, parallax
       ENDREP UNTIL STRCMP(namecheck, name, /FOLD_CASE) OR namecheck EQ 'END'
       IF namecheck EQ 'END' THEN BEGIN
           staruval.rdtrue = 1B
           staruval.ra = -1000.0
           RETURN
       ENDIF
       CLOSE, starlun
       FREE_LUN, starlun
       staruval.ra = ra
       staruval.dec = dec
       staruval.mura = pmra
       staruval.mudec = pmdec
       staruval.parallax = parallax
       staruval.rdtrue = 1B
    END
    0: ; do nothing - no event yet.
ENDCASE

END


PRO GETMOON, moonname, valid, code
ev = WIDGET_EVENT()

CASE ev.id OF
    moonname:BEGIN
       WIDGET_CONTROL, ev.id, GET_VALUE = name
       moonname = name[0]
       CATCH, error
       IF error EQ 0 THEN BEGIN
           CSPICE_BODN2C, moonname, code, found
           valid = 1B
       ENDIF ELSE BEGIN
           moonname = '0'
           void = DIALOG_MESSAGE(["Moon data not found in kernel pool.", $
                           "Ignoring moon selection. Reselect",$
                           "from preferences menu to try again."],$
                          /INFORMATION)
           valid = 1B
           code = -1L
       ENDELSE
       CATCH, /CANCEL
       MESSAGE, /RESET_ERROR_STATE
   END
   0:                           ; do nothing - no event yet.
ENDCASE

END
