@skymap_panel.pro
@skymap_panel_event.pro
@skymap_panel_callback.pro
@skymap_tool.pro
@skymap_tool__define.pro

;;File Version: 1.02
;;Last Modified: November 16, 2004
;;By: JEC

PRO GEOMETER_SKYMAP, data, maininfo, i_time, starlim, STARS = STARS, $
                     UPDATE = UPDATE

;;  >> Abstract <<
;;  Only to be called by GEOMETER.

;;  i_time is the index of the desired time for computing body data.
;;  maininfo is the UVALUE structure from GEOMETER.
;;  Josh Colwell
;; July 12, 2004: Plots pixels, target body limb, and fixed RA/Dec grid.
;; July 16, 2004: fix bug in limb time, and add ALLBODIES keyword.
;; July 26, 2004: First cut at star handling. Draw bodies filled.
;; July 27, 2004: Add rings of Saturn, star magnitude limit.
;; September 8, 2004: Add text labeling.
;; September 10, 2004: Use the zoom switch in mainuval to adjust default zoom.
;; September 18, 2004: Fixed ring drawing.
;; September 26, 2004: Store object ids of certain elements so they can be
;; toggled on and off.
;; October 4, 2004: Add UPDATE keyword.
;; November 16, 2005: Change labeling of pixel for boresight pixel.

;; Create the Skymap iTool with its customized panel:
IF NOT KEYWORD_SET(UPDATE) THEN SKYMAP_TOOL

;; Get the tool's object reference for sending text labels to the vis layer:
idtool = ITGETCURRENT(TOOL=oTool)

;; Change the current_zoom property of the view so that we don't have
;; so much white space around it:
view = idtool+'/WINDOW/VIEW_1'
IF NOT KEYWORD_SET(UPDATE) THEN BEGIN
    void = oTool->doSetProperty(view, 'CURRENT_ZOOM', 1.5)
    oTool->setproperty, PTR_NEW(data), 'DATA'
    oTool->setproperty, PTR_NEW(maininfo), 'MAININFO'
    oTool->setproperty, PTR_NEW(i_time), 'ITIME'
ENDIF

oTool->commitactions

;; Put the plot time on the annotation layer:
time_label = data[i_time].utc
;; Not yet implemented. Need to find method to add text to annotation layer.


npixel = N_ELEMENTS(data[i_time].ra_full[0,*])

;; Indicate to the user which end of the slit is which by drawing the lowest 
;; numbered pixel filled in green:
selected_pixels = WHERE(maininfo.pxvec)
name = 'Pixel '+STRTRIM(STRING(selected_pixels[0]),2)

IF name EQ 'Pixel 64' THEN BEGIN
    CASE maininfo.channel OF
        1: name = 'UVIS EUV'
        2: name = 'UVIS EUV'
        3: name = 'UVIS Solar'
        4: name = 'UVIS FUV'
        5: name = 'UVIS FUV'
        6: name = 'UVIS FUV'
        7: name = 'UVIS HSP'
        8: name = 'UVIS HDAC'
        ELSE: name = 'UVIS'
    ENDCASE
ENDIF

;; Store the object IDs of all plotted pixels.
pixel_ids = ['']
IPLOT, [data[i_time].ra_full[1:*,0],data[i_time].ra_full[1,0]], $
    [data[i_time].dec_full[1:*,0],data[i_time].dec_full[1,0]], $
    NAME = name, /FILL_BACKGROUND, FILL_COLOR = [0,255,0], TRANSPARENCY = 50, $
    XTITLE = 'Right Ascension (Degrees)', YTITLE = 'Declination (Degrees)', $
    IDENTIFIER = skymap_id, /OVERPLOT
pixel_ids = [pixel_ids, (otool->getselecteditems())->getfullidentifier()]

;; Store the object IDs of all labels in the visualization layer.
label_ids = ['']

oTextp = OBJ_NEW('IDLitVisText')
oTool->Add, oTextp
gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
gTextp->SetProperty, STRING = name, LOCATION = $
                     [data[i_time].ra_full[1,0],data[i_time].dec_full[1,0]]
;; Store the ID for the pixel labels too so they toggle with the pixels.
pixel_ids = [pixel_ids, gtextp->getfullidentifier()]
label_ids = [label_ids, gtextp->getfullidentifier()]

;; Get the object reference for the data space so we can control the
;; plotting range.
dataspaceid = 'WINDOW/VIEW_1/VISUALIZATION LAYER/DATA SPACE'
oDataSpace = oTool->getbyidentifier(dataspaceid)
oDataSpace->setproperty, x_auto_update= 0B
oDataSpace->setProperty, y_auto_update= 0B

IF npixel GT 1 THEN BEGIN
    label_flag = 0
    FOR px_index = 1, npixel-1 DO BEGIN
        name = 'Pixel '+STRTRIM(STRING(selected_pixels[px_index]),2)
        IPLOT, [data[i_time].ra_full[1:*,px_index], $
                data[i_time].ra_full[1,px_index]], $
               [data[i_time].dec_full[1:*,px_index], $
                data[i_time].dec_full[1,px_index]], /OVERPLOT, NAME = name
        pixel_ids = [pixel_ids, (otool->getselecteditems())->$
                     getfullidentifier()]
                
        IF label_flag THEN BEGIN
            oTextp = OBJ_NEW('IDLitVisText')
            oTool->Add, oTextp
            gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
            gTextp->SetProperty, STRING = name, LOCATION = $
                                 [data[i_time].ra_full[1,px_index], $
                                  data[i_time].dec_full[1,px_index]]
            label_flag = 0
            label_ids = [label_ids, gtextp->getfullidentifier()]
            pixel_ids = [pixel_ids, gtextp->getfullidentifier()]
        ENDIF ELSE label_flag = 1
    ENDFOR
ENDIF
pixel_ids = pixel_ids[1:*]
otool->setproperty, PTR_NEW(pixel_ids), 'PIXELS'

;;*************************************************************************
;;                              BODIES
;;*************************************************************************
;; Now draw the target body:
;;Get the string value for the target body:
targetid = WIDGET_INFO(maininfo.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, maininfo.targetdrop, GET_VALUE = targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
CSPICE_BODN2C, target, targetid, found
IF NOT found THEN BEGIN
    target = 'SATURN'
    targetid = 699L 
    ;;This should only happen if the target is the rings.
ENDIF

;;Determine a color to fill in the target body on the plot:
CASE target OF
    'SATURN': body_color = [128,128,0]
    'TITAN': body_color = [240,140,0]
    ELSE: body_color = [50,50,150]
ENDCASE

;; The target has a lot of objects: limb, terminator, possibly rings. 
;; So need an array of ids for all the target-related objects.
target_ids = ['']

BODY_LIMB, target, 'CASSINI', data[i_time].et, 90, ra_limb, dec_limb, $
    limberror
IF NOT limberror THEN $
    IPLOT, [ra_limb,ra_limb[0]], [dec_limb,dec_limb[0]], /OVERPLOT, $
    /FILL_BACKGROUND, FILL_COLOR = body_color, TRANSPARENCY = 50, $
    NAME = target
target_ids = [target_ids, (otool->getselecteditems())->getfullidentifier()]

;; Get the target body limb limits and double it for the "zoomed in" option
;; (triple if target is Saturn to allow room for rings).
IF maininfo.skyzoom EQ 2 THEN BEGIN
    IF target EQ 'SATURN' THEN zoomscalefactor = 3. ELSE zoomscalefactor = 2.
    ra_min = MIN(ra_limb)
    ra_max = MAX(ra_limb)
    ra_center = ra_min+.5*(ra_max-ra_min)
    delta_ra = zoomscalefactor*(ra_max-ra_min) < 360.
    dec_min = MIN(dec_limb)
    dec_max = MAX(dec_limb)
    dec_center = dec_min+.5*(dec_max-dec_min)
    delta_dec = zoomscalefactor*(dec_max-dec_min) < 180.

    ra_min = (ra_center-.5*delta_ra) > 0.
    ra_max = (ra_min+delta_ra) < 360.
    dec_min = (dec_center-.5*delta_dec) > (-90.)
    dec_max = (dec_min+delta_dec) < 90.
ENDIF ELSE BEGIN
    ra_min = 0.
    ra_max = 360.
    dec_min = -90.
    dec_max = 90.
ENDELSE

;; Set the data space plot range based on zoom option:
oDataSpace->setProperty, x_minimum = ra_min
oDataSpace->setProperty, x_maximum = ra_max
oDataSpace->setProperty, y_minimum = dec_min
oDataSpace->setProperty, y_maximum = dec_max

;; Draw the terminator on the target body:
ra_term = [0D0]
dec_term = [0D0]
CSPICE_BODVAR, targetid, 'RADII', radii
CSPICE_SPKPOS, target, data[i_time].et, 'J2000', 'LT', 'CASSINI', scpos, ltime
CSPICE_SPKPOS, target, data[i_time].et, 'J2000', 'LT', 'SUN', sunpos, ltimesun
CSPICE_EDLIMB, radii[0], radii[1], radii[2], sunpos, term
;; term is a SPICE ellipse defining the terminator.
CSPICE_EL2CGV, term, center, v1, v2

;; Get RA/Dec points along the terminator of the body:
FOR j = 0, 71 DO BEGIN
    AZ = !DPI*2.0/71D0*j-!DPI
    VEC2 = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    termvec = scpos-vec2
    ;; This is actually a less then fully accurate check to
    ;; see if the terminator line is behind the limb...
    IF CSPICE_VNORM(termvec) LT CSPICE_VNORM(scpos) THEN BEGIN
        CSPICE_RECRAD, termVEC, RANGE, RA1, DEC1
        ra_term = [ra_term, RA1*!RADEG]
        dec_term = [dec_term, DEC1*!RADEG]
    ENDIF
ENDFOR
ra_term = ra_term[1:*]
dec_term = dec_term[1:*]
IPLOT, ra_term, dec_term, /OVERPLOT, NAME = 'Terminator', COLOR = [255,0,0]

target_ids = [target_ids, (otool->getselecteditems())->getfullidentifier()]

;;Label the target body. First, create a new VisText object:
oText1 = OBJ_NEW('IDLitVisText')
;;Add the object to the tool: Note that this puts the text object
;;into the visualization's dataspace rather than into the annotation
;;layer.
oTool->Add, oText1
;;Retrieve a reference to the IDLgrText object that underlies each
;;IDLitVisText object:
gText1 = oText1->Get(/ALL, ISA = 'IDLgrText')
;;Set properties on the grText object. In particular, set the location
;;for the grText object:
gText1->SetProperty, STRING = target, LOCATION = [ra_term[1],dec_term[1]]

;;Save the ID of this label.
label_ids = [label_ids, gtext1->getfullidentifier()]
target_ids = [target_ids, gtext1->getfullidentifier()]

;; Draw the limbs of all bodies:
body_codes = LONG([10, 399, 599, 699, 601, 602, 603, 604, 605, $
    606, 607, 608, 609])
nbodies = N_ELEMENTS(body_codes)
body_ids = ['']
FOR bodycount = 0, nbodies-1 DO BEGIN
    CSPICE_BODC2N, body_codes[bodycount], bodyname, found
    IF found AND bodyname NE target THEN BEGIN
        BODY_LIMB, bodyname, 'CASSINI', data[i_time].et, 36, ra_limb, $
                   dec_limb, limberror
        IF NOT limberror THEN $
            IPLOT, [ra_limb,ra_limb[0]], [dec_limb,dec_limb[0]], /OVERPLOT, $
                   /FILL_BACKGROUND, TRANSPARENCY = 50, NAME = bodyname

    ENDIF
;;If this body is a satellite, save its ID so it can be toggled on and off.
    IF body_codes[bodycount] GT 600 AND $
      body_codes[bodycount] LT 699 THEN BEGIN
        body_ids = [body_ids, (otool->getselecteditems())->$
                    getfullidentifier()]
        itsamoon = 1B
    ENDIF ELSE itsamoon = 0B

    ;;Label the body:
    oTextb = OBJ_NEW('IDLitVisText')
    oTool->Add, oTextb
    gTextb = oTextb->Get(/ALL, ISA = 'IDLgrText')
    gTextb->SetProperty, STRING = bodyname, LOCATION = [ra_limb[0],dec_limb[0]]
    label_ids = [label_ids, gtextb->getfullidentifier()]
    
    ;;Store the body label in the body_id list as well so it gets toggled
    ;;with the body.
    IF itsamoon THEN body_ids = [body_ids, gtextb->getfullidentifier()]
ENDFOR

;; That's it for labels on the vis layer, so set the label_ids property.
label_ids = label_ids[1:*]
otool->setproperty, PTR_NEW(label_ids), 'VISLABELS'

;;*************************************************************************
;;                              RINGS
;;*************************************************************************
;; Do Saturn's rings. For now, hardwire the critical ring boundaries.
;; Would be better eventually to read these values in from PCK.
ringrad = [66970.,74510.,92000.,117580.,122170.,133405.,133730.,136780.,$
    140180.,140270.]
nring = N_ELEMENTS(ringrad)    

CSPICE_SPKPOS, 'SATURN', data[i_time].et, 'J2000', 'LT', 'CASSINI', $
    scpos, lt_corr
;; Get the rotation matrix between J2000 and Saturn:
CSPICE_PXFORM, 'J2000', 'IAU_SATURN', data[i_time].et-lt_corr, inert2p
inert2p = TRANSPOSE(inert2p)
scpos_p = inert2p#scpos
CSPICE_BODVAR, 699L, 'RADII', satrad

nringpoints = 36
;;Use 36 points for each ring

FOR ringindex = 0, nring-1 DO BEGIN
  ringra = DBLARR(nringpoints)
  ringdec = DBLARR(nringpoints)

  FOR ringpoint = 0, nringpoints-1 DO BEGIN
  az = !DPI*2.0*DOUBLE(ringpoint)/DOUBLE(nringpoints)-!DPI
  rvec = [COS(az), SIN(az), 0.0]*ringrad[ringindex]
  ;; RVEC is the vector to the ring in body coordinates.

  vec = TRANSPOSE(inert2p)#(scpos_p+rvec)
  ;; VEC is the vector from the S/C to the ring in IREF.
  ;; Check to see if it intercepts the planet.

  CSPICE_SURFPT,-scpos_p,inert2p#VEC,satrad[0],satrad[1],satrad[2], $
    POINT, found
  HIDDEN = 0B
  IF found THEN BEGIN
    ;; See if the point is behind the planet.
    IF (CSPICE_VNORM(SCPOS) LT CSPICE_VNORM(VEC)) THEN HIDDEN = 1B
  ENDIF
  IF (NOT HIDDEN) THEN BEGIN
    CSPICE_RECRAD, VEC, RANGE, RA1, DEC1
    ringra[ringpoint] = RA1*!RADEG
    ringdec[ringpoint] = DEC1*!RADEG
  ENDIF
  ENDFOR    ;;ringpoint

  ;; Hidden parts of the rings have coordinate values of [0.,0.].
  ;; Identify which points are not hidden:
  good = WHERE(ringra NE 0D0)
  ;; Find where there is a break in good:
  void = MAX(good-SHIFT(good,1), breakloc)
  ;; Rearrange the order of points so the ring is drawn without 
  ;; connecting behind the body:
  IF void GT 1 THEN good = [good[breakloc:*],good[0:breakloc-1]]

  IPLOT, ringra[good], ringdec[good], /OVERPLOT, $
         NAME = 'Ring Edge'+STRING(ringindex)

  IF target EQ 'SATURN' THEN $
    target_ids = [target_ids, (otool->getselecteditems())->$
                  getfullidentifier()] ELSE $
    body_ids = [body_ids, (otool->getselecteditems())->$
                getfullidentifier()] 

ENDFOR  ;;ringindex

;; That's it for the target-related objects.
target_ids = target_ids[1:*]
otool->setproperty, PTR_NEW(target_ids), 'TARGET'

;; That's also it for the toggle-able bodies.
body_ids = body_ids[1:*]
otool->setproperty, PTR_NEW(body_ids), 'MOONS'
;;*************************************************************************
;;                                    GRID
;;*************************************************************************
IF NOT KEYWORD_SET(UPDATE) THEN BEGIN
;; Now draw some RA/Dec gridlines:
ra_inc = 30.
dec_inc = 30.

grid_ids = ['']
FOR ragrid = 360-ra_inc, ra_inc, -ra_inc DO BEGIN
    xcoords = [ragrid,ragrid]
    ycoords = [-90,90]
    IPLOT, xcoords, ycoords, /OVERPLOT, NAME = 'RA Lines'
    ra_id = (otool->getselecteditems())->getfullidentifier()
    grid_ids = [grid_ids, ra_id]
ENDFOR

FOR decgrid = -90+dec_inc, 90-dec_inc, dec_inc DO BEGIN
    IPLOT, [360,0],[decgrid,decgrid], /OVERPLOT, NAME = 'Dec Lines'
    dec_id = (otool->getselecteditems())->getfullidentifier()
    grid_ids = [grid_ids, dec_id]
ENDFOR

grid_ids = grid_ids[1:*]
otool->setproperty, PTR_NEW(grid_ids), 'GRID'

;;**************************************************************************
;;                                      STARS
;;**************************************************************************
IF KEYWORD_SET(STARS) THEN BEGIN
;; Load the star data file:
starfile = FILE_WHICH('star_table.dat')
IF FILE_TEST(starfile) THEN BEGIN
    RESTORE, starfile
    ostarids = WHERE(STRMID(STRTRIM(star_table.sp,2),0,1) EQ 'O')
    ostars = star_table[ostarids]
    bstarids = WHERE(STRMID(STRTRIM(star_table.sp,2),0,1) EQ 'B')
    bstars = star_table[bstarids]
    astarids = WHERE(STRMID(STRTRIM(star_table.sp,2),0,1) EQ 'A')
    astars = star_table[astarids]
    ;;Unfortunately symbol size must be passed in one at a time, so it's 
    ;;not easy to give each star a symbol size directly proportional to
    ;;its magnitude without making a separate iPlot call for each star.
    ;;So, break it down into 4 different magnitude categories:
    
    bright_o = WHERE(ostars.mag LE 1.0 AND ostars.mag LT starlim)
    bright2_o = WHERE(ostars.mag GT 1.0 AND ostars.mag LE 2.0 AND $
        ostars.mag LT starlim)
    mid_o = WHERE(ostars.mag GT 2.0 AND ostars.mag LE 3.0 AND $
        ostars.mag LT starlim)
    faint_o = WHERE(ostars.mag GT 3.0 AND ostars.mag LE starlim)
    
    bright_b = where(bstars.mag LE 1.0 AND bstars.mag LT starlim)
    bright2_b = WHERE(bstars.mag GT 1.0 AND bstars.mag LE 2.0 AND $
        bstars.mag LT starlim)
    mid_b = WHERE(bstars.mag GT 2.0 AND bstars.mag LE 3.0 AND $
        bstars.mag LT starlim)
    faint_b = WHERE(bstars.mag GT 3.0 AND bstars.mag LE starlim AND $
        bstars.mag LT starlim)
        
    bright_a = WHERE(astars.mag LE 1.0 AND astars.mag LT starlim)
    bright2_a = WHERE(astars.mag GT 1.0 AND astars.mag LE 2.0 AND $
        astars.mag LT starlim)
    mid_a = WHERE(astars.mag GT 2.0 AND astars.mag LE 3.0 AND $
        astars.mag LT starlim)
    faint_a = WHERE(astars.mag GT 3.0 AND astars.mag LE starlim)
    
    ;; We're going to have a button to hide and show all stars, so
    ;; in order to do that we need to remember the identifier of each
    ;; star plot object as it is created in the skymap itool. I think
    ;; that after each object is created it is "selected" so we can
    ;; use the getselecteditems method on the itool to get the object
    ;; reference for it. Then need an array to store the identifiers:
    starids = ['']
    ;;O Stars*************************************************************
    IF bright_o[0] NE -1 THEN BEGIN
        IPLOT, ostars[bright_o].ra, ostars[bright_o].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [0,0,255], SYM_SIZE = .7, $
        NAME = 'Very Bright O Stars'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF bright2_o[0] NE -1 THEN BEGIN
        IPLOT, ostars[bright2_o].ra, ostars[bright2_o].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [0,0,255], SYM_SIZE = .5, $
        NAME = 'O Stars 1-2'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF mid_o[0] NE -1 THEN BEGIN
        IPLOT, ostars[mid_o].ra, ostars[mid_o].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [0,0,255], SYM_SIZE = .3, $
        NAME = 'O Stars 2-3'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF faint_o[0] NE -1 THEN BEGIN
        IPLOT, ostars[faint_o].ra, ostars[faint_o].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [0,0,255], SYM_SIZE = .2, $
        NAME = 'O Stars >3'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    ;;B Stars*************************************************************
    IF bright_b[0] NE -1 THEN BEGIN
        IPLOT, bstars[bright_b].ra, bstars[bright_b].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [0,255,0], SYM_SIZE = .7, $
        NAME = 'Very Bright B Stars'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF bright2_b[0] NE -1 THEN BEGIN
        IPLOT, bstars[bright2_b].ra, bstars[bright2_b].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [0,255,0], SYM_SIZE = .5, $
        NAME = 'B Stars 1-2'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF mid_b[0] NE -1 THEN BEGIN
        IPLOT, bstars[mid_b].ra, bstars[mid_b].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [0,255,0], SYM_SIZE = .3, $
        NAME = 'B Stars 2-3'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF faint_b[0] NE -1 THEN BEGIN
        IPLOT, bstars[faint_b].ra, bstars[faint_b].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [0,255,0], SYM_SIZE = .2, $
        NAME = 'B Stars >3'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    ;;A Stars*************************************************************
    IF bright_a[0] NE -1 THEN BEGIN
        IPLOT, astars[bright_a].ra, astars[bright_a].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [255,0,0], SYM_SIZE = .7, $
        NAME = 'Very Bright A Stars'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF bright2_a[0] NE -1 THEN BEGIN
        IPLOT, astars[bright2_a].ra, astars[bright2_a].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [255,0,0], SYM_SIZE = .5, $
        NAME = 'A Stars 1-2'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF mid_a[0] NE -1 THEN BEGIN
        IPLOT, astars[mid_a].ra, astars[mid_a].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [255,0,0], SYM_SIZE = .3, $
        NAME = 'A Stars 2-3'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF faint_a[0] NE -1 THEN BEGIN
        IPLOT, astars[faint_a].ra, astars[faint_a].dec, /OVERPLOT, $
        LINESTYLE=6, SYM_INDEX = 1, SYM_COLOR = [255,0,0], SYM_SIZE = .2, $
        NAME = 'A Stars >3'
        id = (otool->getselecteditems())->getfullidentifier()
        starids = [starids, id]
    ENDIF

    IF N_ELEMENTS(starids) GT 1 THEN starids = starids[1:*]


    ;; Now need to associate this data with the panel user interface
    ;; so it is visible to the button that will allow user to hide
    ;; or show the star objects.
    otool->setproperty, PTR_NEW(starids), 'STARS'

ENDIF ;; If star file found
ENDIF ;; If star keyword set

ENDIF ;; UPDATE keyword set.

END
