@target_panel.pro
@target_panel_event.pro
@target_panel_callback.pro
@target_tool.pro
@target_tool__define.pro

;;File Version: 1.03
;;Last Modified: August 2, 2006
;;By: JEC

PRO GEOMETER_TARGET, data, maininfo, i_time, STARS = starlim, $
                     UPDATE = UPDATE, ZOOM = ZOOM, ALLPIXELS = $
                     ALLPIXELS

;;  >> Abstract <<
;;  Only to be called by GEOMETER.
;; Draws objects in the "U, V" coordinate system, which is the
;; sky plane coordinate system in normalized coordinates. The
;; sky plane is defined as being the plane normal to the line of sight
;; from Cassini to the center of the target body.
;; i_time is the index of the desired time for computing body data.
;; maininfo is the UVALUE structure from GEOMETER.
;; Keywords:
;; ZOOM: A Scale factor for the amount of sky to show, where a value
;; of 1 places the limb of the target body just inside the plot 
;; boundaries.
;; STARS: set equal to the limiting magnitude of stars to be drawn.
;; If it is set to a magnitude of less than -1 then stars will not
;; be drawn.
;; UPDATE: If this is an update, stars are not redrawn since their 
;; positions are fixed. 
;; ALLPIXELS: If set, only pixels are drawn, but pixels are drawn
;; for all times for which data exist.
;; 
;; Author:
;; Josh Colwell
;; Based on GEOMETER_SKYMAP.
;;
;; November 16, 2005: Change labeling of boresight pixel. Turn off iTool
;; axes. Expand grid to include rings.
;; June 14, 2006: Fix bug for non-Saturn targets (used to tie space
;; to ring coordinates).
;; August 2, 2006: Improve terminator drawing.

;; Create the TARGET iTool with its customized panel:
IF NOT KEYWORD_SET(UPDATE) THEN TARGET_TOOL

;; Get the tool's object reference for sending text labels to the vis layer:
idtool = ITGETCURRENT(TOOL=oTool)

;; Change the current_zoom property of the view so that we don't have
;; so much white space around it:
view = idtool+'/WINDOW/VIEW_1'
IF NOT KEYWORD_SET(UPDATE) THEN BEGIN
    void = oTool->doSetProperty(view, 'CURRENT_ZOOM', 1.0)
    oTool->setproperty, PTR_NEW(data), 'DATA'
    oTool->setproperty, PTR_NEW(maininfo), 'MAININFO'
    oTool->setproperty, PTR_NEW(i_time), 'ITIME'
;; The following is not working.
    windowid = idtool+'/WINDOW'
    void = oTool->doSetProperty(windowid, 'WIDTH', 500)
    void = oTool->doSetProperty(windowid, 'HEIGHT', 450)
    oTool->commitactions
ENDIF

;; Get the camera rotation matrix used to create perspective view:
CSPICE_EUL2M, !DPI/2D0, !DPI/2D0-data[i_time].dectarg*!DTOR, $
              data[i_time].ratarg*!DTOR, 3L, 2L, 3L, cam_mat

;;Get the string value for the target body:
targetid = WIDGET_INFO(maininfo.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, maininfo.targetdrop, GET_VALUE = targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
CSPICE_BODN2C, target, targetid, found
IF NOT found THEN BEGIN
    target = 'SATURN'
    targetid = 699L 
    ;;This should only happen if the target is the rings.
ENDIF

npixel = N_ELEMENTS(data[i_time].ra_full[0,*])
selected_pixels = WHERE(maininfo.pxvec)

;; Store the object IDs of all plotted pixels.
pixel_ids = ['']

;; Store the object IDs of all labels in the visualization layer.
label_ids = ['']

name = STRTRIM(STRING(selected_pixels[0]),2)

IF name EQ '64' THEN BEGIN
    CASE maininfo.channel OF
        1: name = 'UVIS EUV'
        2: name = 'UVIS EUV'
        3: name = 'UVIS Solar'
        4: name = 'UVIS FUV'
        5: name = 'UVIS FUV'
        6: name = 'UVIS FUV'
        7: name = 'UVIS HSP'
        8: name = 'UVIS HDAC'
        ELSE: name = 'UVIS'
    ENDCASE
ENDIF

IF KEYWORD_SET(ALLPIXELS) THEN BEGIN
    ntime = N_ELEMENTS(data.utc)
    FOR time_index = 0, ntime-1 DO BEGIN
;; First do the lowest numbered pixel that was selected.
        RD_2_UV, [data[time_index].ra_full[1:*,0], $
                  data[time_index].ra_full[1,0]], $
                 [data[time_index].dec_full[1:*,0], $
                  data[time_index].dec_full[1,0]], $
                 upx1, vpx1, behind, cam_mat
        
        IF NOT behind THEN BEGIN 
            IPLOT, upx1, vpx1, $
                   NAME = name, FILL_BACKGROUND = $
                   (time_index EQ i_time), FILL_COLOR = [200, 200, 0], $
                   TRANSPARENCY = 50, /OVERPLOT
            pixel_ids = [pixel_ids, (otool->getselecteditems())->$
                         getfullidentifier()]

            oTextp = OBJ_NEW('IDLitVisText')
            oTool->Add, oTextp
            gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
            gTextp->SetProperty, STRING = name, LOCATION = [upx1[0],vpx1[0]]
;; Store the ID for the pixel labels too so they toggle with the pixels.
            pixel_ids = [pixel_ids, gtextp->getfullidentifier()]
            label_ids = [label_ids, gtextp->getfullidentifier()]
        ENDIF
;; Now do any other selected pixels. Don't bother labeling all these
;; other pixels to avoid too much clutter.
        IF npixel GT 1 THEN BEGIN
            FOR px_index = 1, npixel-1 DO BEGIN
                RD_2_UV, [data[time_index].ra_full[1:*,px_index], $
                          data[time_index].ra_full[1,px_index]], $
                         [data[time_index].dec_full[1:*,px_index], $
                          data[time_index].dec_full[1,px_index]], $
                         upx, vpx, behind, cam_mat
                IF NOT behind THEN BEGIN
                    IPLOT, upx, vpx, /OVERPLOT, FILL_COLOR = $
                           [200, 200 ,0], $
                           FILL_BACKGROUND = (time_index EQ i_time), $
                           TRANSPARENCY = 50
                    pixel_ids = [pixel_ids, (otool->getselecteditems())->$
                                 getfullidentifier()]
                ENDIF
            ENDFOR
        ENDIF
    ENDFOR
    pixel_ids = pixel_ids[1:*]
    otool->setproperty, PTR_NEW(pixel_ids), 'PIXELS'
    RETURN
ENDIF ELSE BEGIN
;; Not an "ALLPIXELS" update, so we're drawing just the pixels for the
;; current plot time.

    RD_2_UV, [data[i_time].ra_full[1:*,0], data[i_time].ra_full[1,0]], $
             [data[i_time].dec_full[1:*,0], data[i_time].dec_full[1,0]], $
             upx1, vpx1, behind, cam_mat

    IF NOT behind THEN BEGIN 
        IPLOT, upx1, vpx1, $
               NAME = name, /FILL_BACKGROUND, FILL_COLOR = $
               [0,255,0], TRANSPARENCY = 50, $
               XTITLE = 'Right Ascension (Degrees)', $
               YTITLE = 'Declination (Degrees)', $
               IDENTIFIER = skymap_id, /OVERPLOT
        pixel_ids = [pixel_ids, (otool->getselecteditems())->$
                     getfullidentifier()]
    ENDIF

    oTextp = OBJ_NEW('IDLitVisText')
    oTool->Add, oTextp
    gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
    gTextp->SetProperty, STRING = name, LOCATION = [upx1[0],vpx1[0]]
    
;; Store the ID for the pixel labels too so they toggle with the pixels.
    pixel_ids = [pixel_ids, gtextp->getfullidentifier()]
    label_ids = [label_ids, gtextp->getfullidentifier()]
    
    IF npixel GT 1 THEN BEGIN
        label_flag = 0
        FOR px_index = 1, npixel-1 DO BEGIN
            name = STRTRIM(STRING(selected_pixels[px_index]),2)
            RD_2_UV, [data[i_time].ra_full[1:*,px_index], $
                      data[i_time].ra_full[1,px_index]], $
                     [data[i_time].dec_full[1:*,px_index], $
                      data[i_time].dec_full[1,px_index]], $
                     upx, vpx, behind, cam_mat
            IF NOT behind THEN BEGIN
                IPLOT, upx, vpx, /OVERPLOT, NAME = name
                pixel_ids = [pixel_ids, (otool->getselecteditems())->$
                             getfullidentifier()]
                
                IF label_flag THEN BEGIN
                    oTextp = OBJ_NEW('IDLitVisText')
                    oTool->Add, oTextp
                    gTextp = oTextp->Get(/ALL, ISA = 'IDLgrText')
                    gTextp->SetProperty, STRING = name, LOCATION = $
                                         [upx[0],vpx[0]]
                    label_flag = 0
                    label_ids = [label_ids, gtextp->getfullidentifier()]
                    pixel_ids = [pixel_ids, gtextp->getfullidentifier()]
                ENDIF ELSE label_flag = 1
            ENDIF
        ENDFOR
    ENDIF
ENDELSE

;; Get the object reference for the data space so we can control the
;; plotting range.
dataspaceid = 'WINDOW/VIEW_1/VISUALIZATION LAYER/DATA SPACE'
oDataSpace = oTool->getbyidentifier(dataspaceid)
oDataSpace->setProperty, anisotropic_scale_2d = 1.0

oTool->commitactions

pixel_ids = pixel_ids[1:*]
otool->setproperty, PTR_NEW(pixel_ids), 'PIXELS'

;;*************************************************************************
;;                              BODIES
;;*************************************************************************
;; Now draw the target body:
;;Determine a color to fill in the target body on the plot:
CASE target OF
    'SATURN': body_color = [128,128,0]
    'TITAN': body_color = [240,140,0]
    ELSE: body_color = [50,50,150]
ENDCASE

;; The target has a lot of objects: limb, terminator, possibly rings. 
;; So need an array of ids for all the target-related objects.
target_ids = ['']

BODY_LIMB, target, 'CASSINI', data[i_time].et, 90, ra_limb, dec_limb, $
    limberror

IF NOT limberror THEN BEGIN
    RD_2_UV, [ra_limb, ra_limb[0]], [dec_limb, dec_limb[0]], $
             u_limb, v_limb, behind, cam_mat
    IF NOT behind THEN BEGIN
        IPLOT, u_limb, v_limb, /OVERPLOT, $
               /FILL_BACKGROUND, FILL_COLOR = body_color, TRANSPARENCY = 50, $
               NAME = target
        target_ids = [target_ids, (otool->getselecteditems())->$
                      getfullidentifier()]
    ENDIF
ENDIF

;; Draw the terminator on the target body:
ra_term = [0D0]
dec_term = [0D0]
CSPICE_BODVAR, targetid, 'RADII', radii
CSPICE_SPKPOS, target, data[i_time].et, 'J2000', 'LT', 'CASSINI', scpos, ltime
CSPICE_SPKPOS, target, data[i_time].et, 'J2000', 'LT', 'SUN', sunpos, ltimesun
CSPICE_EDLIMB, radii[0], radii[1], radii[2], sunpos, term
;; term is a SPICE ellipse defining the terminator.
CSPICE_EL2CGV, term, center, v1, v2

;; Get RA/Dec points along the terminator of the body:
FOR j = 0, 71 DO BEGIN
    AZ = !DPI*2.0/71D0*j-!DPI
    VEC2 = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    termvec = scpos-vec2
    ;; This is actually a less then fully accurate check to
    ;; see if the terminator line is behind the limb...
    IF CSPICE_VNORM(termvec) LT CSPICE_VNORM(scpos) THEN BEGIN
        CSPICE_RECRAD, termVEC, RANGE, RA1, DEC1
        ra_term = [ra_term, RA1*!RADEG]
        dec_term = [dec_term, DEC1*!RADEG]
    ENDIF
ENDFOR
ra_term = ra_term[1:*]
dec_term = dec_term[1:*]
;; Remove discontinuity across the body:
delta_ra_term = ra_term-SHIFT(ra_term,1)
delta_ra_term = delta_ra_term[1:*]
void = MAX(ABS(delta_ra_term),breakloc)
;; Rearrange to avoid drawing across the body:
targ_ra_size = MAX(ra_limb)-MIN(ra_limb)
IF void GT targ_ra_size/2. THEN BEGIN
    ra_term_temp = [ra_term[breakloc:*], ra_term[0:breakloc-1]]
    dec_term_temp = [dec_term[breakloc:*], dec_term[0:breakloc-1]]
    ra_term = ra_term_temp[1:*]
    dec_term = dec_term_temp[1:*]
ENDIF

delta_dec_term = dec_term-SHIFT(dec_term,1)
delta_dec_term = delta_dec_term[1:*]
void = MAX(ABS(delta_dec_term),breakloc)
;; Rearrange to avoid drawing across the body:
targ_dec_size = MAX(dec_limb)-MIN(dec_limb)
IF void GT targ_dec_size/2. THEN BEGIN
    ra_term_temp = [ra_term[breakloc:*], ra_term[0:breakloc-1]]
    dec_term_temp = [dec_term[breakloc:*], dec_term[0:breakloc-1]]
    ra_term = ra_term_temp[1:*]
    dec_term = dec_term_temp[1:*]
ENDIF
RD_2_UV, ra_term, dec_term, u_term, v_term, behind, cam_mat
IF NOT behind THEN BEGIN
  IPLOT, u_term, v_term, /OVERPLOT, NAME = 'Terminator', COLOR = [255,0,0]
  target_ids = [target_ids, (otool->getselecteditems())->getfullidentifier()]

zoom:
IF KEYWORD_SET(ZOOM) THEN zoomscalefactor = zoom ELSE $
  zoomscalefactor = 2.

u_min = MIN(u_limb)
u_max = MAX(u_limb)
u_center = u_min+.5*(u_max-u_min)
delta_u = zoomscalefactor*(u_max-u_min)
v_min = MIN(v_limb)
v_max = MAX(v_limb)
v_center = v_min+.5*(v_max-v_min)
delta_v = zoomscalefactor*(v_max-v_min)

u_min = (u_center-.5*delta_u)
u_max = (u_min+delta_u)
v_min = (v_center-.5*delta_v)
v_max = (v_min+delta_v)

;; Set the data space plot range:
oDataSpace->setProperty, x_minimum = u_min
oDataSpace->setProperty, x_maximum = u_max
oDataSpace->setProperty, y_minimum = v_min
oDataSpace->setProperty, y_maximum = v_max
oDataSpace->setProperty, x_auto_update= 0B
oDataSpace->setProperty, y_auto_update= 0B

;; Turn off the itool axes:
idaxes = oDataSpace->findIdentifiers('*AXES')
oaxes = oTool->getByIdentifier(idaxes)

;; the name of the property I want to set.
oaxes->setProperty, hide = 1B

;;Label the target body. First, create a new VisText object:
  oText1 = OBJ_NEW('IDLitVisText')
;;Add the object to the tool: Note that this puts the text object
;;into the visualization's dataspace rather than into the annotation
;;layer.
  oTool->Add, oText1
;;Retrieve a reference to the IDLgrText object that underlies each
;;IDLitVisText object:
  gText1 = oText1->Get(/ALL, ISA = 'IDLgrText')
;;Set properties on the grText object. In particular, set the location
;;for the grText object:
  gText1->SetProperty, STRING = target, LOCATION = [u_term[1], v_term[1]]
  
;;Save the ID of this label.
  label_ids = [label_ids, gtext1->getfullidentifier()]
  target_ids = [target_ids, gtext1->getfullidentifier()]
ENDIF  ;; Not behind

;; Plot the sub-solar and sub-spacecraft points:
;; Convert data.subsolarlat, data.subsolarlon to RA/Dec, then to U/J.
CSPICE_SPKPOS, target, data[i_time].et, 'J2000', 'LT', 'CASSINI', scpos, ltime
CSPICE_RECRAD, scpos, dist, ra, dec
RD_2_UV, ra*!RADEG, dec*!RADEG, usub, vsub, behind, cam_mat
IPLOT, [-1000D0, usub], [-1000D0, vsub], /OVERPLOT, SYM_INDEX = 4, $
       SYM_COLOR = [0,0,255], LINESTYLE = 6
target_ids = [target_ids, (otool->getselecteditems())->getfullidentifier()]

;; Draw the limbs of all bodies:
body_codes = LONG([10, 399, 599, 699, 601, 602, 603, 604, 605, $
    606, 607, 608, 609])
nbodies = N_ELEMENTS(body_codes)
body_ids = ['']

for bodycount = 0, nbodies-1 DO BEGIN
    CSPICE_BODC2N, body_codes[bodycount], bodyname, found
    IF found AND bodyname NE target THEN BEGIN
        BODY_LIMB, bodyname, 'CASSINI', data[i_time].et, 36, ra_limb, $
                   dec_limb, limberror
        IF NOT limberror THEN BEGIN
            RD_2_UV, [ra_limb, ra_limb[0]], [dec_limb, dec_limb[0]], $
                     u_limb, v_limb, behind, cam_mat
            IF NOT behind THEN BEGIN
;; Need to put a test in here to see if the body is behind the main
;; target body.
              IPLOT, u_limb, v_limb, /OVERPLOT, $
                   /FILL_BACKGROUND, TRANSPARENCY = 50, NAME = bodyname
;;If this body is a satellite, save its ID so it can be toggled on and off.
              IF body_codes[bodycount] GT 600 AND $
                body_codes[bodycount] LT 699 THEN BEGIN
                  body_ids = [body_ids, (otool->getselecteditems())->$
                              getfullidentifier()]
                  itsamoon = 1B
              ENDIF ELSE itsamoon = 0B
              
              ;;Label the body:
              oTextb = OBJ_NEW('IDLitVisText')
              oTool->Add, oTextb
              gTextb = oTextb->Get(/ALL, ISA = 'IDLgrText')
              gTextb->SetProperty, STRING = bodyname, LOCATION = $
                                   [u_limb[0], v_limb[0]]
              label_ids = [label_ids, gtextb->getfullidentifier()]
              
              ;;Store the body label in the body_id list as well so it 
              ;;gets toggled with the body.
              IF itsamoon THEN body_ids = $
                [body_ids, gtextb->getfullidentifier()]
          ENDIF
      ENDIF
  ENDIF
ENDFOR

;;*************************************************************************
;;                              RINGS
;;*************************************************************************
;; Only do this if Saturn is not behind the viewer:
CSPICE_SPKPOS, 'SATURN', data[i_time].et, 'J2000', 'LT', 'CASSINI', $
    scpos, lt_corr
CSPICE_RECRAD, scpos, dist, rasat, decsat
RD_2_UV, rasat*!RADEG, decsat*!RADEG, usat, vsat, satbehind, cam_mat
IF NOT satbehind THEN BEGIN 
;; Do Saturn's rings. For now, hardwire the critical ring boundaries.
;; Would be better eventually to read these values in from PCK.
ringrad = [66970.,74510.,92000.,117580.,122170.,133405.,133730.,136780.,$
    140180.,140270.]
nring = N_ELEMENTS(ringrad)    

;; Get the rotation matrix between J2000 and Saturn:
CSPICE_PXFORM, 'J2000', 'IAU_SATURN', data[i_time].et-lt_corr, inert2p
inert2p = TRANSPOSE(inert2p)
scpos_p = inert2p#scpos
CSPICE_BODVAR, 699L, 'RADII', satrad

nringpoints = 360
;;Use 36 points for each ring

FOR ringindex = 0, nring-1 DO BEGIN
  ringra = DBLARR(nringpoints)
  ringdec = DBLARR(nringpoints)

  FOR ringpoint = 0, nringpoints-1 DO BEGIN
  az = !DPI*2.0*DOUBLE(ringpoint)/DOUBLE(nringpoints)-!DPI
  rvec = [COS(az), SIN(az), 0.0]*ringrad[ringindex]
  ;; RVEC is the vector to the ring in body coordinates.

  vec = TRANSPOSE(inert2p)#(scpos_p+rvec)
  ;; VEC is the vector from the S/C to the ring in IREF.
  ;; Check to see if it intercepts the planet.

  CSPICE_SURFPT,-scpos_p,inert2p#VEC,satrad[0],satrad[1],satrad[2], $
    POINT, found
  HIDDEN = 0B
  IF found THEN BEGIN
    ;; See if the point is behind the planet.
    IF (CSPICE_VNORM(SCPOS) LT CSPICE_VNORM(VEC)) THEN HIDDEN = 1B
  ENDIF
  IF (NOT HIDDEN) THEN BEGIN
    CSPICE_RECRAD, VEC, RANGE, RA1, DEC1
    ringra[ringpoint] = RA1*!RADEG
    ringdec[ringpoint] = DEC1*!RADEG
  ENDIF
  ENDFOR    ;;ringpoint

  ;; Hidden parts of the rings have coordinate values of [0.,0.].
  ;; Identify which points are not hidden:
  good = WHERE(ringra NE 0D0)
  ;; Find where there is a break in good:
  void = MAX(good-SHIFT(good,1), breakloc)
  ;; Rearrange the order of points so the ring is drawn without 
  ;; connecting behind the body:
  IF void GT 1 THEN good = [good[breakloc:*],good[0:breakloc-1]] ELSE $
    IF N_ELEMENTS(good) EQ nringpoints THEN good = [good, good[0]]

  RD_2_UV, ringra[good], ringdec[good], ringu, ringv, behind, cam_mat
  IPLOT, ringu, ringv, /OVERPLOT, $
         NAME = 'Ring Edge'+STRING(ringindex)

  IF target EQ 'SATURN' THEN $
    target_ids = [target_ids, (otool->getselecteditems())->$
                  getfullidentifier()] ELSE $
    body_ids = [body_ids, (otool->getselecteditems())->$
                getfullidentifier()] 

ENDFOR  ;;ringindex
ENDIF   ;;Saturn and rings not behind viewer.

;; That's it for the target-related objects.
target_ids = target_ids[1:*]
otool->setproperty, PTR_NEW(target_ids), 'TARGET'

;; That's also it for the toggle-able bodies.
IF N_ELEMENTS(body_ids) GT 1 THEN BEGIN
    body_ids = body_ids[1:*]
    otool->setproperty, PTR_NEW(body_ids), 'MOONS'
ENDIF

;;*************************************************************************
;;                                    GRID
;;*************************************************************************
;; Now draw some RA/Dec gridlines:
IF target EQ 'SATURN' THEN BEGIN
    u_min = u_min < MIN(ringu)
    u_max = u_max > MAX(ringu)
    v_min = v_min < MIN(ringv)
    v_max = v_max > MAX(ringv)
ENDIF 

UV_2_RD, u_min, v_min, ra_min, dec_min, cam_mat
UV_2_RD, u_max, v_max, ra_max, dec_max, cam_mat

grid_ids = ['']
;; The RA/Dec grid is distorted in the perspective view of Target, so
;; we need vectors for each line of constant RA and constant Dec.
grid_pnts = 40
racoords = FLTARR(grid_pnts)

ngrid = 5
;; Shift to nearest interval of 0.1 degrees to avoid label clutter:
ramingrid = ROUND(ra_min*10.)/10.
ramaxgrid = ROUND(ra_max*10.)/10.
ra_inc = ROUND((ramaxgrid-ramingrid)/FLOAT(ngrid)*10.)/10.

decmingrid = ROUND(dec_min*10.)/10.
decmaxgrid = ROUND(dec_max*10.)/10.
dec_inc = ROUND((decmaxgrid-decmingrid)/FLOAT(ngrid)*10.)/10.

;; Make sure we have enough grid lines to get to the edge of the plot,
;; after doing the rounding above.
IF ramingrid+ra_inc*(ngrid) LE ramaxgrid THEN nra = ngrid+1 $
  ELSE nra = ngrid
IF decmingrid+dec_inc*(ngrid) LE decmaxgrid THEN ndec = ngrid+1 $
  ELSE ndec = ngrid

;; Calculate the U,V coordinates of the grid lines.
FOR ragrid = 0, nra-1 DO BEGIN
    racoords[*] = ramingrid+ra_inc*ragrid
    deccoords = FINDGEN(grid_pnts)/(grid_pnts-1.)*(decmaxgrid-decmingrid)+$
                decmingrid
    RD_2_UV, racoords, deccoords, ucoords, vcoords, behind, cam_mat
    IPLOT, ucoords, vcoords, /OVERPLOT, NAME = 'RA Lines'+$
           STRTRIM(STRING(ragrid),2), COLOR = [0,255,0]
    ra_id = (otool->getselecteditems())->getfullidentifier()
    ;;Label the grid line:
    oTextb = OBJ_NEW('IDLitVisText')
    oTool->Add, oTextb
    gTextb = oTextb->Get(/ALL, ISA = 'IDLgrText')
    ;;We've rounded the values to nearest 0.1 degree, so just 
    ;;write the label to that precision as well.
    gridstring = STRTRIM(STRING(racoords[0]),2)
    decimalloc = STRPOS(gridstring,'.')
    gridstring = STRMID(gridstring,0,decimalloc+2)
    gTextb->SetProperty, STRING = gridstring, $
                         LOCATION = [ucoords[grid_pnts-1],$
                                     vcoords[grid_pnts-1]]
    label_ids = [label_ids, gtextb->getfullidentifier()]
    grid_ids = [grid_ids, ra_id, gtextb->getfullidentifier()]
ENDFOR

FOR decgrid = 0, ndec-1 DO BEGIN
    racoords = FINDGEN(grid_pnts)/(grid_pnts-1.)*(ramaxgrid-ramingrid)+$
               ramingrid
    deccoords[*] = decmingrid+dec_inc*decgrid
    RD_2_UV, racoords, deccoords, ucoords, vcoords, behind, cam_mat
    IPLOT, ucoords, vcoords, /OVERPLOT, NAME = 'Dec Lines'+$
           STRTRIM(STRING(decgrid),2), COLOR = [0, 255, 0]
    dec_id = (otool->getselecteditems())->getfullidentifier()
    ;;Label the grid line:
    oTextb = OBJ_NEW('IDLitVisText')
    oTool->Add, oTextb
    gTextb = oTextb->Get(/ALL, ISA = 'IDLgrText')
    ;;We've rounded the values to nearest 0.1 degree, so just 
    ;;write the label to that precision as well.
    gridstring = STRTRIM(STRING(deccoords[0]),2)
    decimalloc = STRPOS(gridstring,'.')
    gridstring = STRMID(gridstring,0,decimalloc+2)
    gTextb->SetProperty, STRING = gridstring, $
                         LOCATION = [ucoords[grid_pnts-1],$
                                     vcoords[grid_pnts-1]]
    label_ids = [label_ids, gtextb->getfullidentifier()]
    grid_ids = [grid_ids, dec_id, gtextb->getfullidentifier()]
ENDFOR

grid_ids = grid_ids[1:*]
otool->setproperty, PTR_NEW(grid_ids), 'GRID'

;; That's it for labels on the vis layer, so set the label_ids property.
label_ids = label_ids[1:*]
otool->setproperty, PTR_NEW(label_ids), 'VISLABELS'

;; Turn off axis tick marks.
axisids = oTool->findidentifiers('*AXIS*', /VISUALIZATIONS)
FOR axisnum = 0, N_ELEMENTS(axisids)-1 DO BEGIN
    oaxis = oTool->getbyidentifier(axisids[axisnum])
    oaxis->setProperty, major = 0, minor=0
ENDFOR
oTool->commitactions

;;**************************************************************************
;;                                      STARS
;;**************************************************************************
IF starlim LT -1 THEN stars = 0B ELSE stars = 1B
IF stars THEN BEGIN
;; Load the star data file:
starfile = FILE_WHICH('star_table.dat')
IF FILE_TEST(starfile) THEN BEGIN
    RESTORE, starfile
    ostarids = WHERE(STRMID(STRTRIM(star_table.sp,2),0,1) EQ 'O')
    ostars = star_table[ostarids]
    bstarids = WHERE(STRMID(STRTRIM(star_table.sp,2),0,1) EQ 'B')
    bstars = star_table[bstarids]
    astarids = WHERE(STRMID(STRTRIM(star_table.sp,2),0,1) EQ 'A')
    astars = star_table[astarids]
    ;;Unfortunately symbol size must be passed in one at a time, so it's 
    ;;not easy to give each star a symbol size directly proportional to
    ;;its magnitude without making a separate iPlot call for each star.
    ;;So, break it down into 4 different magnitude categories:
    
    bright_o = WHERE(ostars.mag LE 1.0 AND ostars.mag LT starlim)
    bright2_o = WHERE(ostars.mag GT 1.0 AND ostars.mag LE 2.0 AND $
        ostars.mag LT starlim)
    mid_o = WHERE(ostars.mag GT 2.0 AND ostars.mag LE 3.0 AND $
        ostars.mag LT starlim)
    faint_o = WHERE(ostars.mag GT 3.0 AND ostars.mag LE starlim)
    
    bright_b = where(bstars.mag LE 1.0 AND bstars.mag LT starlim)
    bright2_b = WHERE(bstars.mag GT 1.0 AND bstars.mag LE 2.0 AND $
        bstars.mag LT starlim)
    mid_b = WHERE(bstars.mag GT 2.0 AND bstars.mag LE 3.0 AND $
        bstars.mag LT starlim)
    faint_b = WHERE(bstars.mag GT 3.0 AND bstars.mag LE starlim AND $
        bstars.mag LT starlim)
        
    bright_a = WHERE(astars.mag LE 1.0 AND astars.mag LT starlim)
    bright2_a = WHERE(astars.mag GT 1.0 AND astars.mag LE 2.0 AND $
        astars.mag LT starlim)
    mid_a = WHERE(astars.mag GT 2.0 AND astars.mag LE 3.0 AND $
        astars.mag LT starlim)
    faint_a = WHERE(astars.mag GT 3.0 AND astars.mag LE starlim)
    
    ;; We're going to have a button to hide and show all stars, so
    ;; in order to do that we need to remember the identifier of each
    ;; star plot object as it is created in the skymap itool. I think
    ;; that after each object is created it is "selected" so we can
    ;; use the getselecteditems method on the itool to get the object
    ;; reference for it. Then need an array to store the identifiers:
    starids = ['']

    ;;O Stars*************************************************************
    IF bright_o[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright_o)-1 DO BEGIN
            RD_2_UV, (ostars[bright_o].ra)[star], $
                     (ostars[bright_o].dec)[star], $
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar,u[0]]
                vstar = [vstar,v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, $
                   SYM_COLOR = [0,0,255], SYM_SIZE = .7, $
                   NAME = 'Very Bright O Stars'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF bright2_o[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright2_o)-1 DO BEGIN
            RD_2_UV, (ostars[bright2_o].ra)[star], $
                     (ostars[bright2_o].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [0,0,255], SYM_SIZE = .5, $
                   NAME = 'O Stars 1-2'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF mid_o[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(mid_o)-1 DO BEGIN
            RD_2_UV, (ostars[mid_o].ra)[star], (ostars[mid_o].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [0,0,255], SYM_SIZE = .3, $
                   NAME = 'O Stars 2-3'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF faint_o[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(faint_o)-1 DO BEGIN
            RD_2_UV, (ostars[faint_o].ra)[star], (ostars[faint_o].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [0,0,255], SYM_SIZE = .2, $
                   NAME = 'O Stars >3'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    ;;B Stars*************************************************************
    IF bright_b[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright_b)-1 DO BEGIN
            RD_2_UV, (bstars[bright_b].ra)[star], $
                     (bstars[bright_b].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [0,255,0], SYM_SIZE = .7, $
                   NAME = 'Very Bright B Stars'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF bright2_b[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright2_b)-1 DO BEGIN
            RD_2_UV, (bstars[bright2_b].ra)[star], $
                     (bstars[bright2_b].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [0,255,0], SYM_SIZE = .5, $
                   NAME = 'B Stars 1-2'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF mid_b[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(mid_b)-1 DO BEGIN
            RD_2_UV, (bstars[mid_b].ra)[star], (bstars[mid_b].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [0,255,0], SYM_SIZE = .3, $
                   NAME = 'B Stars 2-3'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF faint_b[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(faint_b)-1 DO BEGIN
            RD_2_UV, (bstars[faint_b].ra)[star], (bstars[faint_b].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [0,255,0], SYM_SIZE = .2, $
                   NAME = 'B Stars >3'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    ;;A Stars*************************************************************
    IF bright_a[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright_a)-1 DO BEGIN
            RD_2_UV, (astars[bright_a].ra)[star], $
                     (astars[bright_a].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [255,0,0], SYM_SIZE = .7, $
                   NAME = 'Very Bright A Stars'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF bright2_a[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright2_a)-1 DO BEGIN
            RD_2_UV, (astars[bright2_a].ra)[star], $
                     (astars[bright2_a].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [255,0,0], SYM_SIZE = .5, $
                   NAME = 'A Stars 1-2'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF mid_a[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(mid_a)-1 DO BEGIN
            RD_2_UV, (astars[mid_a].ra)[star], (astars[mid_a].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [255,0,0], SYM_SIZE = .3, $
                   NAME = 'A Stars 2-3'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF faint_a[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(faint_a)-1 DO BEGIN
            RD_2_UV, (astars[faint_a].ra)[star], (astars[faint_a].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN BEGIN
            IPLOT, ustar[1:*], vstar[1:*], /OVERPLOT, LINESTYLE=6, $
                   SYM_INDEX = 1, SYM_COLOR = [255,0,0], SYM_SIZE = .2, $
                   NAME = 'A Stars >3'
            id = (otool->getselecteditems())->getfullidentifier()
            starids = [starids, id]
        ENDIF
    ENDIF

    IF N_ELEMENTS(starids) GT 1 THEN starids = starids[1:*]


    ;; Now need to associate this data with the panel user interface
    ;; so it is visible to the button that will allow user to hide
    ;; or show the star objects.
    otool->setproperty, PTR_NEW(starids), 'STARS'

ENDIF ;; If star file found
ENDIF ;; If star keyword set

;stop,'end of geometer_target'
END
