PRO TITAN_EXO_LIMB, observer, et, npts, ra_limb, dec_limb, limberror
;; Retrieve the RA and DEC of the Titan exobase
;; as seen from the observer at npts equally spaced
;; points around the limb. Based on BODY_LIMB.
;; Josh Colwell
;; January 3, 2007. 

ra_limb = FLTARR(npts)
dec_limb = FLTARR(npts)

limberror = 0B
CATCH, /CANCEL
CATCH, limb_error_status
IF STRMID(!ERROR_STATE.MSG, 0, 6) EQ 'CSPICE' THEN BEGIN
    limberror = 1B
    CATCH, /CANCEL
    MESSAGE, /RESET_ERROR_STATE
    RETURN
ENDIF

;; Get the state vector from the S/C to Titan
CSPICE_SPKPOS, 'TITAN', et, 'J2000', 'LT', observer, scpos, LT_CORR

radii = [4075., 4075., 4075.]

;; Use the NAIF routine EDLIMB to find the limb of the body
;; from the pov of the S/C.

CSPICE_EDLIMB, RADII(0), RADII(1), RADII(2), SCPOS, BODLIMB
CSPICE_EL2CGV, BODLIMB, CENTER, V1, V2

;; Get RA/Dec points along the limb of the body:
FOR j = 0, npts-1 DO BEGIN
    AZ = !DPI*2.0/DOUBLE(npts)*j-!DPI
    VEC2 = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    limbvec = scpos-vec2
    CSPICE_RECRAD, LIMBVEC, RANGE, RA1, DEC1
    ra_limb[j] = RA1*!RADEG
    dec_limb[j] = DEC1 *!RADEG
ENDFOR

RETURN
END
